﻿<#
.SYNOPSIS
  Vadlidate an XML file against XSD schema.
.DESCRIPTION
  Check the validity of an XML file against an XSD schema file, throw exception if not valid
.EXAMPLE
  ValidateXmlAgainstXsd .\HPCIaaSClusterConfig.xml .\HPCIaaSClusterConfig.xsd
#>
function ValidateXmlAgainstXsd
{
    Param
    (
        # The xml file path
        [Parameter(Mandatory=$true)]
        [String] $XmlFile,
		
        # The schema file path
        [Parameter(Mandatory=$true)]
        [String] $SchemaFile		
    )
	

    [string[]] $Script:XmlValidationErrorLog = @()
    [scriptblock] $xmlValidationEventHandler = {
        $Script:XmlValidationErrorLog += $args[1].Exception.Message
    }

    $xml = New-Object System.Xml.XmlDocument
    $schemaReader = New-Object System.Xml.XmlTextReader $SchemaFile
    try
    {
        $schema = [System.Xml.Schema.XmlSchema]::Read($schemaReader, $xmlValidationEventHandler)
        $xml.Schemas.Add($schema) | Out-Null
        $xml.Load($XmlFile)
        $xml.Validate($xmlValidationEventHandler)
    }
    catch
    {
        throw
    }
    finally
    {
        $schemaReader.Dispose()
    }

    if($Script:XmlValidationErrorLog.Length -gt 0)
    {
        throw ($Script:XmlValidationErrorLog -join "`r`n")
    }
}

function ValidatePasswordComplexity
{
    Param
    (
        [Parameter(Mandatory=$true)]
        [String] $Password
    )

    if($Password.Length -lt 8)
    {
        throw "The password must contain at least 8 characters"
    }

    $chCategories = 0
    if($Password -cmatch "[^0-9a-zA-Z]")
    {
        $chCategories++
    }

    if($Password -cmatch "[0-9]")
    {
        $chCategories++
    }

    if($Password -cmatch "[a-z]")
    {
        $chCategories++
    }

    if($Password -cmatch "[A-Z]")
    {
        $chCategories++
    }

    if($chCategories -lt 3)
    {
        throw "The password must contain 3 of the following: `n a lowercase character `n an uppercase character `n a number `n a special character"
    }
}

function ValidateUserName
{
    Param
    (
        [Parameter(Mandatory=$true)]
        [String] $UserName
    )

    $valid = $false
    do
    {
        if($UserName -match "^admin[1-2]{0,1}$")
        {
            break
        }

        if($UserName -match "^user[1-5]{0,1}$")
        {
            break
        }

        if($UserName -eq "a")
        {
            break
        }

        if($UserName -eq "1")
        {
            break
        }

        if($UserName -match '[\s, \\, /, \[, \], :, |, <, >, +, =, ;, \,, ?, *, @]')
        {
            break
        }

        $valid = $true
    } while ($false)

    if(-not $valid)
    {
        throw "The user name $UserName is not allowed. Please specify another one"
    }
}

function ValidateVNetNaming
{
    Param
    (
        # The VNet Name
        [Parameter(Mandatory=$true)]
        [String] $VNetName,

        # The Subnet Name
        [Parameter(Mandatory=$true)]
        [String] $SubnetName
    )

    $regex = "^[^0-9].{1,62}$"
    if($VNetName -notmatch $regex)
    {
        throw "The virtual network name $VNetName is invalid. It must contain between 2 and 63 characters, and cannot start with a number."
    }
    if($SubnetName -notmatch $regex)
    {
        throw "The subnet Name $SubnetName is invalid. It must contain between 2 and 63 characters, and cannot start with a number."
    }
}

function ValidateServiceNaming
{
    Param
    (
        # The Name
        [Parameter(Mandatory=$true)]
        [String] $ServiceName
    )
    
    $matched = $false
    do 
    {
        $regex = "^[0-9a-zA-Z-]{1,63}$"
        if($ServiceName -cnotmatch $regex)
        {
            break;
        }
        if($ServiceName[0] -eq '-')
        {
            break;
        }
        if($ServiceName[-1] -eq '-')
        {
            break;
        }

        $matched = $true
    } while($false)
    
    if(-not $matched)
    {
        throw "The cloud service name $ServiceName is invalid, It must meet the following criteria: `n 1. It must contain fewer than 63 characters;`n 2. It can contain only letters, numbers, and hyphens, and the first and last character must be a letter or number."
    }
}

function ValidateServiceNamingPattern
{
    Param
    (
        # The Name
        [Parameter(Mandatory=$true)]
        [String] $ServiceNamePattern,

        [Parameter(Mandatory=$true)]
        [int] $Count      
    )

    $matched = $false
    do 
    {
        $regex = "^[0-9a-zA-Z][0-9a-zA-Z-]{0,61}%[0-9]{1,62}%$"
        if($ServiceNamePattern -cnotmatch $regex)
        {
            break
        }
        if($ServiceNamePattern.Length -gt 65)
        {
            break
        }
        $matched = $true
    } while ($false)

    if(-not $matched)
    {
        throw "The cloud service naming pattern $ServiceNamePattern is invalid. It should be <root_name>%<start_number>% and meet the following criteria:`n 1. It must contain between 4 and 65 characters including the percent signs;`n 2. The root_name can contain only letters, numbers, and hyphens, and it must not start with a hyphen; `n 3. The start_number can contain only numbers; `nExample: AzureService%100%"
    }

    $result = ParseNamePattern -NamePattern $ServiceNamePattern
    $availNameCount = [System.Math]::Pow(10, $result[1]) - $result[2]
    if($availNameCount -lt $Count)
    {
        throw "Not enough name space in service naming pattern $ServiceNamePattern, $Count names required, only $availNameCount available"
    }
}

function ValidateStorageAccountNaming
{
    Param
    (
        # The Name
        [Parameter(Mandatory=$true)]
        [String] $StorageAccountName
    )
    
    $regex = "^[0-9a-z]{3,24}$"
    if($StorageAccountName -cnotmatch $regex)
    {
        throw "The storage account name $StorageAccountName is invalid. It must contain between 3 and 24 characters, and can contain only lowercase letters and numbers."
    }
}

function ValidateStorageAccountNamingPattern
{
    Param
    (
        # The Name
        [Parameter(Mandatory=$true)]
        [String] $StorageAccountNamePattern,

        [Parameter(Mandatory=$true)]
        [int] $Count      
    )

    $matched = $false
    do 
    {
        $regex = "^[0-9a-z]{1,23}%[0-9]{1,23}%$"
        if($StorageAccountNamePattern -cnotmatch $regex)
        {
            break
        }
        if(($StorageAccountNamePattern.Length -gt 26) -or ($StorageAccountNamePattern.Length -lt 5))
        {
            break
        }
        $matched = $true
    } while ($false)

    if(-not $matched)
    {
        throw "The storage account naming pattern $StorageAccountNamePattern is invalid. It should be <root_name>%<start_number>% and meet the following criteria:`n 1. It must contain between 5 and 26 characters including the percent signs;`n 2. The root_name can contain only lower-case letters and numbers; `n 3. The start_number can contain only numbers; `nExample: azurestorage%100%"
    }

    $result = ParseNamePattern -NamePattern $StorageAccountNamePattern
    $availNameCount = [System.Math]::Pow(10, $result[1]) - $result[2]
    if($availNameCount -lt $Count)
    {
        throw "Not enough name space in storage account naming pattern $StorageAccountNamePattern, $Count names required, only $availNameCount available"
    }
}

function ValidateAffinityGroupNaming
{
    Param
    (
        # The Name
        [Parameter(Mandatory=$true)]
        [String] $AffinityGroupName
    )
    
    $matched = $false
    do 
    {
        $regex = "^[0-9a-zA-Z][0-9a-zA-Z-]{0,62}$"
        if($AffinityGroupName -cnotmatch $regex)
        {
            break;
        }
        if($AffinityGroupName[-1] -eq '-')
        {
            break;
        }

        $matched = $true
    } while($false)
    
    if(-not $matched)
    {
        throw "The affinity group name $AffinityGroupName is invalid. It must meet the following criteria: `n 1. It must contain fewer than 63 characters;`n 2. It can contain only letters, numbers, and hyphens, and the first and last character must be a letter or number."
    }
}

function ValidateVMNaming
{
    Param
    (
        # The Name
        [Parameter(Mandatory=$true)]
        [String] $VMName
    )

    $regex = "^[a-zA-Z][a-zA-Z0-9-]{1,13}[a-zA-Z0-9]$"
    if($VMName -cnotmatch $regex)
    {
        throw "The VM Name $VMName is invalid. it must meet the following criteria: `n 1. It must contain between 3 and 15 characters; `n 2. It can contain only letters, numbers, and hyphens. `n 3. It must start with a letter and must end with a letter or a number."
    }
}


function ValidateVMPatternNaming
{
    Param
    (
        # The VM Name pattern
        [Parameter(Mandatory=$true)]
        [String] $VMNamePattern,

        [Parameter(Mandatory=$true)]
        [int] $NodeCount      
    )

    $matched = $false
    do 
    {
        $regex = "^[a-zA-Z][a-zA-Z0-9-]{0,13}%[0-9]{1,14}%$"
        if($VMNamePattern -cnotmatch $regex)
        {
            break
        }
        if(($VMNamePattern.Length -gt 17) -or ($VMNamePattern.Length -lt 5))
        {
            break
        }
        $matched = $true
    } while ($false)

    if(-not $matched)
    {
        throw "The VM naming pattern $VMNamePattern is invalid. It should be <root_name>%<start_number>% and meet the following criteria:`n 1. It must contain between 5 and 17 characters including the percent signs;`n 2. The root_name can contain only letters, numbers, and hyphens, and it must start with a letter `n 3. The start_number can contain only numbers; `nExample: AzureCN%1000%"
    }

    $result = ParseNamePattern -NamePattern $VMNamePattern
    $availNameCount = [System.Math]::Pow(10, $result[1]) - $result[2]
    if($availNameCount -lt $NodeCount)
    {
        throw "Not enough name space in VM naming pattern $VMNamePattern, $NodeCount names required, only $availNameCount available"
    }
}

function ValidateVMSize
{
    Param
    (
        # The VM Name
        [Parameter(Mandatory=$true)]
        [String] $VMName,

        [Parameter(Mandatory=$true)]
        [String] $VMSize,
        
        [Parameter(Mandatory=$true)]
        [String[]] $AvailableSizes              
    )

    if($AvailableSizes -cnotcontains $VMSize)
    {
        throw "The VM size '$VMSize' of VM $VMName is invalid. Valid sizes are: $([String]::Join(',', $AvailableSizes))"
    }
}


function ValidateVMExtensionVersion
{
    Param
    (
        [Parameter(Mandatory=$true)]
        [String] $Version
    )

    if(($ext.Version -notmatch "^[0-9]{1,5}\.\*$") -and ($ext.Version -notmatch "^[0-9]{1,5}\.[0-9]{1,5}$"))
    {
        throw "The Version $Version is invalid. It must be <major>.<minor>, and the minor can be '*'"
    }
}

function GetRegionalVNetLocation
{
    Param
    (
        # The Name
        [Parameter(Mandatory=$true)]
        [String] $VNetName
    )

    $uniqueID = [Guid]::NewGuid().ToString()
    $vnetCfgFile = "$env:temp\VNetCfg_$uniqueID.xml"

    InvokeAzureCmdWithRetry -Command "Get-AzureVNetConfig -ExportToFile `"$vnetCfgFile`"" | Out-Null
    if(Test-Path -Path $vnetCfgFile)
    {
        $xmlDoc = [xml](Get-Content $vnetCfgFile)
        Remove-Item $vnetCfgFile -Force

        $vnetSite = $xmlDoc.GetElementsByTagName("VirtualNetworkSite") | Where-Object {$_.name -eq $VNetName}
        if($null -ne $vnetSite)
        {
            if($vnetSite.HasAttribute("Location"))
            {
                return $vnetSite.Location
            }
            else
            {
                throw "Virtual network $VNetName is not a regional virtual network." 
            }
        }
    }

    throw "Virtual network $VNetName doesn't exist"
}

function CheckService
{
    Param
    (
        [Parameter(Mandatory=$true)]
        [String[]] $ServiceName,

        [Parameter(Mandatory=$true)]
        [String] $VNetName,

        [Parameter(Mandatory=$true)]
        [String] $Location,

        [Parameter(Mandatory=$true)]
        [HashTable] $ServiceAffinityGroupTable,
        
        # If not specified, we don't check the deployment label
        [Parameter(Mandatory=$false)]
        [String] $DeploymentLabel=""
    )

    $svcStateTable = @{}
    $errmsgs = @()
    $existingServices = @(InvokeAzureCmdWithRetry -Command "Get-AzureService")
    foreach($svcName in $ServiceName)
    {
        try
        {
            TraceInfo "Checking cloud service $svcName ..."
            $service = $existingServices | ?{$_.ServiceName -eq $svcName}
            if($null -ne $service)
            {
                $agName = $ServiceAffinityGroupTable[$svcName]
                if([String]::IsNullOrEmpty($agName))
                {
                    if($null -ne $service.AffinityGroup)
                    {
                        throw "The cloud service $svcName already exists but in an affinity group $($service.AffinityGroup), not directly in the region $Location."
                    }
                    elseif($service.Location -ne $Location)
                    {
                        throw "The cloud service $svcName already exists but in the region $($service.Location), not in the region $Location."
                    }
                }
                else
                {
                    if($null -eq $service.AffinityGroup)
                    {
                        throw "The cloud service $svcName already exists but not in the affinity group $agName."
                    }
                    if($service.AffinityGroup -ne $agName)
                    {
                        throw "The cloud service $svcName already exists but in affinity group $($service.AffinityGroup), not in affinity group $agName."
                    }
                }

                $deployment = InvokeAzureCmdWithRetry -Command "Get-AzureDeployment -ServiceName $svcName -WarningAction SilentlyContinue"
                if($null -ne $deployment)
                {
                    # Check VNet
                    if($deployment.VNetName -ne $VNetName)
                    {
                        throw "The production deployment of cloud service $ServiceName is not in virtual network $VNetName, but in virtual network $($deployment.VNetName)."
                    }
            
                    if((-not [String]::IsNullOrEmpty($DeploymentLabel)) -and ((DecodeLabelIfBase64Encoded $deployment.Label) -ne $DeploymentLabel))
                    {
                         throw "The cloud service $svcName is already in use by other production deployment."
                    }

                    TraceInfo "The cloud service $svcName already exists."
                    $svcStateTable[$svcName] = "Deployed"
                }
                else
                {
                    TraceInfo "The cloud service $svcName already exists."
                    $svcStateTable[$svcName] =  "Created"
                }
            }
            else
            {
                if((InvokeAzureCmdWithRetry -Command "Test-AzureName -Service -Name $svcName") -eq $true)
                {
                    throw "The cloud service name $svcName is not available, it is already in use."
                }

                TraceInfo "Cloud service $svcName doesn't exist"
                $svcStateTable[$svcName] = "NotCreated"
            }
        }
        catch
        {
            $errmsgs += $_.Exception.Message
        }
    }

    if($errmsgs.Count -ne 0)
    {
        throw ($errmsgs -join "`r`n")
    }

    return $svcStateTable
}

function ValidateVMConfig
{
    Param
    (
        [Parameter(Mandatory=$true)]
        [String] $VMName,

        [Parameter(Mandatory=$true)]
        [String] $ServiceName,

        [Parameter(Mandatory=$true)]
        [String] $SubnetName
    )

    $vmexists = $false
    $vm = InvokeAzureCmdWithRetry -Command "Get-AzureVM -Name $VMName -ServiceName $ServiceName"
    if($null -ne $vm)
    {
        $vmexists = $true
        $vmInSubnet = Get-AzureSubnet -VM $vm -ErrorAction SilentlyContinue
        if($vmInSubnet -ne $SubnetName)
        {
            throw "The virtual machine with same name $VMName already exists in cloud service $ServiceName, but not in Subnet $SubnetName"
        }
        TraceInfo "The virtual machine $VMName already exists"
    }
    else
    {
        TraceInfo "The virtual machine $VMName doesn't exist"
    }

    return $vmexists
}

<#
.SYNOPSIS
  Vadlidate input Cluster Config file
.DESCRIPTION
  Validate whether the input cluster config file is valid, throw exception if not valid
.EXAMPLE
  ConfigFileSemanticsCheck .\HPCIaaSClusterConfig.xml
#>
function ConfigFileSemanticsCheck
{
    Param
    (
        # Prefix name to name resources
        [Parameter(Mandatory=$true)]
        [String] $ConfigFile
    )
	
    if (-not (Test-Path -Path $ConfigFile))
    {
        throw "Config File $ConfigFile does not exist"
    }

    if(-not (Test-Path -Path "$PSScriptRoot\HPCIaaSClusterConfig.xsd"))
    {
        throw "XML Schema File HPCIaaSClusterConfig.xsd does not exist"
    }
    
    $xmlFullPath = (Resolve-Path -Path $ConfigFile).Path

    ## Validate the config file against schema
    ValidateXmlAgainstXsd -XmlFile $xmlFullPath -SchemaFile "$PSScriptRoot\HPCIaaSClusterConfig.xsd" | Out-Null

    $errormsgs = @()
    $warningmsgs = @()
    $config = ([Xml] (Get-Content $xmlFullPath)).IaaSClusterConfig
    $commonConfig = GetCommonConfig $config
    $domainConfig = GetDomainConfig $config.Domain
    $dbConfig = GetDatabaseConfig $config.Database
    $hnConfig = GetHeadNodeConfig $config.HeadNode
    $cnConfig = $null
    $bnConfig = $null
    $lnConfig = $null
    if($null -ne $config["ComputeNodes"])
    {
        $cnConfig = GetNodeGroupConfig $config.ComputeNodes
    }
    if($null -ne $config["BrokerNodes"])
    {
        $bnConfig = GetNodeGroupConfig $config.BrokerNodes
    }
    if($null -ne $config["LinuxComputeNodes"])
    {
        $lnConfig = GetNodeGroupConfig $config.LinuxComputeNodes
    }

    ## Validate the subscription
    $subNameExists = $false
    if($null -ne $commonConfig.SubscriptionName)
    {
        $subNameExists = $true 
    }

    if($null -eq $commonConfig.SubscriptionId)
    {
        if(-not $subNameExists)
        {
            $errormsgs += "Either 'SubscriptionName' or 'SubscriptionId' must be specified in 'Subscription'"
        }
    }
    else
    {
        if($subNameExists)
        {
            $errormsgs += "You cannot specify both 'SubscriptionName' and 'SubscriptionId', specify either of them in 'Subscription'"
        }
        else
        {
            try
            {
                [System.Guid]::Parse($commonConfig.SubscriptionId) | Out-Null
            }
            catch
            {
                $errormsgs += "The 'SubscriptionId' in 'Subscription' is not a valid Guid: $($commonConfig.SubscriptionId)"
            }
        }
    }

    ## Validate Domain options
    if($domainConfig.DCOption -eq "NewDC")
    {
        if($null -eq $domainConfig.DomainController)
        {
            $errormsgs += "'DomainController' must be specified when 'DCOption' is specified as 'NewDC'"
        }
    }
    # Workaround we allow ExistingDC to give DC VM information
    elseif($domainConfig.DCOption -eq "HeadNodeAsDC")
    {
        if($null -ne $domainConfig.DomainController)
        {
            $errormsgs += "'DomainController' must not be specified when 'DCOption' is specified as 'ExistingDC' or 'HeadNodeAsDC'"
        }
    }

    if($null -ne $domainConfig.DNSForwarder)
    {
        if($domainConfig.DCOption -eq "ExistingDC")
        {
            $errormsgs += "'DNSForwarder' must not be specified when 'DCOption' is specified as 'ExistingDC'"
        }
        else
        {
            try
            {
                 [System.Net.IPAddress]::Parse($domainConfig.DNSForwarder) | Out-Null
            }
            catch
            {
                $errormsgs += "'DNSForwarder' under 'Domain' is not a valid IP address: $($domainConfig.DNSForwarder)"
            }
        }
    }

    if($domainConfig.NoWindowsAutoUpdate -and ($domainConfig.DCOption -eq "ExistingDC"))
    {
        $errormsgs += "'NoWindowsAutoUpdate' must not be specified when 'DCOption' is specified as 'ExistingDC'"
    }
    

    ## Validate Database options

    if ($dbConfig.DBOption -eq "ExistingDB")
    {
        if($domainConfig.DCOption -ne "ExistingDC")
        {
            $errormsgs += "'ExistingDB' can only be specified when you already have an existing domain forest"
        }
        if($null -ne $dbConfig.DBVersion)
        {
            $errormsgs += "'DBVersion' must not be specified when 'DBOption' is specified as 'ExistingDB'"
        }
        if($null -eq $dbConfig.DBConnection)
        {
            $errormsgs += "'DBConnection' must be specified when 'DBOption' is specified as 'ExistingDB'"
        }
        else
        {
            if([String]::IsNullOrEmpty($dbConfig.DBConnection.ServerInstance))
            {
                $errormsgs += "'ServerInstance' must be specified when 'DBOption' is specified as 'ExistingDB'"
            }
        }
    }
    elseif($dbConfig.DBOption -eq "NewRemoteDB")
    {
        if($null -eq $dbConfig.DBServer)
        {
            $errormsgs += "'DBServer' must be specified when 'DBOption' is specified as 'NewRemoteDB'"
        }
        else
        {
            if(($dbConfig.DBServer.DataDiskSizeInGB -lt 100) -or ($dbConfig.DBServer.DataDiskSizeInGB -gt 1000))
            {
                $errormsgs += "The allowed value range of 'DataDiskSizeInGB' is 100 - 1000."
            }
        }

        if($null -eq $dbConfig.DBVersion)
        {
            $errormsgs += "'DBVersion' must be specified when 'DBOption' is specified as 'NewRemoteDB'"
        }
    }
    elseif($dbConfig.DBOption -eq "LocalDB")
    {
        if($null -ne $dbConfig.DBVersion)
        {
            $errormsgs += "'DBVersion' must not be specified when 'DBOption' is specified as 'LocalDB'"
        }
        if($null -ne $dbConfig.DBServer)
        {
            $errormsgs += "'DBServer' must not be specified when 'DBOption' is specified as 'LocalDB'"
        }
        if($null -ne $dbConfig.DBConnection)
        {
            $errormsgs += "'DBConnection' must not be specified when 'DBOption' is specified as 'LocalDB'"
        }
    }

    ## Validate head node options
    if($null -ne $hnConfig.ImageFamily)
    {
        if($hnConfig.ImageFamily -notmatch "HPC Pack")
        {
            $errormsgs += "The 'ImageFamily' in 'HeadNode' is not a valid HPC image family: $($hnConfig.ImageFamily)"
        }
    }

    ## Validate the VM Size Combination in Affinity Group
    $AGVMSizeMap = @{}
    $configDefaultAGName = ""
    if($null -ne $config["AffinityGroup"])
    {
        $configDefaultAGName = $config.AffinityGroup.Trim()
    }

    $vmConfigList = @($domainConfig.DomainController, $dbConfig.DBServer, $hnConfig, $cnConfig, $bnConfig, $lnConfig)
    foreach($vmConfig in $vmConfigList)
    {
        if($null -eq $vmConfig)
        {
            continue
        }

        $vmAGName = $vmConfig.AffinityGroup
        if($null -eq $vmConfig.AffinityGroup)
        {
            $vmAGName = $configDefaultAGName
        }

        $AGVMSizeMap[$vmAGName] += @($vmConfig.VMSize)
        $AGVMSizeMap[$vmAGName] = @($AGVMSizeMap[$vmAGName] | select -Unique)
    }
    
    $agErr = @()
    foreach($agname in $AGVMSizeMap.Keys)
    {
        if($agname -eq "")
        {
            continue
        }

        # All in RDMASizes or none in RDMASizes
        if(@($AGVMSizeMap[$agname] | % {IsVMSizeWithRDMA -VMSize $_} | select -Unique).Count -ne 1)
        {
            $agErr += ("$agname : {0}`n" -f ($AGVMSizeMap[$agname] -join " "))
        }
    }

    if($agErr.Count -gt 0)
    {
        $errormsgs += "VM size combination error in the following affinity group(s), A8/A9 sizes cannot co-exist with other sizes: `n$agErr"
    }

    ## Validate the naming of the config file
    $agCount = 0
    foreach($agname in $AGVMSizeMap.Keys)
    {
        if($agname -eq "")
        {
            continue
        }

        $agCount++
        try
        {
            ValidateAffinityGroupNaming -AffinityGroupName $agname | Out-Null
        }
        catch
        {
            $errormsgs += $_.Exception.Message
        }
    }

    if($agCount -gt 0)
    {
        $warningmsgs += "Affinity group is no longer recommended as a result of Azure network architectural improvements. It is recommended to remove 'AffinityGroup' from the configuration file."
    }

    try
    {
        ValidateStorageAccountNaming -StorageAccountName $config.Subscription.StorageAccount.Trim() | Out-Null
    }
    catch
    {
        $errormsgs += $_.Exception.Message
    }

    if($null -ne $config["VNet"])
    {
        $configVNetName = $config.VNet.VNetName.Trim()
        $configSubnetName = $config.VNet.SubnetName.Trim()
        try
        {
            ValidateVNetNaming -VNetName $configVNetName -SubnetName $configSubnetName | Out-Null
        }
        catch
        {
            $errormsgs += $_.Exception.Message
        }
    }

    if($null -ne $domainConfig.DomainController)
    {
        try
        {
            ValidateVMNaming -VMName $domainConfig.DomainController.VMName | Out-Null
        }
        catch
        {
            $errormsgs += $_.Exception.Message
        }

        try
        {
            ValidateServiceNaming -ServiceName $domainConfig.DomainController.ServiceName | Out-Null
        }
        catch
        {
            $errormsgs += $_.Exception.Message
        }
    }

    if($null -ne $dbConfig.DBServer)
    {
        try
        {
            ValidateVMNaming -VMName $dbConfig.DBServer.VMName | Out-Null
        }
        catch
        {
            $errormsgs += $_.Exception.Message
        }

        try
        {
            ValidateServiceNaming -ServiceName $dbConfig.DBServer.ServiceName | Out-Null
        }
        catch
        {
            $errormsgs += $_.Exception.Message
        }
    }

    if($null -ne $hnConfig)
    {
        try
        {
            ValidateVMNaming -VMName $hnConfig.VMName | Out-Null
        }
        catch
        {
            $errormsgs += $_.Exception.Message
        }

        try
        {
            ValidateServiceNaming -ServiceName $hnConfig.ServiceName | Out-Null
        }
        catch
        {
            $errormsgs += $_.Exception.Message
        }

        if($null -ne $hnConfig.PreConfigScript)
        {
            $configHNPreConfigScript = $hnConfig.PreConfigScript
            if (-not (Test-Path -Path $configHNPreConfigScript -PathType Leaf))
            {
                $errormsgs += "The 'PreConfigScript' file does not exist: $configHNPreConfigScript"
            }
            elseif(-not $configHNPreConfigScript.ToLower().EndsWith(".ps1"))
            {
                $errormsgs += "The 'PreConfigScript' file is not a PowerShell script file: $configHNPreConfigScript"
            }
        }

        if($null -ne $hnConfig.PostConfigScript)
        {
            $configHNPostConfigScript = $hnConfig.PostConfigScript
            if (-not (Test-Path -Path $configHNPostConfigScript -PathType Leaf))
            {
                $errormsgs += "The 'PostConfigScript' file does not exist: $configHNPostConfigScript"
            }
            elseif(-not $configHNPostConfigScript.ToLower().EndsWith(".ps1"))
            {
                $errormsgs += "The 'PostConfigScript' file is not a PowerShell script file: $configHNPostConfigScript"
            }
        }
    }

    foreach($nodeGroupConfig in @($cnConfig, $bnConfig, $lnConfig))
    {
        if($null -eq $nodeGroupConfig)
        {
            continue
        }

        try
        {
            ValidateVMPatternNaming -VMNamePattern $nodeGroupConfig.VMNamePattern -NodeCount $nodeGroupConfig.NodeCount | Out-Null
        }
        catch
        {
            $errormsgs += $_.Exception.Message            
        }

        if($null -ne $nodeGroupConfig.ServiceName)
        {
            try
            {
                ValidateServiceNaming -ServiceName $nodeGroupConfig.ServiceName | Out-Null
            }
            catch
            {
                $errormsgs += $_.Exception.Message            
            }

            if($nodeGroupConfig.NodeCount -gt 50)
            {
                $errormsgs += "The 'NodeCount' of $($nodeGroupConfig.NodeGroupName) has exceeded the maximum value 50: $($nodeGroupConfig.NodeCount)"
            }
        }
        else
        {
            # Currently, 'ServiceNamePattern' not supported for 'BrokerNodes'
            if($nodeGroupConfig.NodeGroupName -eq 'BrokerNodes')
            {
                $errormsgs += "'ServiceNamePattern' is not supported in $($nodeGroupConfig.NodeGroupName), please specify 'ServiceName' instead."
            }
            else
            {
                $svcNamePattern = $nodeGroupConfig.ServiceNamePattern
                $maxNodeCountPerSvc = $nodeGroupConfig.MaxNodeCountPerService
                $storageNamePattern = $nodeGroupConfig.StorageAccountNamePattern
                if(($maxNodeCountPerSvc -eq 0) -or ($maxNodeCountPerSvc -gt 50))
                {
                    $errormsgs += "The 'MaxNodeCountPerService' of $($nodeGroupConfig.NodeGroupName) must be between 1 and 50"
                }
                else
                {
                    $nodeSetCount = [Math]::Ceiling($nodeGroupConfig.NodeCount/$maxNodeCountPerSvc)
                    try
                    {
                        ValidateServiceNamingPattern -ServiceNamePattern $svcNamePattern -Count $nodeSetCount | Out-Null
                    }
                    catch
                    {
                        $errormsgs += $_.Exception.Message            
                    }
                    try
                    {
                        ValidateStorageAccountNamingPattern -StorageAccountNamePattern $storageNamePattern -Count $nodeSetCount | Out-Null
                    }
                    catch
                    {
                        $errormsgs += $_.Exception.Message            
                    }
                }
            }
        }

        if($null -ne $nodeGroupConfig.VMExtensions)
        {
            foreach($ext in $nodeGroupConfig.VMExtensions)
            {
                try
                {
                    ValidateVMExtensionVersion -Version $ext.Version | Out-Null
                }
                catch
                {
                    $errormsgs += ("Error in VMExtension $($ext.ExtensionName): " + $_.Exception.Message)
                }
            }
        }
    }

    if($null -ne $lnConfig)
    {
        if($null -ne $lnConfig.OSVersion)
        {
            $errormsgs +=  "'OSVersion' not supported for LinuxNodes."
        }
        if($null -eq $lnConfig.SSHKeyPairForRoot)
        {
            $warningmsgs += "'SSHKeyPairForRoot' is not specified in 'LinuxComputeNodes'. The script will generate an SSH key pair for test purposes. It is recommended to specify 'SSHKeyPairForRoot' for production deployments."
        }
    }

    ## Validate the Service/VM Name conflict
    $services = @{}
    $vmNames = @()
    # Collect the service/VM name info of DC, DB, HN
    foreach($vmConfig in @($domainConfig.DomainController, $dbConfig.DBServer, $hnConfig))
    {
        if($null -ne $vmConfig)
        {
            $vmAGName = $vmConfig.AffinityGroup
            if($null -eq $vmAGName)
            {
                $vmAGName = $configDefaultAGName
            }

            $services[$vmConfig.ServiceName] += @($vmAGName)
            $services[$vmConfig.ServiceName] = @($services[$vmConfig.ServiceName] | select -Unique)
            if($vmNames -contains $vmConfig.VMName)
            {
                $errormsgs += "There is more than one VM with same name: $($vmConfig.VMName)."
            }
            else
            {
                $vmNames += $vmConfig.VMName
            }
        }
    }

    # Validate the services /VM name for Node groups
    $svcNamePatterns = @()
    $vmNamePatterns = @()
    $nodeGroupSingleSvcs = @()
    foreach($nodeGroupConfig in @($cnConfig, $bnConfig, $lnConfig))
    {
        if($null -ne $nodeGroupConfig)
        {
            # Check VM Name confliction
            foreach($vmname in $vmNames)
            {
                if(NameMatchesPattern -Name $vmname -NamePattern $nodeGroupConfig.VMNamePattern)
                {
                    $errormsgs += "The VM name $vmname conflicts with 'VMNamePattern' of $($nodeGroupConfig.NodeGroupName): $($nodeGroupConfig.VMNamePattern)."
                }                
            }
            
            foreach($pattern in $vmNamePatterns)
            {
                if(NamePatternOverlapped -NamePattern1 $pattern -NamePattern2 $nodeGroupConfig.VMNamePattern)
                {
                    $errormsgs += "The 'VMNamePattern' of $($nodeGroupConfig.NodeGroupName) may overlap with other VMNamePattern: $($nodeGroupConfig.VMNamePattern) vs $pattern."
                }
            }

            $vmNamePatterns += $nodeGroupConfig.VMNamePattern

            # Check service confliction
            $vmAGName = $nodeGroupConfig.AffinityGroup
            if($null -eq $vmAGName)
            {
                $vmAGName = $configDefaultAGName
            }

            if($null -ne $nodeGroupConfig.ServiceName)
            {
                if($nodeGroupSingleSvcs -contains $nodeGroupConfig.ServiceName)
                {
                    $errormsgs += "ComputeNodes/BrokerNodes/LinuxComputeNodes cannot share cloud services with eatch other. The 'ServiceName' $($nodeGroupConfig.ServiceName) for $($nodeGroupConfig.NodeGroupName) is already used by another node group."
                }
                else
                {
                    if($services.Keys -contains $nodeGroupConfig.ServiceName)
                    {
                        $warningmsgs += "The 'ServiceName' $($nodeGroupConfig.ServiceName) for $($nodeGroupConfig.NodeGroupName) is already used by other node(s). It is recommended to specifiy a dedicated cloud service for $($nodeGroupConfig.NodeGroupName)"
                    }

                    $nodeGroupSingleSvcs += $nodeGroupConfig.ServiceName
                    $services[$nodeGroupConfig.ServiceName] += @($vmAGName)
                    $services[$nodeGroupConfig.ServiceName] = @($services[$nodeGroupConfig.ServiceName] | select -Unique)
                }

                foreach($pattern in $svcNamePatterns)
                {
                    if(NameMatchesPattern -Name $nodeGroupConfig.ServiceName -NamePattern $pattern)
                    {
                        $errormsgs += "ComputeNodes/BrokerNodes/LinuxComputeNodes cannot share cloud services with eatch other. The 'ServiceName' for $($nodeGroupConfig.NodeGroupName) conflicts with the 'ServiceNamePattern' for another node group: $($nodeGroupConfig.ServiceName) vs $pattern."
                    }
                }
            }
            else
            {
                foreach($svc in $services.Keys)
                {
                    if(NameMatchesPattern -Name $svc -NamePattern $nodeGroupConfig.ServiceNamePattern)
                    {
                        if($nodeGroupSingleSvcs -contains $svc)
                        {
                            $errormsgs += "ComputeNodes/BrokerNodes/LinuxComputeNodes cannot share cloud services with eatch other. The 'ServiceNamePattern' for $($nodeGroupConfig.NodeGroupName) conflicts with the 'ServiceName' for another node group: $($nodeGroupConfig.ServiceNamePattern) vs $svc."
                        }
                        else
                        {
                            $warningmsgs += "The 'ServiceNamePattern' for $($nodeGroupConfig.NodeGroupName) conflicts with the 'ServiceName' for another node group: $($nodeGroupConfig.ServiceNamePattern) vs $svc. It is recommended to specify dedicated cloud services for $($nodeGroupConfig.NodeGroupName)."
                        }
                    }
                }

                foreach($pattern in $svcNamePatterns)
                {
                    if(NamePatternOverlapped -NamePattern1 $pattern -NamePattern2 $nodeGroupConfig.ServiceNamePattern)
                    {
                        $errormsgs += "ComputeNodes/BrokerNodes/LinuxComputeNodes cannot share cloud services with eatch other, the 'ServiceNamePattern' for $($nodeGroupConfig.NodeGroupName) may overlap with that for another node group: $($nodeGroupConfig.ServiceNamePattern) vs $pattern."
                    }
                }

                $svcNamePatterns += $nodeGroupConfig.ServiceNamePattern
            }
        }
    }

    foreach($svc in $services.Keys)
    {
        if($services[$svc].Count -gt 1)
        {
            $errormsgs += ("The cloud service $svc cannot be specified in multiple affnity groups: " + $services[$svc])
        }
    }

    ### Check certificates
    $certList = @(GetCertificatesConfig -IaaSClusterConfig $config)
    $certIds = @($certList | %{$_.Id})
    $pfxFiles = @($certList | %{$_.PfxFile.ToLower()})
    foreach($group in ($certIds | group))
    {
        if($group.Count -gt 1)
        {
            $errormsgs += "[Certificate] There are two or more certificates with the same Id: " + $group.Name
        }
    }

    foreach($group in ($pfxFiles | group))
    {
        if($group.Count -gt 1)
        {
            $errormsgs += "[Certificate] The following certificate is specified multiple times: " + $group.Name
        }
    }

    if($certList.Count -gt 1)
    {
        if($null -ne $config["Certificates"] -and ($null -ne $config["AzureBurst"]) -and ($null -ne $config.AzureBurst["Certificate"]))
        {
            $warningmsgs += "The 'Certificate' elements are specified in both 'Certificates' and 'AzureBurst'. Please move all certificates from 'AzureBurst' to 'Certificates'"
        }
    }

    ### Check AzureBurst
    if($null -ne $config["AzureBurst"])
    {
        $errormsgs += CheckAzureBurstConfig -AzureBustConfig $config["AzureBurst"] -Certificates $certList
    }

    ### Check AutoGrowShink
    if($null -ne $config["AutoGrowShrink"])
    {
        $autoGrownShrinkConfig =  GetAutoGrowShrinkConfig $config["AutoGrowShrink"]
        if($certIds -notcontains $autoGrownShrinkConfig.CertificateId)
        {
            $errormsgs += "[AutoGrowShink] The 'CertificateId' $($autoGrownShrinkConfig.CertificateId) cannot be found."
        }
        if($autoGrownShrinkConfig.GrowInterval -ne 0 -and $autoGrownShrinkConfig.GrowInterval -lt 1)
        {
            $errormsgs += "[AutoGrowShink] The 'GrowInterval' should be no less than 1 minutes."
        }
        if($autoGrownShrinkConfig.ShrinkInterval -ne 0 -and $autoGrownShrinkConfig.ShrinkInterval -lt 2)
        {
            $errormsgs += "[AutoGrowShink] The 'ShrinkInterval' should be no less than 2 minutes."
        }
    }

    ### Conclusion
    return @{"Error"=$errormsgs;"Warning"=$warningmsgs}
}

function CheckStorageAccount
{
    param 
    (
        [Parameter(Mandatory=$true)]
        [String[]] $StorageAccountName,

        [Parameter(Mandatory=$true)]
        [String] $Location
    )

    $accountStatus = @{}
    $errMsg = @()
    $existingStorageAccounts = @(InvokeAzureCmdWithRetry -Command "Get-AzureStorageAccount")
    foreach($stg in $StorageAccountName)
    {
        TraceInfo "Checking storage account $stg"
        $storageAccount = $existingStorageAccounts | ?{$_.StorageAccountName -eq $stg}
        if($null -ne $storageAccount)
        {
            if($storageAccount.GeoPrimaryLocation -ne $Location)
            {
                $errMsg += "The storage account $stg already exists, but in region $($storageAccount.GeoPrimaryLocation), not in region $Location."
                continue
            }

            if($null -eq $storageAccount.AffinityGroup)
            {
                TraceInfo "The storage account $stg already exists in region $Location."
            }
            else
            {
                TraceInfo "The storage account $stg already exists in Affinity Group $($storageAccount.AffinityGroup)."
            }
        }
        else
        {
            if((InvokeAzureCmdWithRetry -Command "Test-AzureName -Storage -Name $stg") -eq $true)
            {
                $errMsg += "The storage account name $stg is not available. It is already in use."
                continue
            }
        }

        $accountStatus[$stg] = ($null -ne $storageAccount)
    }

    if($errMsg.Count -gt 0)
    {
        throw ($errMsg -join "`r`n")
    }

    return $accountStatus
}

function CheckLocationAndAffinityGroup
{
    param 
    (
        [Parameter(Mandatory=$true)]
        [String] $Location,

        [Parameter(Mandatory=$false)]
        [String[]] $AGNames
    )

    $locations = InvokeAzureCmdWithRetry -Command "Get-AzureLocation" | ForEach-Object {$_.Name}
    if($locations -notcontains $Location)
    {
        throw "Invalid Location: $Location."
    }
    else
    {
        TraceInfo "$Location is a valid Location."
    }

    $agStates = @{}
    if(($null -ne $AGNames) -and ($AGNames.Count -gt 0))
    {
        $errmsgs = @()
        foreach($agname in $AGNames)
        {
            $ag = InvokeAzureCmdWithRetry -Command "Get-AzureAffinityGroup -Name $agname"
            if($null -ne $ag)
            {
                if($ag.Location -ne $Location)
                {
                    $errmsgs += "The affinity group $agname already exists in region $($ag.Location), but not in region $Location."
                }
                else
                {
                    TraceInfo "The affinity group $agname already exists."
                    $agStates[$agname] = $true
                }
            }
            else
            {
                TraceInfo "The affinity group $agname doesn't exist yet."
                $agStates[$agname] = $false
            }
        }
        if($errmsgs.Count -gt 0)
        {
            throw ($errmsgs -join "`r`n")
        }
    }

    return $agStates
}

function CheckVNet
{
    param 
    (
        [Parameter(Mandatory=$true)]
        [String] $VNetName,

        [Parameter(Mandatory=$true)]
        [String] $SubnetName,

        [Parameter(Mandatory=$true)]
        [String] $Location
    )

    $vnetExists = $false
    $vnet = Get-AzureVNetSiteWithRetry -VNetName $VNetName -ErrorAction SilentlyContinue
    if ($null -ne $vnet)
    {
        $vnetExists = $true
        # The VNet already exists, check whether the settings are consistent
        # Decide whether it is a Regional VNet by Property AffinityGroup, for Regional VNet, AffinityGroup is $null
        if($null -eq $vnet.AffinityGroup)
        {
            # Check the location information for Regional VNet
            # Currently, we cannot get the Location by Get-AzureVNetSite command for Regional VNet
            $vnetLocation = GetRegionalVNetLocation -VNetName $VNetName
            if($vnetLocation -ne $Location)
            {
                throw "The virtual network $VNetName already exists but in region $vnetLocation, not in $Location."
            }

            TraceInfo "The virtual network $VNetName already exists in region $Location."                    
        }
        else
        {
            $vnetAG = $vnet.AffinityGroup
            throw "The virtual network $VNetName already exists but is associated with an affinity group $vnetAG."
        }

        ### Validate Subnet
        $subnet = $vnet.Subnets | Where-Object Name -eq $SubnetName
        if($null -eq $subnet)
        {
            throw "The virtual network $VNetName already exists but no subnet $SubnetName configured in it."
        }
        else
        {
            TraceInfo "The subnet $SubnetName already configured in $VNetName."
        }
    }
    else
    {
        TraceInfo "The virtual network $VNetName doesn't exist."
    }

    return $vnetExists
}

function CheckAzureVMImage
{
    param
    (
        [Parameter(Mandatory=$true)]
        [String] $ImageName,

        [Parameter(Mandatory=$true)]
        [ValidateSet("Windows","Linux")]
        [String] $OS,

        [Parameter(Mandatory=$false)]
        [String] $Location = "",

        # The image list to be searched, if not specified, use Get-AzureVMImage to get the image.
        [Parameter(Mandatory = $false)]
        [Object[]] $FromImageList = @()
    )

    if(($null -eq $FromImageList) -or ($FromImageList.Count -eq 0))
    {
        $image = InvokeAzureCmdWithRetry -Command "Get-AzureVMImage -ImageName $ImageName"
    }
    else
    {
        $image = $FromImageList | ?{$_.ImageName -eq $ImageName}
    }

    if($null -eq $image)
    {
        throw "The image $ImageName does not exist"
    }
    
    if($image.OS -ne $OS)
    {
        throw "The image $ImageName is not a $OS image, but a $($image.OS) image."
    }
    
    if(-not [String]::IsNullOrEmpty($Location))
    {
        $imgLocations = @()        
        if($null -ne $image.Location)
        {
            $imgLocations += $image.Location.Split(';')
        }
        elseif($null -ne $image.AffinityGroup)
        {
            $agName = $image.AffinityGroup
            $ag = InvokeAzureCmdWithRetry -Command "Get-AzureAffinityGroup -Name $agName"
            if($null -ne $ag)
            {
                $imgLocations += $ag.Location
            }
        }

        if($imgLocations -notcontains $Location)
        {
            throw "The image $ImageName found, but not in region $Location"
        }
    }
}

function CheckNodesInService
{
    param
    (
        [Parameter(Mandatory=$true)]
        [String] $ServiceName, 

        [Parameter(Mandatory=$true)]
        [String[]] $TargetNodeNames,

        [Parameter(Mandatory=$true)]
        [psobject] $NodeGroupConfig,

        [Parameter(Mandatory=$true)]
        [Switch] $UsePublicHPCImage
    )

    $ImageName = $NodeGroupConfig.ImageName
    TraceInfo "Checking virtual machines in cloud service $ServiceName"
    $VMsInService = @(InvokeAzureCmdWithRetry -Command "Get-AzureVM -ServiceName $ServiceName")
    # Check whether there is any VM other than the target VMs in the deployment
    $otherVMNames = @($VMsInService | ?{$TargetNodeNames -notcontains $_.Name} | %{$_.Name})
    if($otherVMNames.Count -gt 0)
    {
        TraceWarning "There are some other virtual machines in the cloud service $ServiceName : $otherVMNames"
    }

    $existingNodes = @($VMsInService | ?{$TargetNodeNames -contains $_.Name})
    # Check the source image of the existing VMs
    $imgNotMatch = @()
    $srcImgNames = @()
    foreach($vm in $existingNodes)
    {
        # There are two types of VM images: VMImage and OSImage
        # If the VM was created from VMImage, the SourceImageName is VMImageContext.OSDiskConfiguration.Name (typically as <ImageName>-os-yyyy-MM-dd)
        # If the VM was created from OSImage, the SourceImageName is OSImageContext.ImageName
        $srcImgName = $vm.VM.OSVirtualHardDisk.SourceImageName

        if($NodeGroupConfig.NodeType -eq "LinuxNode")
        {
            # For Linux Node, we just check whether VM was created from the given image or its replicas.
            if(-not [String]::IsNullOrEmpty($ImageName))
            {
                # if the given image is HPC installed, the source image name shall be <ImageName>[_<StorageAccountName>] or <ImageName>-os-yyyy-MM-dd
                if(($srcImgName -notmatch "^$ImageName(_[0-9a-z]{3,24})?$") -and ($srcImgName -notmatch "^$ImageName-os-"))
                {
                    $imgNotMatch += $vm.Name
                }
            }
        }
        else
        {
            $isPublicHpcImage = $false
            if($UsePublicHPCImage)
            {
                if($srcImgName -match "^[0-9a-f]{32}__HPC-Pack-")
                {
                    $isPublicHpcImage = $true
                }
                else
                {
                    TraceWarning "Virtual machine $($vm.Name) is not deployed with public HPC Image as expected. Checking script-generated image."
                }
            }
            
            # no need to return the image name or do any more check if the vm is provisioned from public HPC CN image
            if(-not $isPublicHpcImage)
            {
                # return the source image names if the node group is CN(WITHOUT default compute node image) or BN
                $srcImgNames += $srcImgName
                if(-not [String]::IsNullOrEmpty($ImageName) -and $NodeGroupConfig.ImageHPCPackInstalled)
                {
                    # if the given image is HPC installed, the source image name shall be <ImageName>[_<StorageAccountName>] or <ImageName>-os-yyyy-MM-dd
                    if(($srcImgName -notmatch "^$ImageName(_[0-9a-z]{3,24})?$") -and ($srcImgName -notmatch "^$ImageName-os-"))
                    {
                        $imgNotMatch += $vm.Name
                    }
                }
                else
                {
                    # if the given image is not HPC installed, means we will create the base HPC node image and create nodes with this base image.
                    # The source image name shall be HPC-Pack-<NodeType>_<DateTime>[_<StorageAccountName>] or HPC-Pack-<NodeType>_<DateTime>-os-yyyy-MM-dd
                    if(-not (IsBaseHPCNodeImage -VMSrcImageName $srcImgName -NodeType $NodeGroupConfig.NodeType))
                    {
                        $imgNotMatch += $vm.Name
                    }
                }
            }
        }
    }
    if($imgNotMatch.Count -gt 0)
    {
        throw ("The following virtual machines already exist in cloud service $ServiceName but not created by this script: " + ($imgNotMatch -join ","))
    }

    $readyVMNames = @()
    if($existingNodes.Count -gt 0)
    {
        $srcImgNames = @($srcImgNames | select -Unique | Sort-Object -Descending)
        $readyVMNames = @($existingNodes | %{$_.Name})
    }

    return @{"ReadyVMNames"=$readyVMNames;"VMImages"=$srcImgNames}
}

function CheckNodeGroup
{
    param
    (
        [Parameter(Mandatory=$true)]
        [psobject] $NodeGroupConfig, 

        [Parameter(Mandatory=$true)]
        [String[]] $AvailableVMSizes,
        
        [Parameter(Mandatory=$true)]
        [Object[]] $AvailableVMImages,

        [Parameter(Mandatory=$true)]
        [HashTable] $ServiceStateTable, 
        
        [Parameter(Mandatory=$true)]
        [String] $Location,
        
        [Parameter(Mandatory = $false)]
        [String] $Language = ""
    )

    try
    {
        ValidateVMSize -VMName $NodeGroupConfig.VMNamePattern -VMSize $NodeGroupConfig.VMSize -AvailableSizes $AvailableVMSizes
        
        $usePublicHPCImage = $false
        if(-not [String]::IsNullOrEmpty($NodeGroupConfig.ImageName))
        {
            if($NodeGroupConfig.NodeType -eq "LinuxNode")
            {
                CheckAzureVMImage -ImageName $NodeGroupConfig.ImageName -OS Linux -Location $Location -FromImageList $AvailableVMImages
            }
            elseif(IsWellKnownHPCImageName -ImageName $NodeGroupConfig.ImageName -NodeType $NodeGroupConfig.NodeType)
            {
                $realImgName = Get-LatestHPCImageName -WellKnownName $NodeGroupConfig.ImageName -FromImageList $AvailableVMImages -Language $Language
                if([string]::IsNullOrEmpty($realImgName))
                {
                    throw "The image $($NodeGroupConfig.ImageName) does not exist"
                }

                $NodeGroupConfig.ImageName = $realImgName
                $NodeGroupConfig.ImageHPCPackInstalled = $true
                $usePublicHPCImage = $true
            }
            else
            {
                CheckAzureVMImage -ImageName $NodeGroupConfig.ImageName -OS Windows -Location $Location -FromImageList $AvailableVMImages
            }            
        }
        else
        {
            # We use the default public HPC compute node image when both criteria were met: 1. ImageName not specified. 2. OSVersion not specified or specified as "WindowsServer2012R2" 
            if(($NodeGroupConfig.NodeType -eq "ComputeNode") -and (($null -eq $NodeGroupConfig.OSVersion) -or ("WindowsServer2012R2" -eq $NodeGroupConfig.OSVersion)))
            {
                $usePublicHPCImage = $true
            }
        }

        $srcImageNames = @()
        $serviceVMTable = @{}
        if($null -eq $NodeGroupConfig.ServiceName)
        {
            $nodeNameSets = @(GetNameSetsByPattern -NamePattern $NodeGroupConfig.VMNamePattern -Count $NodeGroupConfig.NodeCount -MaxCountPerSet $NodeGroupConfig.MaxNodeCountPerService)
            $nodeSetCount = $nodeNameSets.Count
            $serviceNames = @(GetNamesByPattern -NamePattern $NodeGroupConfig.ServiceNamePattern -Count $nodeSetCount)
            $storageNames = @(GetNamesByPattern -NamePattern $NodeGroupConfig.StorageAccountNamePattern -Count $nodeSetCount)
        }
        else
        {
            $nodeNameSets = @(, @(GetNamesByPattern -NamePattern $NodeGroupConfig.VMNamePattern -Count $NodeGroupConfig.NodeCount))
            $nodeSetCount = 1
            $serviceNames = @($NodeGroupConfig.ServiceName)
        }

        $totalReadyVMCount = 0
        for($i=0;$i -lt $nodeSetCount; $i++)
        {
            $serviceName = $serviceNames[$i]
            $targetNodeNames = $nodeNameSets[$i]
            if($ServiceStateTable[$serviceName] -eq "Deployed")
            {
                $result = CheckNodesInService -ServiceName $serviceName -TargetNodeNames $targetNodeNames -NodeGroupConfig $NodeGroupConfig -UsePublicHPCImage:$usePublicHPCImage
                $nodeNamesToCreate = @($targetNodeNames | ?{$result["ReadyVMNames"] -notcontains $_})
                $srcImageNames += $result["VMImages"]
                if($nodeNamesToCreate.Count -gt 0)
                {
                    $serviceVMTable[$serviceName] = $nodeNamesToCreate
                }

                $totalReadyVMCount += $result["ReadyVMNames"].Count
            }
            else
            {
                $serviceVMTable[$serviceName] = $targetNodeNames
            }
        }

        $hpcNodeImageName = ""
        if($NodeGroupConfig.ImageHPCPackInstalled -or ($NodeGroupConfig.NodeType -eq "LinuxNode"))
        {
            $hpcNodeImageName = $NodeGroupConfig.ImageName
        }
        elseif($srcImageNames.Count -gt 0)
        {
            $imageList = @(Get-BaseHPCNodeImage -NodeType $NodeGroupConfig.NodeType -Location $Location)
            $srcImages = @()
            if($imageList.Count -gt 0)
            {
                $srcImageNames = $srcImageNames | select -Unique
                foreach($srcImgName in $srcImageNames)
                {
                    $img = GetVMImageByVMSourceImageName -SourceImageName $srcImgName -FromImageList $imageList
                    if(($null -ne $img) -and (IsBaseHPCNodeImage -ImageName $img.ImageName -NodeType $NodeGroupConfig.NodeType))
                    {
                        $srcImages += $img
                    }
                }
            }

            if($srcImages.Count -gt 0)
            {
                # The format of the hpc node base image name is HPC-Pack-<NodeType>_<DateTime>[_<StorageAccountName>]
                # Sort the images in descending order by name(<DateTime>[_<StorageAccountName>] part), so that the latest image can be selected.
                $srcImages = $srcImages | Sort-Object -Property ImageName -Descending
                # The first image is probably with tailing _<StorageAccountName>, try to find an image with same <DateTime> but without _<StorageAccountName>.
                $firstImgStg = GetStorageAccountFromMediaLink -MediaLink (GetVMImageMediaLink $srcImages[0])
                $hpcNodeImageName = $srcImages[0].ImageName
                if($hpcNodeImageName -match "_$firstImgStg$")
                {
                    $firstImgNameRoot = $hpcNodeImageName.Substring(0, $hpcNodeImageName.Length - $firstImgStg.Length - 1)
                    $imgWithRootName = $srcImages | ?{$_.ImageName -eq $firstImgNameRoot}
                    if($null -ne $imgWithRootName)
                    {
                        $hpcNodeImageName = $imgWithRootName.ImageName
                    }                
                }
            }
        }

        TraceInfo "$totalReadyVMCount node(s) already exist, $($NodeGroupConfig.NodeCount - $totalReadyVMCount) node(s) will be created."

        return @($serviceVMTable, $hpcNodeImageName)
    }
    catch
    {
        throw
    }
}


function GetVMImageByVMSourceImageName
{
    param
    (
        [Parameter(Mandatory=$true)]
        [String]$SourceImageName,

        [Parameter(Mandatory=$true)]
        [Object[]] $FromImageList
    )
  
    if($null -ne $FromImageList -and $FromImageList.Count -gt 0)
    {
        foreach($img in $FromImageList)
        {
            if($img.ImageName -eq $SourceImageName)
            {
                return $img
            }
            if($img.PSObject.Properties.Name -contains "OSDiskConfiguration" -and $img.OSDiskConfiguration.PSObject.Properties.Name -contains "Name")
            {
                if($img.OSDiskConfiguration.Name -eq $SourceImageName)
                {
                    return $img
                }
            }
        }
    }

    return $null
}

function CheckAzureBurstConfig
{
    param
    (
        [Parameter(Mandatory=$true)]
        [System.Xml.XmlElement] $AzureBustConfig,

        [Parameter(Mandatory=$false)]
        [Object[]] $Certificates = @()
    )

    $errMsgs = @()
    $certIds = @()
    if($null -ne $Certificates)
    {
        $certIds = @($Certificates | %{$_.Id})
    }
    $tempNames = @()
    $svcNames = @()
    foreach($xmlTemp in $AzureBustConfig.AzureNodeTemplate)
    {
        $tempErrMsgs = @()
        $tempName = $xmlTemp.TemplateName.Trim()
        $tempNames += $tempName

        $svcName = $xmlTemp.ServiceName.Trim()
        try
        {
            ValidateServiceNaming -ServiceName $svcName | Out-Null
        }
        catch
        {
            $tempErrMsgs += $_.Exception.Message
        }

        try
        {
            ValidateStorageAccountNaming -StorageAccountName $xmlTemp.StorageAccount.Trim() | Out-Null
        }
        catch
        {
            $tempErrMsgs += $_.Exception.Message
        }

        $subId = $xmlTemp.SubscriptionId.Trim()
        try
        {
            [System.Guid]::Parse($subId) | Out-Null
        }
        catch
        {
            $tempErrMsgs += "The 'SubscriptionId' is not a valid Guid: $subId"
        }

        $svcNames += $svcName

        $certId = [Int]$xmlTemp.CertificateId.Trim()
        if($certIds -notcontains $certId)
        {
            $tempErrMsgs += "The 'CertificateId' $certId cannot be found."
        }

        if(($null -ne $xmlTemp["NodeCount"]) -and ($null -eq $xmlTemp["RoleSize"]))
        {
            $tempErrMsgs += "'RoleSize' must be specified when 'NodeCount' is specified."
        }

        if($null -ne $xmlTemp["Proxy"])
        {
            if($xmlTemp.Proxy.UsesStaticProxyCount.Trim() -eq "true")
            {
                if($null -ne $xmlTemp.Proxy["ProxyMultiplicity"])
                {
                    $proxyNum = [int]$xmlTemp.Proxy.ProxyMultiplicity.Trim()
                    if($proxyNum -gt 10 -or $proxyNum -lt 2)
                    {
                        $tempErrMsgs += "The value of 'ProxyMultiplicity' is invalid: $proxyNum (allowed range 2 - 10)."
                    }
                }
            }
            else
            {
                if($null -ne $xmlTemp.Proxy["ProxyRatio"])
                {
                    $prxoyRatio = [int]$xmlTemp.Proxy.ProxyRatio.Trim()
                    if($prxoyRatio -gt 400 -or $prxoyRatio -lt 100)
                    {
                        $tempErrMsgs += "The value of 'ProxyRatio' is invalid: $prxoyRatio (allowed range 100 - 400)."
                    }
                }
            
                if($null -ne $xmlTemp.Proxy["ProxyRatioBase"])
                {
                    $prxoyRatioBase = [int]$xmlTemp.Proxy.ProxyRatioBase.Trim()
                    if($prxoyRatioBase -gt 600 -or $prxoyRatioBase -lt 200)
                    {
                        $tempErrMsgs += "The value of 'ProxyRatioBase' is invalid: $prxoyRatioBase (allowed range 200 - 600)."
                    }
                }
            }
        }

        foreach($msg in $tempErrMsgs)
        {
            $errMsgs += "[AzureNodeTemplate $tempName] $msg"
        }

        $tempErrMsgs.Clear()
    }

    foreach($group in ($tempNames | group))
    {
        if($group.Count -gt 1)
        {
            $errMsgs += "[AzureNodeTemplate] There are two or more AzureNodeTemplates with the same 'TemplateName': " + $group.Name
        }
    }

    foreach($group in ($svcNames | group))
    {
        if($group.Count -gt 1)
        {
            $errMsgs += "[AzureNodeTemplate] The cloud service name $($group.Name) is specified in two or more AzureNodeTemplates."
        }
    } 
    
    return $errMsgs
}

function GetAutoGrowShrinkConfig
{
    param
    (
        [Parameter(Mandatory=$true)]
        [System.Xml.XmlElement] $AutoGrowShrinkConfig
    )
    $configInfo = @{}
    $configInfo["CertificateId"] = [Int]$AutoGrowShrinkConfig.CertificateId.Trim()
    ## 0 means default grow/shrink interval
    $configInfo["GrowInterval"] = 0
    $configInfo["ShrinkInterval"] = 0
    if($null -ne $AutoGrowShrinkConfig["GrowInterval"])
    {
        $configInfo["GrowInterval"] = [Int]$AutoGrowShrinkConfig.GrowInterval.Trim()
    }
    if($null -ne $AutoGrowShrinkConfig["ShrinkInterval"])
    {
        $configInfo["ShrinkInterval"] = [Int]$AutoGrowShrinkConfig.ShrinkInterval.Trim()
    }
    $configInfo['Enable'] = $true
    if(($null -ne $AutoGrowShrinkConfig["Enable"]) -and ($AutoGrowShrinkConfig.Enable.Trim() -eq "false"))
    {
        $configInfo['Enable'] = $false
    }

    $configPSObject = New-Object psobject -Property $configInfo
    return $configPSObject
}

function GetCertificatesConfig
{
    param
    (
        [Parameter(Mandatory=$true)]
        [System.Xml.XmlElement] $IaaSClusterConfig
    )
    $certList = @()
    $certElements = @()
    if($null -ne $IaaSClusterConfig["Certificates"])
    {
        $certElements += $IaaSClusterConfig.Certificates.Certificate
    }
    if(($null -ne $IaaSClusterConfig["AzureBurst"]) -and ($null -ne $IaaSClusterConfig.AzureBurst["Certificate"]))
    {
        $certElements += $IaaSClusterConfig.AzureBurst.Certificate
    }
    
    foreach($elem in $certElements)
    {
        $cert = @{}
        $cert["Id"] = [Int]$elem.Id.Trim()
        $cert["PfxFile"] = $elem.PfxFile.Trim()
        $cert["Password"] = $null
        if($null -ne $elem["Password"])
        {
            $cert["Password"] = $elem.Password.Trim()
        }

        $certConfig = New-Object psobject -Property $cert
        $certList += $certConfig
    }

    return $certList  
}


function PromptForPfxFilePassword
{
    param
    (
        [Parameter(Mandatory=$true)]
        [String] $FilePath, 

        [Parameter(Mandatory=$false)]
        [String] $Password = ""
    )
    
    $retry = 0
    $pswToTry = $Password
    while($true)
    {
        try
        {
            $pfxCert = GetCertificate $FilePath $pswToTry
            break
        }
        catch
        {
            if($_.Exception.HResult -eq 0x80070056)
            {
                if([String]::IsNullOrEmpty($pswToTry))
                {
                    $prompt = "The certificate file $FilePath is password protected. Please input the password"
                }
                else
                {
                    $prompt = "The password for the certificate file $FilePath is incorrect. Please input the correct password"
                }
                    
                if($retry -lt 3)
                {
                    $secPsw = Read-Host -Prompt $prompt -AsSecureString
                    $pswToTry = ConvertSecureStrToPlain -SecurePassword $secPsw 
                    $retry++
                }
                else
                {
                    throw "The password for the certificate file $FilePath is incorrect."
                }
            }
            elseif($_.Exception.HResult -eq 0x80092009)
            {
                throw "The file $FilePath is not a valid PFX file."
            }
            else
            {
                throw
            }
        }
    }
    
    if(-not $pfxCert.HasPrivateKey)
    {
        throw "The certificate file $pfxFile doesn't contain a private key."
    }
    return $pswToTry
}

function ValidateCertificates
{
    param
    (
        [Parameter(Mandatory=$true)]
        [Object[]] $Certificates
    )
    
    $certInfoList = @()
    foreach($cert in $Certificates)
    {   
        $pfxFile = $cert.PfxFile
        $password = ""
        if(-not [string]::IsNullOrEmpty($cert.Password))
        {
            $password = $cert.Password
        }

        $retry = 0
        while($true)
        {
            try
            {
                $pfxCert = GetCertificate $pfxFile $password
                break
            }
            catch
            {
                if($_.Exception.HResult -eq 0x80070056)
                {
                    if([String]::IsNullOrEmpty($password))
                    {
                        $prompt = "The certificate file $pfxFile is password protected. Please input the password"
                    }
                    else
                    {
                        $prompt = "The password for the certificate file $pfxFile is incorrect. Please input the correct password"
                    }
                    
                    if($retry -lt 3)
                    {
                        $secPsw = Read-Host -Prompt $prompt -AsSecureString
                        $password = ConvertSecureStrToPlain -SecurePassword $secPsw 
                        $retry++
                    }
                    else
                    {
                        throw "The password for the certificate file $pfxFile is incorrect."
                    }
                }
                elseif($_.Exception.HResult -eq 0x80092009)
                {
                    throw "The file $pfxFile is not a valid PFX file."
                }
                else
                {
                    throw
                }
            }
        }

        if(-not $pfxCert.HasPrivateKey)
        {
            throw "The certificate file $pfxFile doesn't contain a private key."
        }
        
        $certInfo = @{}
        $certInfo["Id"] = $cert.Id
        $certInfo["CertName"] = [IO.Path]::GetFileNameWithoutExtension($pfxFile)
        $certInfo["PfxContent"] = [IO.File]::ReadAllBytes($pfxFile)
        $certInfo["Password"] = $password
        $certInfo["Thumbprint"] = $pfxCert.Thumbprint
        $certInfoPSObject = New-Object psobject -Property $certInfo
        $certInfoList += $certInfoPSObject
    }

    return $certInfoList
}

function GetCommonConfig
{
    param
    (
        [Parameter(Mandatory=$true)]
        [System.Xml.XmlElement] $IaaSClusterConfig
    )
    
    $configInfo = @{}
    $configInfo['SubscriptionId'] = $null
    $configInfo['SubscriptionName'] = $null
    if($null -ne $IaaSClusterConfig.Subscription["SubscriptionName"])
    {
        $configInfo['SubscriptionName'] = $IaaSClusterConfig.Subscription.SubscriptionName.Trim()
    }

    if($null -ne $IaaSClusterConfig.Subscription["SubscriptionId"])
    {
        $configInfo['SubscriptionId'] = $IaaSClusterConfig.Subscription.SubscriptionId.Trim()
    }
    
    $configInfo['DefaultStorageAccount'] = $IaaSClusterConfig.Subscription.StorageAccount.Trim()
    $configInfo['Location'] = $IaaSClusterConfig.Location.Trim()
    $configInfo['Language'] = $null
    if($null -ne $config["Language"])
    {
        $configInfo['Language'] = $IaaSClusterConfig.Language.Trim()
    }
    $configInfo['DefaultAffinityGroup'] = $null
    if($null -ne $config["AffinityGroup"])
    {
        $configInfo['DefaultAffinityGroup'] = $IaaSClusterConfig.AffinityGroup.Trim()
    }

    $configInfo['VNetName'] = $IaaSClusterConfig.VNet.VNetName.Trim()
    $configInfo['SubnetName'] = $IaaSClusterConfig.VNet.SubnetName.Trim()

    # We put the DomainFQDN in 'Domain' to CommonConfig
    $configInfo['DomainFQDN'] = $IaaSClusterConfig.Domain.DomainFQDN.Trim()
    $configInfo['DomainNetBios'] = $configInfo['DomainFQDN'].Split(".")[0].ToUpper()

    $configPSObject = New-Object psobject -Property $configInfo
    return $configPSObject
}

function GetDomainConfig
{
    param
    (
        [Parameter(Mandatory=$true)]
        [System.Xml.XmlElement] $DomainConfig
    )
    
    $configInfo = @{}
    $configInfo['DomainFQDN'] = $DomainConfig.DomainFQDN.Trim()
    $configInfo['DomainNetBios'] = $configInfo['DomainFQDN'].Split(".")[0].ToUpper()
    $configInfo['DCOption'] = $DomainConfig.DCOption.Trim()
    $configInfo['DomainController'] = $null
    if($null -ne $DomainConfig['DomainController'])
    {
        $dcInfo = @{}
        $dcInfo['VMName'] = $DomainConfig.DomainController.VMName.Trim()
        $dcInfo['ServiceName'] = $DomainConfig.DomainController.ServiceName.Trim()
        $dcInfo['VMSize'] = $DomainConfig.DomainController.VMSize.Trim()
        $dcInfo['AffinityGroup'] = $null
        if($null -ne $DomainConfig.DomainController["AffinityGroup"])
        {
            $dcInfo['AffinityGroup'] = $DomainConfig.DomainController.AffinityGroup.Trim()
        }

        $configInfo['DomainController'] = New-Object psobject -Property $dcInfo
    }

    $configInfo['DNSForwarder'] = $null
    if($null -ne $DomainConfig['DNSForwarder'])
    {
        $configInfo['DNSForwarder'] = $DomainConfig.DNSForwarder.Trim()
    }

    $configInfo['NoWindowsAutoUpdate'] = $false
    if(($null -ne $DomainConfig["NoWindowsAutoUpdate"]) -and ($DomainConfig.NoWindowsAutoUpdate.Trim() -ne "false"))
    {
        $configInfo['NoWindowsAutoUpdate'] = $true
    }
    
    $configPSObject = New-Object psobject -Property $configInfo
    return $configPSObject
}

function GetDatabaseConfig
{
    param
    (
        [Parameter(Mandatory=$true)]
        [System.Xml.XmlElement] $DBConfig
    )
    
    $configInfo = @{}
    $configInfo['DBOption'] = $DBConfig.DBOption.Trim()
    $configInfo['DBServer'] = $null
    $configInfo['DBConnection'] = $null
    $configInfo['DBVersion'] = $null
    if($null -ne $DBConfig['DBVersion'])
    {
        $configInfo['DBVersion'] = $DBConfig.DBVersion.Trim()
    }

    if($null -ne $DBConfig['DBServer'])
    {
        $dbServerInfo = @{}
        $dbServerInfo['VMName'] = $DBConfig.DBServer.VMName.Trim()
        $dbServerInfo['ServiceName']  = $DBConfig.DBServer.ServiceName.Trim()
        $dbServerInfo['VMSize']  = $DBConfig.DBServer.VMSize.Trim()
        $dbServerInfo['AffinityGroup']  = $null
        if($null -ne $DBConfig.DBServer["AffinityGroup"])
        {
            $dbServerInfo['AffinityGroup'] = $DBConfig.DBServer.AffinityGroup.Trim()
        }

        $dbServerInfo['DataDiskSizeInGB'] = 200
        if($null -ne $DBConfig.DBServer["DataDiskSizeInGB"])
        {
            $dbServerInfo['DataDiskSizeInGB'] = [int]$DBConfig.DBServer.DataDiskSizeInGB.Trim()
        }

        $configInfo['DBServer'] = New-Object psobject -Property $dbServerInfo
    }
    
    if($null -ne $DBConfig['DBConnection'])
    {
        $dbConnectionInfo = @{}
        $dbConnectionInfo['ServerInstance'] = $DBConfig.DBConnection.ServerInstance.Trim()
        $dbConnectionInfo['ManagementDB'] = "HpcManagement"
        $dbConnectionInfo['SchedulerDB'] = "HpcScheduler"
        $dbConnectionInfo['MonitoringDB'] = "HpcMonitoring"
        $dbConnectionInfo['ReportingDB'] = "HpcReporting"
        $dbConnectionInfo['DiagnosticsDB'] = "HpcDiagnostics"
        if($null -ne $DBConfig.DBConnection["ManagementDB"])
        {
            $dbConnectionInfo['ManagementDB'] = $DBConfig.DBConnection.ManagementDB.Trim()
        }

        if($null -ne $DBConfig.DBConnection["SchedulerDB"])
        {
            $dbConnectionInfo['SchedulerDB'] = $DBConfig.DBConnection.SchedulerDB.Trim()
        }

        if($null -ne $DBConfig.DBConnection["MonitoringDB"])
        {
            $dbConnectionInfo['MonitoringDB'] = $DBConfig.DBConnection.MonitoringDB.Trim()
        }

        if($null -ne $DBConfig.DBConnection["ReportingDB"])
        {
            $dbConnectionInfo['ReportingDB'] = $DBConfig.DBConnection.ReportingDB.Trim()
        }

        if($null -ne $DBConfig.DBConnection["DiagnosticsDB"])
        {
            $dbConnectionInfo['DiagnosticsDB'] = $DBConfig.DBConnection.DiagnosticsDB.Trim()
        }

        $configInfo['DBConnection'] = New-Object psobject -Property $dbConnectionInfo
    }
    
    $configPSObject = New-Object psobject -Property $configInfo
    return $configPSObject
}


function GetHeadNodeConfig
{
    param
    (
        [Parameter(Mandatory=$true)]
        [System.Xml.XmlElement] $HeadNodeConfig
    )
    
    $configInfo = @{}
    $configInfo['VMName'] = $HeadNodeConfig.VMName.Trim()
    $configInfo['ServiceName'] = $HeadNodeConfig.ServiceName.Trim()
    $configInfo['VMSize'] = $HeadNodeConfig.VMSize.Trim()
    $configInfo['AffinityGroup'] = $null
    if($null -ne $HeadNodeConfig["AffinityGroup"])
    {
        $configInfo['AffinityGroup'] = $HeadNodeConfig.AffinityGroup.Trim()
    }

    $configInfo['ImageFamily'] = $null
    if($null -ne $HeadNodeConfig["ImageFamily"])
    {
        $configInfo['ImageFamily'] = $HeadNodeConfig.ImageFamily.Trim()
    }

    $configInfo['EnableRESTAPI'] = $false
    if(($null -ne $HeadNodeConfig["EnableRESTAPI"]) -and ($HeadNodeConfig.EnableRESTAPI.Trim() -ne "false"))
    {
        $configInfo['EnableRESTAPI'] = $true
    }

    $configInfo['EnableWebPortal'] = $false
    if(($null -ne $HeadNodeConfig["EnableWebPortal"]) -and ($HeadNodeConfig.EnableWebPortal.Trim() -ne "false"))
    {
        $configInfo['EnableWebPortal'] = $true
    }

    $configInfo['PreConfigScript'] = $null
    if($null -ne $HeadNodeConfig["PreConfigScript"])
    {
        $configInfo['PreConfigScript'] = $HeadNodeConfig.PreConfigScript.Trim()
    }

    $configInfo['PostConfigScript'] = $null
    if($null -ne $HeadNodeConfig["PostConfigScript"])
    {
        $configInfo['PostConfigScript'] = $HeadNodeConfig.PostConfigScript.Trim()
    }
    
    $configPSObject = New-Object psobject -Property $configInfo
    return $configPSObject
}

function GetNodeGroupConfig
{
    param
    (
        [Parameter(Mandatory=$true)]
        [System.Xml.XmlElement] $NodeGroupConfig
    )
    
    $configInfo = @{}
    $configInfo['NodeGroupName'] = $NodeGroupConfig.Name
    if($NodeGroupConfig.Name -eq "ComputeNodes")
    {
        $configInfo['NodeType'] = "ComputeNode"
    }
    elseif($NodeGroupConfig.Name -eq "BrokerNodes")
    {
        $configInfo['NodeType'] = "BrokerNode"
    }
    elseif($NodeGroupConfig.Name -eq "LinuxComputeNodes")
    {
        $configInfo['NodeType'] = "LinuxNode"
    }

    $configInfo['VMNamePattern'] = $NodeGroupConfig.VMNamePattern.Trim()
    $configInfo['VMSize'] = $NodeGroupConfig.VMSize.Trim()
    $configInfo['NodeCount'] = [int]$NodeGroupConfig.NodeCount.Trim()
    if($null -ne $NodeGroupConfig["ServiceName"])
    {
        $configInfo['ServiceName'] = $NodeGroupConfig.ServiceName.Trim()
        $configInfo['ServiceNamePattern'] = $null
    }
    else
    {
        $configInfo['ServiceName'] = $null
        $configInfo['ServiceNamePattern'] = $NodeGroupConfig.ServiceNamePattern.Trim()
        $configInfo['MaxNodeCountPerService'] = [int]$NodeGroupConfig.MaxNodeCountPerService.Trim()
        $configInfo['StorageAccountNamePattern'] = $NodeGroupConfig.StorageAccountNamePattern.Trim()
    }

    $configInfo['AffinityGroup'] = $null
    if($null -ne $NodeGroupConfig["AffinityGroup"])
    {
        $configInfo['AffinityGroup'] = $NodeGroupConfig.AffinityGroup.Trim()
    }

    $configInfo['ImageName'] = $null
    $configInfo['ImageHPCPackInstalled'] = $false
    $configInfo['OSVersion'] = $null
    if($null -ne $NodeGroupConfig["ImageName"])
    {
        $configInfo['ImageName'] = $NodeGroupConfig["ImageName"].InnerText.Trim()
        if($NodeGroupConfig["ImageName"].HasAttribute("HPCPackInstalled"))
        {
            if($NodeGroupConfig["ImageName"].GetAttribute("HPCPackInstalled").Trim() -eq "true")
            {
                $configInfo['ImageHPCPackInstalled'] = $true
            }
        }
    }
    elseif($null -ne $NodeGroupConfig["OSVersion"])
    {
        $configInfo['OSVersion'] = $NodeGroupConfig.OSVersion.Trim()
    }
    
    $configInfo['VMExtensions'] = $null
    if($null -ne $NodeGroupConfig["VMExtensions"])
    {
        $extensions = @()
        foreach($xmlExt in $NodeGroupConfig.VMExtensions.VMExtension)
        {
            $extName = $xmlExt.ExtensionName.Trim()
            $extPublisher = $xmlExt.Publisher.Trim()
            $extVersion = $xmlExt.Version.Trim()
            $extPubConfig = $null
            $extPrivateConfig = $null
            if($null -ne $xmlExt["PublicConfiguration"])
            {
                $extPubConfig = $xmlExt.PublicConfiguration.Trim()
            }

            if($null -ne $xmlExt["PrivateConfiguration"])
            {
                $extPrivateConfig = $xmlExt.PrivateConfiguration.Trim()
            }
           
            $extensions += NewVMExtensionConfig -ExtensionName $extName -Publisher $extPublisher -Version $extVersion -PublicConfiguration $extPubConfig -PrivateConfiguration $extPrivateConfig
        }

        $configInfo['VMExtensions'] = $extensions
    }

    $configInfo['SSHKeyPairForRoot'] = $null
    if($null -ne $NodeGroupConfig["SSHKeyPairForRoot"])
    {
        $sshCertificate = @{}
        $sshCertificate["PfxFile"] = $NodeGroupConfig.SSHKeyPairForRoot.PfxFile.Trim()
        if($null -ne $NodeGroupConfig.SSHKeyPairForRoot["Password"])
        {
            $sshCertificate["Password"] = $NodeGroupConfig.SSHKeyPairForRoot.Password.Trim()
        }
        else
        {
            $sshCertificate["Password"] = $null
        }

        $configInfo['SSHKeyPairForRoot'] = New-Object psobject -Property $sshCertificate
    }
        
    $configPSObject = New-Object psobject -Property $configInfo
    return $configPSObject
}

function GetNodeGroupServiceStorageTable
{
    param
    (
        [Parameter(Mandatory=$true)]
        [psobject] $NodeGroup,

        [Parameter(Mandatory=$true)]
        [string] $DefaultStorageAccount
    )

    $serviceStorageTable = @{}
    if($null -ne $NodeGroup.ServiceName)
    {
        $serviceStorageTable[$NodeGroup.ServiceName] = $DefaultStorageAccount
    }
    else
    {
        $nodeSetCount = [Math]::Ceiling($NodeGroup.NodeCount/$NodeGroup.MaxNodeCountPerService)
        $serviceNames = @(GetNamesByPattern -NamePattern $NodeGroup.ServiceNamePattern -Count $nodeSetCount)
        $storageAccounts = @(GetNamesByPattern -NamePattern $NodeGroup.StorageAccountNamePattern -Count $nodeSetCount)
        for($i=0; $i -lt $nodeSetCount; $i++)
        {
            $serviceStorageTable[$serviceNames[$i]] = $storageAccounts[$i]
        }
    }

    return $serviceStorageTable
}
# SIG # Begin signature block
# MIIaxgYJKoZIhvcNAQcCoIIatzCCGrMCAQExCzAJBgUrDgMCGgUAMGkGCisGAQQB
# gjcCAQSgWzBZMDQGCisGAQQBgjcCAR4wJgIDAQAABBAfzDtgWUsITrck0sYpfvNR
# AgEAAgEAAgEAAgEAAgEAMCEwCQYFKw4DAhoFAAQUSW0TKb67GaT8HZuDK++V3i2e
# poqgghWBMIIEwjCCA6qgAwIBAgITMwAAAJJMoq9VJwgudQAAAAAAkjANBgkqhkiG
# 9w0BAQUFADB3MQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4G
# A1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSEw
# HwYDVQQDExhNaWNyb3NvZnQgVGltZS1TdGFtcCBQQ0EwHhcNMTUxMDA3MTgxNDE0
# WhcNMTcwMTA3MTgxNDE0WjCBsjELMAkGA1UEBhMCVVMxEjAQBgNVBAgTCVdhc2lu
# Z3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBv
# cmF0aW9uMQ0wCwYDVQQLEwRNT1BSMScwJQYDVQQLEx5uQ2lwaGVyIERTRSBFU046
# N0QyRS0zNzgyLUIwRjcxJTAjBgNVBAMTHE1pY3Jvc29mdCBUaW1lLVN0YW1wIFNl
# cnZpY2UwggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQC6WVT9G7wUxF8u
# /fnFTnid7MCYX4X58613PUnaf2uYaz291cpmbxNeEsx+HZ8xrgjCHkMC3U9rTUfl
# oyhWqlW3ZdZQdn97Qa++X7wXa/ybE8FeY0Qphe8K0w9hbhxRjbII4fInEEkM4GAd
# HLqPqQw+U+Ul/gAC8U64SnklxtsjxN2faP98po9YqDYGH/IGaej0Y9ojGA2aEpVh
# J6n3TezIbXNZDBZW1ODKX1W0OmKPNvTdGqFYAHCr6osCrVLyg4ROozoI9GnsvjC7
# f9ACbPJf6Xy1B2v0teYREkUmpqc+OC/rZpApjgtL2Y5ymgeuihuSUj/XaKNtDa0Z
# ERONWgyLAgMBAAGjggEJMIIBBTAdBgNVHQ4EFgQUBsPfWqqHee6gVxN8Wohmb0CT
# pgMwHwYDVR0jBBgwFoAUIzT42VJGcArtQPt2+7MrsMM1sw8wVAYDVR0fBE0wSzBJ
# oEegRYZDaHR0cDovL2NybC5taWNyb3NvZnQuY29tL3BraS9jcmwvcHJvZHVjdHMv
# TWljcm9zb2Z0VGltZVN0YW1wUENBLmNybDBYBggrBgEFBQcBAQRMMEowSAYIKwYB
# BQUHMAKGPGh0dHA6Ly93d3cubWljcm9zb2Z0LmNvbS9wa2kvY2VydHMvTWljcm9z
# b2Z0VGltZVN0YW1wUENBLmNydDATBgNVHSUEDDAKBggrBgEFBQcDCDANBgkqhkiG
# 9w0BAQUFAAOCAQEAjgD2Z96da+Ze+YXIxGUX2pvvvX2etiR572Kwk6j6aXOFJrbB
# FaNelpipwJCRAY/V9qLIqUh+KfQFBKQYlRBf50WrCcXz+sx0BxyG597HjjGCmL4o
# Y0j/F0KATLMw60EcOh2I1hotO1a1W5fHB661OxD+T5KC6D9JN9TTP8vxap080i/V
# uNKyr2QubnfuOvs7jTjDJP5l5ZUEAFcxuliihARHhKnyoWxWcvje/fI463+pmRhF
# /nBuA3jTiCC5DWI3vST9I0l/BwrVDVMwvvnn5xf0vHb1U3TrJVeo2VRpHsqsoCA0
# 35Vuya6u01jEDkKhrZHuuMnxTAgCVuIFeXh9xDCCBOwwggPUoAMCAQICEzMAAAEK
# LHmu13l7pqwAAQAAAQowDQYJKoZIhvcNAQEFBQAweTELMAkGA1UEBhMCVVMxEzAR
# BgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1JlZG1vbmQxHjAcBgNVBAoTFU1p
# Y3Jvc29mdCBDb3Jwb3JhdGlvbjEjMCEGA1UEAxMaTWljcm9zb2Z0IENvZGUgU2ln
# bmluZyBQQ0EwHhcNMTUwNjA0MTc0MjQ1WhcNMTYwOTA0MTc0MjQ1WjCBgzELMAkG
# A1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1JlZG1vbmQx
# HjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjENMAsGA1UECxMETU9QUjEe
# MBwGA1UEAxMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMIIBIjANBgkqhkiG9w0BAQEF
# AAOCAQ8AMIIBCgKCAQEAkvxvNrvhA7ko1kbRomG6pUb7YxY+LvlH0sfs7ceZsSZR
# db3azA8GFBGkbBmewF9NVInkTVDYjc2hYaV3E5ocp+0NdPenVnnoKPdT0rF6Y+D1
# lJe37NlH+Gw98yWIs7wKxQSnjnyFSHYcYUaqGcR6YovBwjWq+1hvxWMLk0kwgRt7
# 3398T7RHbV94HK+295YTUu+50U055XPeSE48FKqXDTnMi1HhXNE78I5n6jBgqU1a
# nUO92yO6wA/XSxCdnE3wUaKEquScpz3Wo+8KGEio+rFOpZgOS7/wFPMAyLBI1lv+
# ONeJES0FukMCTyIAliWtfTVuhGirIBg4KP4cohCxpQIDAQABo4IBYDCCAVwwEwYD
# VR0lBAwwCgYIKwYBBQUHAwMwHQYDVR0OBBYEFIn+CjHqJs3mbceRkbmUjKIYYTc0
# MFEGA1UdEQRKMEikRjBEMQ0wCwYDVQQLEwRNT1BSMTMwMQYDVQQFEyozMTU5NSsw
# NDA3OTM1MC0xNmZhLTRjNjAtYjZiZi05ZDJiMWNkMDU5ODQwHwYDVR0jBBgwFoAU
# yxHoytK0FlgByTcuMxYWuUyaCh8wVgYDVR0fBE8wTTBLoEmgR4ZFaHR0cDovL2Ny
# bC5taWNyb3NvZnQuY29tL3BraS9jcmwvcHJvZHVjdHMvTWljQ29kU2lnUENBXzA4
# LTMxLTIwMTAuY3JsMFoGCCsGAQUFBwEBBE4wTDBKBggrBgEFBQcwAoY+aHR0cDov
# L3d3dy5taWNyb3NvZnQuY29tL3BraS9jZXJ0cy9NaWNDb2RTaWdQQ0FfMDgtMzEt
# MjAxMC5jcnQwDQYJKoZIhvcNAQEFBQADggEBAKaoU5HfOwH+YV0GWutm8AzY99mE
# 0wBRDhDRpNEXKOeMMzgr2EPGA4t16zOS9KniZ/0C26UdTkNFXRtJ4+BPFvB+j/CI
# EcqCraT7pqlf9ZdgyHvEv3ybae0fgsHxzY54S2L11w0c11MS1pZS7zW9GY6gQJOh
# CqUtFpzCRnQI69v02KVJNlQSEVUDs3sW+0f++2j8wEVc4j8SeTOi74LF3kAZB+4V
# xQqbWQVBqdCXnoGeA1v8SuMaLgWsUEcvjPp52B4g+AX7KWsYFPpyBLcLp5pk7BFd
# T0VJjSkaLfvQtglTXzSU4BbEuepzNehX7+HrFsMYxwazO89hhOK2RImUo4YwggW8
# MIIDpKADAgECAgphMyYaAAAAAAAxMA0GCSqGSIb3DQEBBQUAMF8xEzARBgoJkiaJ
# k/IsZAEZFgNjb20xGTAXBgoJkiaJk/IsZAEZFgltaWNyb3NvZnQxLTArBgNVBAMT
# JE1pY3Jvc29mdCBSb290IENlcnRpZmljYXRlIEF1dGhvcml0eTAeFw0xMDA4MzEy
# MjE5MzJaFw0yMDA4MzEyMjI5MzJaMHkxCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpX
# YXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQg
# Q29ycG9yYXRpb24xIzAhBgNVBAMTGk1pY3Jvc29mdCBDb2RlIFNpZ25pbmcgUENB
# MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAsnJZXBkwZL8dmmAgIEKZ
# dlNsPhvWb8zL8epr/pcWEODfOnSDGrcvoDLs/97CQk4j1XIA2zVXConKriBJ9PBo
# rE1LjaW9eUtxm0cH2v0l3511iM+qc0R/14Hb873yNqTJXEXcr6094CholxqnpXJz
# VvEXlOT9NZRyoNZ2Xx53RYOFOBbQc1sFumdSjaWyaS/aGQv+knQp4nYvVN0UMFn4
# 0o1i/cvJX0YxULknE+RAMM9yKRAoIsc3Tj2gMj2QzaE4BoVcTlaCKCoFMrdL109j
# 59ItYvFFPeesCAD2RqGe0VuMJlPoeqpK8kbPNzw4nrR3XKUXno3LEY9WPMGsCV8D
# 0wIDAQABo4IBXjCCAVowDwYDVR0TAQH/BAUwAwEB/zAdBgNVHQ4EFgQUyxHoytK0
# FlgByTcuMxYWuUyaCh8wCwYDVR0PBAQDAgGGMBIGCSsGAQQBgjcVAQQFAgMBAAEw
# IwYJKwYBBAGCNxUCBBYEFP3RMU7TJoqV4ZhgO6gxb6Y8vNgtMBkGCSsGAQQBgjcU
# AgQMHgoAUwB1AGIAQwBBMB8GA1UdIwQYMBaAFA6sgmBAVieX5SUT/CrhClOVWeSk
# MFAGA1UdHwRJMEcwRaBDoEGGP2h0dHA6Ly9jcmwubWljcm9zb2Z0LmNvbS9wa2kv
# Y3JsL3Byb2R1Y3RzL21pY3Jvc29mdHJvb3RjZXJ0LmNybDBUBggrBgEFBQcBAQRI
# MEYwRAYIKwYBBQUHMAKGOGh0dHA6Ly93d3cubWljcm9zb2Z0LmNvbS9wa2kvY2Vy
# dHMvTWljcm9zb2Z0Um9vdENlcnQuY3J0MA0GCSqGSIb3DQEBBQUAA4ICAQBZOT5/
# Jkav629AsTK1ausOL26oSffrX3XtTDst10OtC/7L6S0xoyPMfFCYgCFdrD0vTLqi
# qFac43C7uLT4ebVJcvc+6kF/yuEMF2nLpZwgLfoLUMRWzS3jStK8cOeoDaIDpVbg
# uIpLV/KVQpzx8+/u44YfNDy4VprwUyOFKqSCHJPilAcd8uJO+IyhyugTpZFOyBvS
# j3KVKnFtmxr4HPBT1mfMIv9cHc2ijL0nsnljVkSiUc356aNYVt2bAkVEL1/02q7U
# gjJu/KSVE+Traeepoiy+yCsQDmWOmdv1ovoSJgllOJTxeh9Ku9HhVujQeJYYXMk1
# Fl/dkx1Jji2+rTREHO4QFRoAXd01WyHOmMcJ7oUOjE9tDhNOPXwpSJxy0fNsysHs
# cKNXkld9lI2gG0gDWvfPo2cKdKU27S0vF8jmcjcS9G+xPGeC+VKyjTMWZR4Oit0Q
# 3mT0b85G1NMX6XnEBLTT+yzfH4qerAr7EydAreT54al/RrsHYEdlYEBOsELsTu2z
# dnnYCjQJbRyAMR/iDlTd5aH75UcQrWSY/1AWLny/BSF64pVBJ2nDk4+VyY3YmyGu
# DVyc8KKuhmiDDGotu3ZrAB2WrfIWe/YWgyS5iM9qqEcxL5rc43E91wB+YkfRzojJ
# uBj6DnKNwaM9rwJAav9pm5biEKgQtDdQCNbDPTCCBgcwggPvoAMCAQICCmEWaDQA
# AAAAABwwDQYJKoZIhvcNAQEFBQAwXzETMBEGCgmSJomT8ixkARkWA2NvbTEZMBcG
# CgmSJomT8ixkARkWCW1pY3Jvc29mdDEtMCsGA1UEAxMkTWljcm9zb2Z0IFJvb3Qg
# Q2VydGlmaWNhdGUgQXV0aG9yaXR5MB4XDTA3MDQwMzEyNTMwOVoXDTIxMDQwMzEz
# MDMwOVowdzELMAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAOBgNV
# BAcTB1JlZG1vbmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEhMB8G
# A1UEAxMYTWljcm9zb2Z0IFRpbWUtU3RhbXAgUENBMIIBIjANBgkqhkiG9w0BAQEF
# AAOCAQ8AMIIBCgKCAQEAn6Fssd/bSJIqfGsuGeG94uPFmVEjUK3O3RhOJA/u0afR
# TK10MCAR6wfVVJUVSZQbQpKumFwwJtoAa+h7veyJBw/3DgSY8InMH8szJIed8vRn
# HCz8e+eIHernTqOhwSNTyo36Rc8J0F6v0LBCBKL5pmyTZ9co3EZTsIbQ5ShGLies
# hk9VUgzkAyz7apCQMG6H81kwnfp+1pez6CGXfvjSE/MIt1NtUrRFkJ9IAEpHZhEn
# KWaol+TTBoFKovmEpxFHFAmCn4TtVXj+AZodUAiFABAwRu233iNGu8QtVJ+vHnhB
# MXfMm987g5OhYQK1HQ2x/PebsgHOIktU//kFw8IgCwIDAQABo4IBqzCCAacwDwYD
# VR0TAQH/BAUwAwEB/zAdBgNVHQ4EFgQUIzT42VJGcArtQPt2+7MrsMM1sw8wCwYD
# VR0PBAQDAgGGMBAGCSsGAQQBgjcVAQQDAgEAMIGYBgNVHSMEgZAwgY2AFA6sgmBA
# VieX5SUT/CrhClOVWeSkoWOkYTBfMRMwEQYKCZImiZPyLGQBGRYDY29tMRkwFwYK
# CZImiZPyLGQBGRYJbWljcm9zb2Z0MS0wKwYDVQQDEyRNaWNyb3NvZnQgUm9vdCBD
# ZXJ0aWZpY2F0ZSBBdXRob3JpdHmCEHmtFqFKoKWtTHNY9AcTLmUwUAYDVR0fBEkw
# RzBFoEOgQYY/aHR0cDovL2NybC5taWNyb3NvZnQuY29tL3BraS9jcmwvcHJvZHVj
# dHMvbWljcm9zb2Z0cm9vdGNlcnQuY3JsMFQGCCsGAQUFBwEBBEgwRjBEBggrBgEF
# BQcwAoY4aHR0cDovL3d3dy5taWNyb3NvZnQuY29tL3BraS9jZXJ0cy9NaWNyb3Nv
# ZnRSb290Q2VydC5jcnQwEwYDVR0lBAwwCgYIKwYBBQUHAwgwDQYJKoZIhvcNAQEF
# BQADggIBABCXisNcA0Q23em0rXfbznlRTQGxLnRxW20ME6vOvnuPuC7UEqKMbWK4
# VwLLTiATUJndekDiV7uvWJoc4R0Bhqy7ePKL0Ow7Ae7ivo8KBciNSOLwUxXdT6uS
# 5OeNatWAweaU8gYvhQPpkSokInD79vzkeJkuDfcH4nC8GE6djmsKcpW4oTmcZy3F
# UQ7qYlw/FpiLID/iBxoy+cwxSnYxPStyC8jqcD3/hQoT38IKYY7w17gX606Lf8U1
# K16jv+u8fQtCe9RTciHuMMq7eGVcWwEXChQO0toUmPU8uWZYsy0v5/mFhsxRVuid
# cJRsrDlM1PZ5v6oYemIp76KbKTQGdxpiyT0ebR+C8AvHLLvPQ7Pl+ex9teOkqHQ1
# uE7FcSMSJnYLPFKMcVpGQxS8s7OwTWfIn0L/gHkhgJ4VMGboQhJeGsieIiHQQ+kr
# 6bv0SMws1NgygEwmKkgkX1rqVu+m3pmdyjpvvYEndAYR7nYhv5uCwSdUtrFqPYmh
# dmG0bqETpr+qR/ASb/2KMmyy/t9RyIwjyWa9nR2HEmQCPS2vWY+45CHltbDKY7R4
# VAXUQS5QrJSwpXirs6CWdRrZkocTdSIvMqgIbqBbjCW/oO+EyiHW6x5PyZruSeD3
# AWVviQt9yGnI5m7qp5fOMSn/DsVbXNhNG6HY+i+ePy5VFmvJE6P9MYIErzCCBKsC
# AQEwgZAweTELMAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAOBgNV
# BAcTB1JlZG1vbmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEjMCEG
# A1UEAxMaTWljcm9zb2Z0IENvZGUgU2lnbmluZyBQQ0ECEzMAAAEKLHmu13l7pqwA
# AQAAAQowCQYFKw4DAhoFAKCByDAZBgkqhkiG9w0BCQMxDAYKKwYBBAGCNwIBBDAc
# BgorBgEEAYI3AgELMQ4wDAYKKwYBBAGCNwIBFTAjBgkqhkiG9w0BCQQxFgQUpx6b
# sRMUc/UnkgE5bc1RAoUYHXswaAYKKwYBBAGCNwIBDDFaMFigNoA0AE0AaQBjAHIA
# bwBzAG8AZgB0ACAASABQAEMAIABQAGEAYwBrACAAMgAwADEAMgAgAFIAMqEegBxo
# dHRwOi8vd3d3Lm1pY3Jvc29mdC5jb20vSFBDMA0GCSqGSIb3DQEBAQUABIIBABwt
# KqVEiOthimUezsV2UFxQw6j9vZuODUdRv1FU+nNn/zqYzMEA8JjzQ24MIUuVd+Tk
# cZoRRXwMtmPtqCMNfwWl65JS2z0coxLQHE+eDwVGlJdSFLBRIbZGyoKGyedA0FcO
# zpJ0dn0Y0Yq/DSPK0MSSQFnRSUGFI2ICgaY45HrJdVW5wNhbyScORmtuT/EiJkwy
# ohaVOx6F75i6/MIHR9Md5jHpf75II0dMBqlikjxZD9stmSGkR1XrTHTHItXqUezN
# 37/u45zNLr8/pMkZ69nLDXEIeJkfLlq3OKaIJ2xLuTnGork1qZDfgVcfV1SLEoHx
# 1yvPCMGm/QJYipuvWuihggIoMIICJAYJKoZIhvcNAQkGMYICFTCCAhECAQEwgY4w
# dzELMAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1Jl
# ZG1vbmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEhMB8GA1UEAxMY
# TWljcm9zb2Z0IFRpbWUtU3RhbXAgUENBAhMzAAAAkkyir1UnCC51AAAAAACSMAkG
# BSsOAwIaBQCgXTAYBgkqhkiG9w0BCQMxCwYJKoZIhvcNAQcBMBwGCSqGSIb3DQEJ
# BTEPFw0xNTExMTMwMjQ1MTBaMCMGCSqGSIb3DQEJBDEWBBTF8FlrTrjLRGi5Nq9l
# FODlpucvFjANBgkqhkiG9w0BAQUFAASCAQB0IgR6Ja6ApgKKVd8n/A8PfdkuD3i4
# 3aSuJTEEBQcSp0eHwFXyxZJUtXC2Zt56mYqVfVG4Ox6bKycdzZZrd1YSa2YW8E8H
# NcOSvsJ//1buLTsfsIMheagvG2fx6KdXimxSlrj6ZN/hsQa1xDsNIn7VEjvy/oxA
# IkZdeNVmg73C9XYuFp5MPVn3eEPHZp8dOPit5fDQI+x2xuyjyZzNhtM0HOnroqPQ
# zjePLmCE+/e+ElXyvW+tHqlEOBVtIv4zZwoy1NbQMnbhxu5K2YDcOd9OChLeHgKj
# qpaY1ox1z8jhs0FBYj2PhuGmYTK1cqMLTbKizGoiTzi//1xahH0GPLjp
# SIG # End signature block
