﻿<#
.Synopsis
   Write the trace log
.DESCRIPTION
   Write the trace log. 
.EXAMPLE
   TraceLog -TraceLevel "This is an error message"

.INPUTS
   None
.OUTPUTS
   None
#>
function TraceLog
{
    Param
    (
        # 
        [Parameter(Mandatory=$true)]
        [ValidateSet("Error","Warning","Information","Detail")]
        [String] $TraceLevel,
		
        # Prefix name to name resources
        [Parameter(Mandatory=$true)]
        [String] $Message		
    )

    $date = Get-Date -Format 'MM/dd/yyyy HH:mm:ss'
    $newmsg =  "[$TraceLevel]$date - $Message"
    Write-Verbose $newmsg -Verbose
}

<#
.Synopsis
   Write the detail level trace log
.DESCRIPTION
   Write the detail level trace log. 
#>
function TraceDetail 
{
    Param
    (
	    [Parameter(Mandatory=$true, ValueFromPipeline = $true)]
	    [string] $Message
    )

    TraceLog -TraceLevel "Detail" -Message $Message
}

<#
.Synopsis
   Write the information level trace log
.DESCRIPTION
   Write the information level trace log. 
#>
function TraceInfo 
{
    Param
    (
	    [Parameter(Mandatory=$true, ValueFromPipeline = $true)]
	    [string] $Message
    )

    TraceLog -TraceLevel "Information" -Message $Message
}

<#
.Synopsis
   Write the warning level trace log
.DESCRIPTION
   Write the warning level trace log. 
#>
function TraceWarning 
{
    Param
    (
	    [Parameter(Mandatory=$true, ValueFromPipeline = $true)]
	    [string] $Message
    )

    TraceLog -TraceLevel "Warning" -Message $Message
}

<#
.Synopsis
   Write the error level trace log
.DESCRIPTION
   Write the error level trace log. 
#>
function TraceError 
{
    Param
    (
	    [Parameter(Mandatory=$true)]
	    [string] $Message
    )

    TraceLog -TraceLevel "Error" -Message $Message
}

<#
.SYNOPSIS
  Returns the latest image for a given image family name filter.
.DESCRIPTION
  Will return the latest image based on a filter match on the ImageFamilyName and
  PublisedDate of the image.  The more specific the filter, the more control you have
  over the object returned.
.EXAMPLE
  The following example will return the latest SQL Server image.  It could be SQL Server
  2014, 2012 or 2008
    
    Get-LatestVMImage -ImageFamilyNameFilter "*SQL Server*"

  The following example will return the latest SQL Server 2014 image. This function will
  also only select the image from images published by Microsoft.  
   
    Get-LatestVMImage -ImageFamilyNameFilter "*SQL Server 2014*" -OnlyMicrosoftImages

  The following example will return $null because Microsoft doesn't publish Ubuntu images.
   
    Get-LatestVMImage -ImageFamilyNameFilter "*Ubuntu*" -OnlyMicrosoftImages

  The following example will return latest SQL Server 2012 SPx Standard image.
   
    Get-LatestVMImage -ImageFamilyNameFilter "SQL Server 2012 SP[1-9] Standard on Windows Server 2012" -OnlyMicrosoftImages -RegexMatch

  The following example will return latest CentOS 7.0 image.
   
    Get-LatestVMImage -ImageNameFilter "5112500ae3b842c8b9c604889f8753c3__OpenLogic-CentOS-70*"

#>
function Get-LatestVMImage
{
    param
    (
        # filters for selecting the image family listed by preference order.
        # For example, "Windows Server 2012*", "*2012 Datacenter*", "*SQL*, "Sharepoint*"
        [Parameter(Mandatory = $true, ParameterSetName="ByImageFamily")]
        [String[]] $ImageFamilyNameFilter,

        # the filter for selecting the image name.
        # For example, "a699494373c04fc0bc8f2bb1389d6106__Windows-Server-2012-Datacenter-*"
        [Parameter(Mandatory = $true, ParameterSetName="ByImageName")]
        [String] $ImageNameFilter,

        # A switch to indicate whether or not to select the latest image where the publisher is Microsoft.
        # If this switch is not specified, then images from all possible publishers are considered.
        [Parameter(Mandatory = $false)]
        [switch] $OnlyMicrosoftImages,

        # The image list to be filtered, if not specified, use Get-AzureVMImage to get the list.
        [Parameter(Mandatory = $false)]
        [Object[]] $FromImageList = @(),

        [Parameter(Mandatory = $false)]
        [Switch] $RegexMatch
    )

    if(($null -eq $FromImageList) -or ($FromImageList.Count -eq 0))
    {
        # Get a list of all available images.
        $imageList = InvokeAzureCmdWithRetry -Command "Get-AzureVMImage"
    }
    else
    {
        $imageList = $FromImageList
    }
    
    if($OnlyMicrosoftImages.IsPresent)
    {
        $imageList = @($imageList | ?{$_.Category -eq "Public" -and ($_.PublisherName -ilike "Microsoft*")})
    }
    else
    {
        $imageList = @($imageList | ?{$_.Category -eq "Public"})
    }

    if($PsCmdlet.ParameterSetName -eq "ByImageFamily")
    {
        foreach($filter in $ImageFamilyNameFilter)
        {
            if($RegexMatch.IsPresent)
            {
                $foundImageList = @($imageList | ?{$_.ImageFamily -imatch $filter})
            }
            else
            {
                $foundImageList = @($imageList | ?{$_.ImageFamily -ilike $filter})
            }

            if($foundImageList.Count -gt 0)
            {
                return ($foundImageList | Sort-Object -Descending -Property ImageFamily, PublishedDate | Select-Object -First(1))
            }
        }
    }
    else
    {
        if($RegexMatch.IsPresent)
        {
            $foundImageList = @($imageList | ?{$_.ImageName -imatch $ImageNameFilter})
        }
        else
        {
            $foundImageList = @($imageList | ?{$_.ImageName -ilike $ImageNameFilter})
        }

        if($foundImageList.Count -gt 0)
        {
            return ($foundImageList | Sort-Object -Descending -Property PublishedDate | Select-Object -First(1))
        }       
    }

    return $null
}

<#
.SYNOPSIS
  Get the latest Windows Server Data Center image
.DESCRIPTION
  Will return the latest Windows Server Data Center image.
.EXAMPLE
    Get-LatestWindowsServer2012R2Image
#>
function Get-LatestWindowsServerImageName
{
    param
    (
        [Parameter(Mandatory = $true)]
        [ValidateSet("WindowsServer2008R2","WindowsServer2012","WindowsServer2012R2")]
        [String] $Version,

        [Parameter(Mandatory = $false)]
        [String] $Language = "",

        [Parameter(Mandatory = $false)]
        [Object[]] $FromImageList = @()
    )

    $baseFilter = ""
    switch($Version)
    {
        "WindowsServer2008R2"   {$baseFilter = "Windows Server 2008 R2 SP[1-9]"; break}
        "WindowsServer2012"     {$baseFilter = "Windows Server 2012( SP[1-9])? Datacenter"; break}
        "WindowsServer2012R2"   {$baseFilter = "Windows Server 2012 R2( SP[1-9])? Datacenter"; break}
        default                 {throw  "Get-LatestWindowsServerDatacenterImage: unsupported Windows Server Version $Version"}
    }

    $imgFilters = @()
    if([String]::IsNullOrEmpty($Language))
    {
        $imgFilters += $baseFilter
        $imgFilters += "$baseFilter \(en-us\)"
    }
    else
    {
        $imgFilters += "$baseFilter \($Language\)"
        $imgFilters += $baseFilter
    }

    $imgFilters = @($imgFilters | %{"^${_}$"})
    $image = Get-LatestVMImage -ImageFamilyNameFilter $imgFilters -OnlyMicrosoftImages -FromImageList $FromImageList -RegexMatch
    if ($null -eq $image)
    {
        throw "Unable to find an image for $Version."
    }

    return $image.ImageName
}

<#
.SYNOPSIS
  Get the latest SQL Server Image
.DESCRIPTION
  Will return the latest SQL Server image for given type.
.EXAMPLE
  The following example will return the latest SQL Server 2012 SP1 Standard image
    Get-LatestSQLServerImageName -Version "SQLServer2014_Standard"
#>
function Get-LatestSQLServerImageName
{
    param
    (
        [Parameter(Mandatory = $true)]
        [ValidateSet("SQLServer2012_Standard","SQLServer2012_Enterprise","SQLServer2014_Standard","SQLServer2014_Enterprise")]
        [String] $Version,

        [Parameter(Mandatory = $false)]
        [String] $Language = "",

        [Parameter(Mandatory = $false)]
        [Object[]] $FromImageList = @()
    )
	
    $baseFilter = ""
    switch($Version)
    {
        "SQLServer2012_Standard"   {$baseFilter = "SQL Server 2012 SP[1-9] Standard on Windows Server 2012"; break}
        "SQLServer2012_Enterprise" {$baseFilter = "SQL Server 2012 SP[1-9] Enterprise on Windows Server 2012"; break}
        "SQLServer2014_Standard"      {$baseFilter = "SQL Server 2014 (SP[1-9]|RTM) Standard on Windows Server 2012 R2"; break}
        "SQLServer2014_Enterprise"    {$baseFilter = "SQL Server 2014 (SP[1-9]|RTM) Enterprise on Windows Server 2012 R2"; break}
        default                       {throw  "Get-LatestSQLServerImageName: unknown SQL Server Version $Version"}
    }

    $imgFilters = @()
    if([String]::IsNullOrEmpty($Language))
    {
        $imgFilters += $baseFilter
        $imgFilters += "$baseFilter \(en-us\)"
    }
    else
    {
        $imgFilters += "$baseFilter \($Language\)"
        $imgFilters += $baseFilter
    }

    $imgFilters = @($imgFilters | %{"^${_}$"})
    $image = Get-LatestVMImage -ImageFamilyNameFilter $imgFilters -OnlyMicrosoftImages -FromImageList $FromImageList -RegexMatch
    if ($null -eq $image)
    {
        throw "Unable to find a SQL Server image."
    }

    return $image.ImageName
}

<#
.SYNOPSIS
  Get the latest public HPC Image
.DESCRIPTION
  Will return the latest HPC Pack image name
#>
function Get-LatestHPCImageName
{
    [CmdletBinding(DefaultParametersetName="ByNodeType")]
    param
    (
        [Parameter(Mandatory = $true, ParameterSetName="ByWellKnownName")]
        [String] $WellKnownName,

        [Parameter(Mandatory = $true, ParameterSetName="ByNodeType")]
        [ValidateSet("HeadNode", "ComputeNode", "BrokerNode")]
        [String] $NodeType,

        [Parameter(Mandatory = $false)]
        [String] $Language = "",

        [Parameter(Mandatory = $false)]
        [Object[]] $FromImageList = @()
    )

    if($PsCmdlet.ParameterSetName -eq "ByWellKnownName")
    {
        $baseFilter = GetWellKnownHPCImageFamily -ImageName $WellKnownName
        if([String]::IsNullOrEmpty($baseFilter))
        {
            TraceWarning "Unknown HPC Image name $WellKnownName"
            return $null
        }
    }
    else
    {
        if($NodeType -eq "HeadNode")
        {
            $baseFilter = "HPC Pack 2012 R2 on Windows Server 2012 R2"
        }
        elseif($NodeType -eq "ComputeNode")
        {
            $baseFilter = "HPC Pack 2012 R2 Compute Node on Windows Server 2012 R2"
        }
        else
        {
            TraceWarning "No HPC Image available for $NodeType"
            return $null
        }
    }

    $imgFilters = @()
    if([String]::IsNullOrEmpty($Language))
    {
        $imgFilters += $baseFilter
        $imgFilters += "$baseFilter (en-us)"
    }
    else
    {
        $imgFilters += "$baseFilter ($Language)"
        $imgFilters += $baseFilter
    }

    $image = Get-LatestVMImage -ImageFamilyNameFilter $imgFilters -OnlyMicrosoftImages -FromImageList $FromImageList
    if ($null -ne $image)
    {
        return $image.ImageName
    }

    return $null
}

function GetVMImageMediaLink
{
    Param
    (
        [Parameter(Mandatory=$true)]
        $Image
    )

    if($Image.GetType().Name -eq "OSImageContext")
    {
        $mediaLink = $Image.MediaLink.ToString()
    }
    else
    {
        $mediaLink = $Image.OSDiskConfiguration.MediaLink.ToString()
    }

    return $mediaLink
}

function GetVMImageTimeStamp
{
    Param
    (
        [Parameter(Mandatory=$true)]
        $Image
    )

    if($Image.GetType().Name -eq "OSImageContext")
    {
        $timeStamp = $Image.PublishedDate
    }
    else
    {
        $timeStamp = $Image.ModifiedTime
    }

    return $timeStamp
}

function Get-BaseHPCNodeImage
{
    param
    (
        [Parameter(Mandatory = $true)]
        [ValidateSet("ComputeNode","BrokerNode")]
        [String] $NodeType,

        [Parameter(Mandatory = $false)]
        [String] $Location = "",

        [Parameter(Mandatory = $false)]
        [Object[]] $FromImageList = @()
    )

    if(($null -eq $FromImageList) -or ($FromImageList.Count -eq 0))
    {
        # Get a list of all available images.
        $imageList = @(InvokeAzureCmdWithRetry -Command "Get-AzureVMImage")
    }
    else
    {
        $imageList = $FromImageList
    }

    $imageList = @($imageList | ?{($_.Category -eq "user") -and (IsBaseHPCNodeImage -ImageName $_.ImageName -NodeType $NodeType)})
    if(-not [String]::IsNullOrEmpty($Location))
    {
        if(($imageList | ?{$null -eq $_.Location} | measure).Count -gt 0)
        {
            #If some images are in the affinity groups, we need to check whether the affinity groups are in the location
            # Get the images directly in $Location
            $tmpImageList = @($imageList | ?{$null -ne $_.Location} | ?{$_.Location.Split(';') -contains $Location})
            $ags = @(InvokeAzureCmdWithRetry -Command "Get-AzureAffinityGroup")
            foreach($img in @($srcImages | ?{$null -eq $_.Location}))
            {
                $ag = $ags | ?{$_.Name -eq $img.AffinityGroup}
                if(($null -ne $ag) -and ($ag.Location -eq $Location))
                {
                    $tmpImageList += $img
                }
            }

            $imageList = $tmpImageList
        }
        else
        {
            $imageList = @($imageList | ?{($_.Location.Split(';') -contains $Location)})
        }
    }

    return $imageList
}

function IsBaseHPCNodeImage
{
    param
    (
        [Parameter(Mandatory = $true, ParameterSetName="ImageName")]
        [String] $ImageName,

        [Parameter(Mandatory = $true, ParameterSetName="VMSrcImageName")]
        [String] $VMSrcImageName,

        [Parameter(Mandatory = $true)]
        [ValidateSet("ComputeNode","BrokerNode", "Any")]
        [String] $NodeType
    )

    # The format of base HPC node image name is HPC-Pack-<NodeType>_yyyyMMdd-HHmmss.fffffff[_<StorageAccountName>]
    if($NodeType -eq "Any")
    {
        $imageNameHead = "HPC-Pack-(ComputeNode|BrokerNode)_"
    }
    else
    {
        $imageNameHead = "HPC-Pack-{0}_" -f $NodeType
    }

    $pattern = "^$imageNameHead[0-9]{8}-[0-9]{6}.[0-9]{7}(_[0-9a-z]{3,24})?$"
    if($PsCmdlet.ParameterSetName -eq "ImageName")
    {
        $matches = $ImageName -match $pattern
    }
    else
    {
        # There are two types of VM images: VMImage and OSImage
        # If a VM was created from OSImage, the value of VM.OSVirtualHardDisk.SourceImageName was fetched from OSImageContext.ImageName
        # If a VM was created from VMImage, the value of VM.OSVirtualHardDisk.SourceImageName was fetched from VMImageContext.OSDiskConfiguration.Name, the format is <ImageName>-os-yyyy-MM-dd
        $matches = $VMSrcImageName -match $pattern
        $pattern = "^$imageNameHead[0-9]{8}-[0-9]{6}.[0-9]{7}-os-"
        $matches = $matches -or ($VMSrcImageName -match $pattern)
    }

    return $matches
}

function NewBaseHPCNodeImageName
{
    param
    (
        [Parameter(Mandatory = $true)]
        [String] $NodeType
    )

    $imageNameHead = "HPC-Pack-{0}_" -f $NodeType
    $imageName = $imageNameHead + (Get-Date).ToUniversalTime().ToString("yyyyMMdd-HHmmss.fffffff")
    return $imageName
}

function ParseBaseHPCNodeImageName
{
    param
    (
        [Parameter(Mandatory = $true)]
        [String] $ImageName
    )

    # The format of base HPC node image name is HPC-Pack-<NodeType>_yyyyMMdd-HHmmss.fffffff[_<StorageAccountName>]
    if(-not (IsBaseHPCNodeImage -ImageName $ImageName -NodeType Any))
    {
        throw "The image name $ImageName is not a valid base HPC node image name."
    }

    $index = $ImageName.IndexOf('_')
    $nodeType = $ImageName.Substring(9, $index-9)

    $leftPart = $ImageName.Substring($index+1)
    $index = $leftPart.IndexOf('_')
    if($index -lt 0)
    {
        $dateTimePart = $leftPart
        $storagePart = ""
    }
    else
    {
        $dateTimePart = $leftPart.Substring(0, $index)
        $storagePart = $leftPart.Substring($index+1)
    }

    $dateTime = [datetime]::ParseExact($dateTimePart, "yyyyMMdd-HHmmss.fffffff", [CultureInfo]::InvariantCulture, "AssumeUniversal")

    return @{"NodeType"=$nodeType;"DateTime"=$dateTime;"StorageAccount"=$storagePart}
}

function ParseNamePattern
{
    Param
    (
        [Parameter(Mandatory=$true)]
        [String] $NamePattern
    )

    $firstper = $NamePattern.IndexOf('%')
    $nameSpaceStrLen = $NamePattern.Length - $firstper - 2
    $rootName = $NamePattern.Substring(0, $firstper)
    $nameSpaceStr = $NamePattern.Substring($firstper+1, $nameSpaceStrLen)
    $startNameIndex = [int]$nameSpaceStr

    return @($rootName, $nameSpaceStrLen, $startNameIndex)
}

function GetNamesByPattern
{
    Param
    (
        [Parameter(Mandatory=$true)]
        [String] $NamePattern,

        [Parameter(Mandatory=$true)]
        [int] $Count
    )

    $names = @()
    if($Count -gt 0)
    {
        $parseResult = ParseNamePattern -NamePattern $NamePattern
        $rootName = $parseResult[0]
        $indexWidth = $parseResult[1]
        $startIndex = $parseResult[2]
        for ($i=0; $i -lt $Count; $i++)
        {
            $name = $rootName + ($i + $startIndex).ToString("D" + $indexWidth)
            $names += $name
        }
    }
    
    return $names  
}

function GetNameSetsByPattern
{
    Param
    (
        [Parameter(Mandatory=$true)]
        [String] $NamePattern,

        [Parameter(Mandatory=$true)]
        [int] $Count,

        [Parameter(Mandatory=$true)]
        [int] $MaxCountPerSet
    )

    $parseResult = ParseNamePattern -NamePattern $NamePattern
    $rootName = $parseResult[0]
    $indexWidth = $parseResult[1]
    $startIndex = $parseResult[2]
    $setNum = [Math]::Ceiling($Count/$MaxCountPerSet)

    $nameSets = @()
    for ($i=0; $i -lt $setNum; $i++)
    {
        $nameSet = @()
        for($j=0; $j -lt $MaxCountPerSet; $j++)
        {
            $seqNum = $i * $MaxCountPerSet + $j
            if($seqNum -ge $Count)
            {
                break
            }
            $nameSet += $rootName + ($seqNum + $startIndex).ToString("D" + $indexWidth)
        }

        $nameSets += ,$nameSet
    }
    
    return $nameSets
}

function NamePatternOverlapped
{
    Param
    (
        [Parameter(Mandatory=$true)]
        [String] $NamePattern1,

        [Parameter(Mandatory=$true)]
        [String] $NamePattern2
    )

    if($NamePattern1.Length -ne $NamePattern2.Length)
    {
        return $false
    }

    $parseResult1 = ParseNamePattern -NamePattern $NamePattern1.ToLower()
    $rootName1 = $parseResult1[0]
    $indexWidth1 = $parseResult1[1]
    $startIndex1 = $parseResult1[2]
    $parseResult2 = ParseNamePattern -NamePattern $NamePattern2.ToLower()
    $rootName2 = $parseResult2[0]
    $indexWidth2 = $parseResult2[1]
    $startIndex2 = $parseResult2[2]

    if($indexWidth1 -eq $indexWidth2)
    {
        return ($rootName1 -eq $rootName2)
    }
    
    if(($rootName1 -notmatch "^$rootName2[0-9]{1,}$") -and ($rootName2 -notmatch "^$rootName1[0-9]{1,}$"))
    {
        return $false
    }
    
    # To check whether <rootname>%abcd% and <rootname>wx%yz% are overlapped, we will compare wx against ab, 
    # if wx greater than or equal to ab, then they can be overlapped.
    # for example,
    #    test%1210% and test11%20% are not overlapped
    #    test%1100% and test11%20% are overlapped
    #    test%0900% and test11%20% are overlapped
    if($indexWidth1 -lt $indexWidth2)
    {
        $wxPart = [int]$rootName1.Substring($rootName2.Length)
        $abPart = [Math]::Floor($startIndex2/[Math]::Pow(10, $indexWidth1))
    }
    else
    {
        $wxPart = [int]$rootName2.Substring($rootName1.Length)
        $abPart = [Math]::Floor($startIndex1/[Math]::Pow(10, $indexWidth2))
    }

    if($wxPart -ge $abPart)
    {
        return $true
    }
    
    return $false 
}

function NameMatchesPattern
{
    Param
    (
        [Parameter(Mandatory=$true)]
        [String] $Name,

        [Parameter(Mandatory=$true)]
        [String] $NamePattern
    )

    $parseResult = ParseNamePattern -NamePattern $NamePattern
    $rootName = $parseResult[0]
    $indexWidth = $parseResult[1]
    $startIndex = $parseResult[2]

    if($Name -match "^$rootName[0-9]{$indexWidth}$")
    {
        $nameIndex = [int]$Name.Substring($rootName.Length)
        if($nameIndex -ge $startIndex)
        {
            return $true
        }
    }

    return $false
}

<#
.SYNOPSIS
  Test whether PowerShell is run as administrator
.DESCRIPTION
  return True if PowerShell is run as adinistrator
#>
function Test-Administrator  
{  
    $curUser = [Security.Principal.WindowsIdentity]::GetCurrent();
    (New-Object Security.Principal.WindowsPrincipal $curUser).IsInRole([Security.Principal.WindowsBuiltinRole]::Administrator)  
}

<#
.SYNOPSIS
   Installs a WinRm certificate to the local store.
.DESCRIPTION
   Gets the WinRM certificate from the Virtual Machine deployed to the specified cloud service, and 
   installs it on the LocalMachine's Root store. The WinRm certificate is stored on the cloud service 
   that hosts the VM, and can be retrieved with Get-AzureCertificate cmdlet. The certificate is used for
   authenticating the service. 
.EXAMPLE
    Install-WinRmCertificate -ServiceName testservice -vmName testVm
.INPUTS
   None
.OUTPUTS
   None
#>
function Install-WinRmCertificate
{
    param
    (
        [Parameter(Mandatory = $true)]
        [String] $ServiceName,
        
        [Parameter(Mandatory = $true)]
        [String] $VMName
    )

    $vm = InvokeAzureCmdWithRetry -Command "Get-AzureVM -ServiceName $ServiceName -Name $VMName"
    # Find the thumbprint used for the WinRM access
    $winRmCertificateThumbprint = $vm.VM.DefaultWinRMCertificateThumbprint
    $installedCert = Get-Item Cert:\CurrentUser\Root\$winRmCertificateThumbprint -ErrorAction SilentlyContinue
    
    if ($null -eq $installedCert)
    {
        # Retrieve the certificate
        $winRmCertificate = InvokeAzureCmdWithRetry -Command "Get-AzureCertificate -ServiceName $ServiceName -Thumbprint $winRmCertificateThumbprint -ThumbprintAlgorithm sha1"

        $certMutex = New-Object System.Threading.Mutex $false,'HPCIaaSClusterCertStoreMutex'
        $mutexAcquired = $false
        try
        {
            try
            {
                $mutexAcquired = $certMutex.WaitOne()
            }
            catch [System.Threading.AbandonedMutexException]
            {
                $mutexAcquired = $true
            }
            $installedCert = Get-Item Cert:\CurrentUser\Root\$winRmCertificateThumbprint -ErrorAction SilentlyContinue
            if($null -ne $installedCert)
            {
                return
            }
            # Read in the certificate to a memory buffer to import it to a X509 certificate object.
            $certBytes = [System.Convert]::FromBase64String($winRmCertificate.Data)
            $x509Cert = New-Object System.Security.Cryptography.X509Certificates.X509Certificate
            $x509Cert.Import($certBytes)
        
            TraceInfo ("Adding Certificate to Cert:\LocalMachine\Root : Thumbprint={0}, Subject={1}" -f $winRmCertificateThumbprint, $x509Cert.Subject)
            # Add the X509 certificate to the LocalMachine\Root store
            $store = New-Object System.Security.Cryptography.X509Certificates.X509Store "Root", "LocalMachine"
            $store.Open("ReadWrite")
            $store.Add($x509Cert)
            $store.Close()
            if(-not [String]::IsNullOrEmpty($env:HPCTempWinRMCertListFile))
            {
                "Cert:\LocalMachine\Root\$winRmCertificateThumbprint" | Out-File $env:HPCTempWinRMCertListFile -Append
            }            

            TraceInfo ("Added Certificate to Cert:\LocalMachine\Root : Thumbprint={0}, Subject={1}" -f $winRmCertificateThumbprint, $x509Cert.Subject)                      
        }
        finally
        {
            if($mutexAcquired)
            {
                $certMutex.ReleaseMutex()
            }
        }
    }

    return $winRmCertificateThumbprint
}

function CombineScriptBlock
{
    param
    (
        [Parameter(Mandatory=$true, Position=0)]
        [ScriptBlock] $ScriptBlock,

        [Parameter(Mandatory=$true, Position=1)]
        [ScriptBlock] $InitScript
    )

    $scriptStr = $ScriptBlock.ToString()
    $insertedScriptStr = $InitScript.ToString()
    if($scriptStr.TrimStart() -match "^param")
    {
        $index = $scriptStr.IndexOf('(')
        if($index -lt 0)
        {
            throw "Invalid syntax: $scriptStr"
        }

        $openParenthesis = 1
        $index++
        while($index -lt $scriptStr.Length)
        {
            if($scriptStr[$index] -eq '(')
            {
                $openParenthesis++
            }
            elseif($scriptStr[$index] -eq ')')
            {
                $openParenthesis--
            }
            if($openParenthesis -eq 0)
            {
                break
            }
            $index++            
        }

        if($openParenthesis -ne 0)
        {
            throw "Invalid syntax: $scriptStr"
        }
        
        $parampart = $scriptStr.Substring(0, $index + 1)
        $cmdpart = $scriptStr.Substring($index + 1)
        $newScriptBlock = [ScriptBlock]::Create("$parampart`n" + $insertedScriptStr + $cmdpart)
    }
    else
    {
        $newScriptBlock = [ScriptBlock]::Create($insertedScriptStr + "`n" + $scriptStr)
    }

    return $newScriptBlock
}

function Start-HpcIaaSJob
{
    param
    (
        [Parameter(Mandatory=$true)]
        [String] $Name,

        [Parameter(Mandatory=$true)]
        [ScriptBlock] $ScriptBlock,

        [Parameter(Mandatory=$true)]
        [ScriptBlock] $InitializationScript,

        [Parameter(Mandatory=$false)]
        [Object[]] $ArgumentList = @()
    )

    $newScriptBlock = CombineScriptBlock $ScriptBlock -InitScript $InitializationScript 
    $job = Start-Job -Name $Name -ScriptBlock $newScriptBlock -ArgumentList $ArgumentList
    return $job
}

function RemotePshCall
{
    param
    (
        [Parameter(Mandatory=$true)]
        [String] $VMName,

        [Parameter(Mandatory=$true)]
        [String] $ServiceName,

        [Parameter(Mandatory=$true)]
        [PSCredential] $Credential,

        [Parameter(Mandatory=$true)]
        [ScriptBlock] $ScriptBlock,

        [Parameter(Mandatory=$false)]
        [ScriptBlock] $InitializationScript,

        [Parameter(Mandatory=$false)]
        [Object[]] $ArgumentList = @(),

        [Parameter(Mandatory=$false)]
        [int[]] $IgnoreErrorCodes = @(),

        [Parameter(Mandatory=$false)]
        [int[]] $NoRetryErrorCodes = @(),

        [Parameter(Mandatory=$false)]
        [Int] $RetryTimes = 30,

        [Parameter(Mandatory=$false)]
        [Int] $RetryInterval = 30
    )

    $WriteLogFuncStr = @'
    function WriteLog
    {
        param 
        (
            [Parameter(ValueFromPipeline = $true)]
            [String] $Message
        )

        try
        {
            $tz = [System.TimeZoneInfo]::FindSystemTimeZoneById("{LogTimeZone}")
            $utcTime = (Get-Date).ToUniversalTime()
            $logTime = [System.TimeZoneInfo]::ConvertTimeFromUtc($utcTime, $tz)
        }
        catch
        {
            $logTime = Get-Date
        }

        $logTimeStr = $logTime.ToString("MM/dd/yyyy HH:mm:ss")
        Write-Verbose "[Information]$logTimeStr - $Message" -Verbose
    }
'@

    $timeZoneName = (Get-WmiObject win32_timezone).StandardName
    $WriteLogFuncStr = $WriteLogFuncStr.Replace("{LogTimeZone}", $timeZoneName)

    if($null -ne $InitializationScript)
    {
        $insertedScriptStr = $WriteLogFuncStr + "`n" + $InitializationScript.ToString()
    }
    else
    {
        $insertedScriptStr = $WriteLogFuncStr
    }

    $initBlock = [ScriptBlock]::Create($insertedScriptStr)
    $newScriptBlock = CombineScriptBlock $ScriptBlock $initBlock

    $uri = InvokeAzureCmdWithRetry -Command "Get-AzureWinRMUri -ServiceName $ServiceName -Name $VMName" -RetryTimes 5
    if($null -eq $uri)
    {
        throw "Failed to retrieve the WinRM https listener URI of the virtual machine $VMName in cloud service $ServiceName."
    }

    $skipCACheck = [String]::IsNullOrEmpty($env:HPCTempWinRMCertListFile)
    $certInstalled = $false
    $retry = 0
    while($true)
    {
        try
        {
            if($skipCACheck)
            {
                $sessOption = New-PSSessionOption -SkipCACheck
                $ret = Invoke-Command -ConnectionUri $uri -Credential $Credential -ScriptBlock $newScriptBlock -ArgumentList $ArgumentList -SessionOption $sessOption -ErrorAction Stop
            }
            else
            {
                $ret = Invoke-Command -ConnectionUri $uri -Credential $Credential -ScriptBlock $newScriptBlock -ArgumentList $ArgumentList -ErrorAction Stop
            }

            return $ret
        }
        catch [System.Management.Automation.Remoting.PSRemotingTransportException]
        {
            $psRemotingErrCode = $_.Exception.ErrorCode
            $errMsg = "Remote PowerShell Call to $VMName failed (ErrorCode=$psRemotingErrCode): " + $_
            if($IgnoreErrorCodes -contains $psRemotingErrCode)
            {
                TraceInfo $errMsg
                TraceInfo "Treat this error as success because the error code is in IgnoreErrorCodes list."
                return
            }
            if($NoRetryErrorCodes -contains $psRemotingErrCode)
            {
                TraceInfo $errMsg
                TraceInfo "No retry for this error because the error code is in NoRetryErrorCodes list."
                throw
            }
            if(($psRemotingErrCode -eq 0x2f8f) -and (-not $certInstalled) -and (-not $skipCACheck))
            {
                # The certificate is not installed, install the certificate and retry again immediately
                Install-WinRmCertificate -ServiceName $ServiceName -VMName $VMName
                $certInstalled = $true
                continue                    
            }
            else
            {
                if($retry -lt $RetryTimes)
                {
                    TraceWarning $errMsg
                    TraceInfo "Retry to connect $VMName after $RetryInterval seconds"
                    Start-Sleep -Seconds $RetryInterval
                    $retry++
                }
                else
                {
                    throw
                }
            }
        }
    }
}

function GetExceptionTypeInfo
{
    Param
    (
        [Parameter(Mandatory=$true)]
        $Exception
    )

    if($null -eq $Exception)
    {
        return ""
    }

    $infoStr = "Exception Type: {0}" -f $Exception.GetType()
    $errCodeColl = $Exception.PSObject.Properties.Match("ErrorCode")
    if($null -ne $errCodeColl -and ($errCodeColl.Count -gt 0))
    {
        if($null -ne $Exception.ErrorCode)
        {
            $infoStr += ", ErrorCode: {0}" -f $Exception.ErrorCode
        }
    }

    return $infoStr
}

function CheckJobStatus
{
    param
    (
        [Parameter(Mandatory=$true)]
        [System.Management.Automation.Job[]]$Jobs
    )

    $ErrMsgs = @()
    foreach($job in $Jobs)
    {
        if($job.State -eq "Failed")
        {
            $ErrMsgs += "[{0}] {1}" -f $job.Name, $job.ChildJobs[0].JobStateInfo.Reason.Message
        }
    }
    
    if($ErrMsgs.Count -ne 0)
    {
        throw ($ErrMsgs -join "`r`n")
    }
}

function WriteJobLogsToFile
{
    Param
    (
        [Parameter(Mandatory=$true)]
        [System.Management.Automation.Job[]]$Jobs,

        [Parameter(Mandatory=$true)]
        [String] $LogFile
    )
    
    try
    {
        "" | Out-File $LogFile
        foreach($job in $jobArray)
        {
            "------------------------------Job '{0}' : {1} ---------------------------------" -f $job.Name, $job.State | Out-File $LogFile -Append
            if($job.ChildJobs[0].Verbose.Count -ne 0)
            {
                Add-Content $LogFile $job.ChildJobs[0].Verbose
            }

            if($job.ChildJobs[0].Error.Count -ne 0)
            {
                $job.ChildJobs[0].Error | %{Add-Content $LogFile ($_ | Out-String); Add-Content $LogFile (GetExceptionTypeInfo $_.Exception)}
            }

            if ($null -ne $job.ChildJobs[0].JobStateInfo.Reason)
            {
                Add-Content $LogFile $job.ChildJobs[0].JobStateInfo.Reason
                Add-Content $LogFile (GetExceptionTypeInfo $job.ChildJobs[0].JobStateInfo.Reason)
                if($null -ne $job.ChildJobs[0].JobStateInfo.Reason.SerializedRemoteInvocationInfo)
                {
                    Add-Content $LogFile ($job.ChildJobs[0].JobStateInfo.Reason.SerializedRemoteInvocationInfo | Out-String)
                }
            }

            Add-Content $LogFile ""
        }
    }
    catch
    {
        #Ignore the exception
    }
}

function GetCertificate
{
    param
    (
        [Parameter(Mandatory=$true)]
        [String] $FilePath, 

        [Parameter(Mandatory=$false)]
        [String] $Password = ""
    )

    if(-not (Test-Path -Path $FilePath -PathType Leaf))
    {
        throw "The certificate file $FilePath doesn't exist."
    }

    try
    {
        $cert = New-Object System.Security.Cryptography.X509Certificates.X509Certificate2 -ArgumentList $FilePath, $Password
        return $cert
    }
    catch [System.Management.Automation.MethodInvocationException]
    {
        throw $_.Exception.InnerException
    }
}

function GetThumbprintFromCertificateFile
{
    param
    (
        [Parameter(Mandatory=$true)]
        [String] $FilePath, 

        [Parameter(Mandatory=$false)]
        [String] $Password = ""
    )

    try
    {
        $cert = GetCertificate $FilePath $Password
        return $cert.Thumbprint 
    }
    catch
    {
        throw ("Failed to get the certificate thumbprint from $FilePath : " + $_)
    }
}

function CreateSelfSignedCertificate
{
    param
    (
        [Parameter(Mandatory=$true)]
        [String] $DnsName,

        [Parameter(Mandatory=$false)]
        [DateTime] $ExpiryDate,

        [Parameter(Mandatory=$false)]
		[Switch]$Exportable
    )

    $startDate = Get-Date
    if(($null -ne $ExpiryDate) -and ($ExpiryDate -le $startDate.AddDays(7)))
    {
        throw "The Expiry date is too close: $ExpiryDate"
    }

    $thumbprint = $null
    
    # GENERATE self signed certificate
    $name = new-object -com X509Enrollment.CX500DistinguishedName
    $name.Encode("CN=$DnsName", 0)

    # http://msdn.microsoft.com/en-us/library/aa378921(VS.85).aspx
    $key = new-object -com X509Enrollment.CX509PrivateKey
    $key.ProviderName = "Microsoft Enhanced Cryptographic Provider v1.0"
    $key.KeySpec = 1
    $key.Length = 2048
    if($Exportable.IsPresent)
    {
        $key.ExportPolicy = 1
    }
    else
    {
        $key.ExportPolicy = 0
    }

    #True for LocalMachine, and False for CurrentUser
    $key.MachineContext = $false
    $key.Create()

    $serverauthoid = new-object -com X509Enrollment.CObjectId
    $serverauthoid.InitializeFromValue("1.3.6.1.5.5.7.3.1")
    $clientauthoid = new-object -com X509Enrollment.CObjectId
    $clientauthoid.InitializeFromValue("1.3.6.1.5.5.7.3.2")
    $ekuoids = new-object -com X509Enrollment.CObjectIds
    $ekuoids.add($serverauthoid)
    $ekuoids.add($clientauthoid)
    $ekuext = new-object -com "X509Enrollment.CX509ExtensionEnhancedKeyUsage"
    $ekuext.InitializeEncode($ekuoids)

	$sigoid = New-Object -ComObject X509Enrollment.CObjectId
	$sigoid.InitializeFromValue(([Security.Cryptography.Oid]"SHA256").Value)

    # http://msdn.microsoft.com/en-us/library/aa377124(VS.85).aspx
    $cert = new-object -com "X509Enrollment.CX509CertificateRequestCertificate"
    $cert.InitializeFromPrivateKey(1, $key, "")
    $cert.Subject = $name
    $cert.Issuer = $cert.Subject
    $cert.NotBefore = $startDate
    if($null -eq $ExpiryDate)
    {
        $cert.NotAfter = $cert.NotBefore.AddYears(1)
    }
    else
    {
        $cert.NotAfter = $ExpiryDate
    }
    $cert.X509Extensions.Add($ekuext)
    $cert.SignatureInformation.HashAlgorithm = $sigoid
    $cert.Encode()

    # interface: http://msdn.microsoft.com/en-us/library/aa377809(VS.85).aspx
    $enrollment = new-object -com "X509Enrollment.CX509Enrollment"
    $enrollment.InitializeFromRequest($cert)
    $certdata = $enrollment.CreateRequest(1)
    [Byte[]]$bytes = [System.Convert]::FromBase64String($certdata)
    [Security.Cryptography.X509Certificates.X509Certificate2] $cert = [Security.Cryptography.X509Certificates.X509Certificate2]$bytes
    $res = $enrollment.InstallResponse(2, $certdata, 1, "")
    
    if($cert -ne $null)
    {   
        $thumbprint = $cert.Thumbprint
    }
    
    return $thumbprint
}
# SIG # Begin signature block
# MIIaxgYJKoZIhvcNAQcCoIIatzCCGrMCAQExCzAJBgUrDgMCGgUAMGkGCisGAQQB
# gjcCAQSgWzBZMDQGCisGAQQBgjcCAR4wJgIDAQAABBAfzDtgWUsITrck0sYpfvNR
# AgEAAgEAAgEAAgEAAgEAMCEwCQYFKw4DAhoFAAQUgIqAMGkJfP5SRKlOKsqb0HiR
# GH+gghWBMIIEwjCCA6qgAwIBAgITMwAAAJJMoq9VJwgudQAAAAAAkjANBgkqhkiG
# 9w0BAQUFADB3MQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4G
# A1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSEw
# HwYDVQQDExhNaWNyb3NvZnQgVGltZS1TdGFtcCBQQ0EwHhcNMTUxMDA3MTgxNDE0
# WhcNMTcwMTA3MTgxNDE0WjCBsjELMAkGA1UEBhMCVVMxEjAQBgNVBAgTCVdhc2lu
# Z3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBv
# cmF0aW9uMQ0wCwYDVQQLEwRNT1BSMScwJQYDVQQLEx5uQ2lwaGVyIERTRSBFU046
# N0QyRS0zNzgyLUIwRjcxJTAjBgNVBAMTHE1pY3Jvc29mdCBUaW1lLVN0YW1wIFNl
# cnZpY2UwggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQC6WVT9G7wUxF8u
# /fnFTnid7MCYX4X58613PUnaf2uYaz291cpmbxNeEsx+HZ8xrgjCHkMC3U9rTUfl
# oyhWqlW3ZdZQdn97Qa++X7wXa/ybE8FeY0Qphe8K0w9hbhxRjbII4fInEEkM4GAd
# HLqPqQw+U+Ul/gAC8U64SnklxtsjxN2faP98po9YqDYGH/IGaej0Y9ojGA2aEpVh
# J6n3TezIbXNZDBZW1ODKX1W0OmKPNvTdGqFYAHCr6osCrVLyg4ROozoI9GnsvjC7
# f9ACbPJf6Xy1B2v0teYREkUmpqc+OC/rZpApjgtL2Y5ymgeuihuSUj/XaKNtDa0Z
# ERONWgyLAgMBAAGjggEJMIIBBTAdBgNVHQ4EFgQUBsPfWqqHee6gVxN8Wohmb0CT
# pgMwHwYDVR0jBBgwFoAUIzT42VJGcArtQPt2+7MrsMM1sw8wVAYDVR0fBE0wSzBJ
# oEegRYZDaHR0cDovL2NybC5taWNyb3NvZnQuY29tL3BraS9jcmwvcHJvZHVjdHMv
# TWljcm9zb2Z0VGltZVN0YW1wUENBLmNybDBYBggrBgEFBQcBAQRMMEowSAYIKwYB
# BQUHMAKGPGh0dHA6Ly93d3cubWljcm9zb2Z0LmNvbS9wa2kvY2VydHMvTWljcm9z
# b2Z0VGltZVN0YW1wUENBLmNydDATBgNVHSUEDDAKBggrBgEFBQcDCDANBgkqhkiG
# 9w0BAQUFAAOCAQEAjgD2Z96da+Ze+YXIxGUX2pvvvX2etiR572Kwk6j6aXOFJrbB
# FaNelpipwJCRAY/V9qLIqUh+KfQFBKQYlRBf50WrCcXz+sx0BxyG597HjjGCmL4o
# Y0j/F0KATLMw60EcOh2I1hotO1a1W5fHB661OxD+T5KC6D9JN9TTP8vxap080i/V
# uNKyr2QubnfuOvs7jTjDJP5l5ZUEAFcxuliihARHhKnyoWxWcvje/fI463+pmRhF
# /nBuA3jTiCC5DWI3vST9I0l/BwrVDVMwvvnn5xf0vHb1U3TrJVeo2VRpHsqsoCA0
# 35Vuya6u01jEDkKhrZHuuMnxTAgCVuIFeXh9xDCCBOwwggPUoAMCAQICEzMAAAEK
# LHmu13l7pqwAAQAAAQowDQYJKoZIhvcNAQEFBQAweTELMAkGA1UEBhMCVVMxEzAR
# BgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1JlZG1vbmQxHjAcBgNVBAoTFU1p
# Y3Jvc29mdCBDb3Jwb3JhdGlvbjEjMCEGA1UEAxMaTWljcm9zb2Z0IENvZGUgU2ln
# bmluZyBQQ0EwHhcNMTUwNjA0MTc0MjQ1WhcNMTYwOTA0MTc0MjQ1WjCBgzELMAkG
# A1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1JlZG1vbmQx
# HjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjENMAsGA1UECxMETU9QUjEe
# MBwGA1UEAxMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMIIBIjANBgkqhkiG9w0BAQEF
# AAOCAQ8AMIIBCgKCAQEAkvxvNrvhA7ko1kbRomG6pUb7YxY+LvlH0sfs7ceZsSZR
# db3azA8GFBGkbBmewF9NVInkTVDYjc2hYaV3E5ocp+0NdPenVnnoKPdT0rF6Y+D1
# lJe37NlH+Gw98yWIs7wKxQSnjnyFSHYcYUaqGcR6YovBwjWq+1hvxWMLk0kwgRt7
# 3398T7RHbV94HK+295YTUu+50U055XPeSE48FKqXDTnMi1HhXNE78I5n6jBgqU1a
# nUO92yO6wA/XSxCdnE3wUaKEquScpz3Wo+8KGEio+rFOpZgOS7/wFPMAyLBI1lv+
# ONeJES0FukMCTyIAliWtfTVuhGirIBg4KP4cohCxpQIDAQABo4IBYDCCAVwwEwYD
# VR0lBAwwCgYIKwYBBQUHAwMwHQYDVR0OBBYEFIn+CjHqJs3mbceRkbmUjKIYYTc0
# MFEGA1UdEQRKMEikRjBEMQ0wCwYDVQQLEwRNT1BSMTMwMQYDVQQFEyozMTU5NSsw
# NDA3OTM1MC0xNmZhLTRjNjAtYjZiZi05ZDJiMWNkMDU5ODQwHwYDVR0jBBgwFoAU
# yxHoytK0FlgByTcuMxYWuUyaCh8wVgYDVR0fBE8wTTBLoEmgR4ZFaHR0cDovL2Ny
# bC5taWNyb3NvZnQuY29tL3BraS9jcmwvcHJvZHVjdHMvTWljQ29kU2lnUENBXzA4
# LTMxLTIwMTAuY3JsMFoGCCsGAQUFBwEBBE4wTDBKBggrBgEFBQcwAoY+aHR0cDov
# L3d3dy5taWNyb3NvZnQuY29tL3BraS9jZXJ0cy9NaWNDb2RTaWdQQ0FfMDgtMzEt
# MjAxMC5jcnQwDQYJKoZIhvcNAQEFBQADggEBAKaoU5HfOwH+YV0GWutm8AzY99mE
# 0wBRDhDRpNEXKOeMMzgr2EPGA4t16zOS9KniZ/0C26UdTkNFXRtJ4+BPFvB+j/CI
# EcqCraT7pqlf9ZdgyHvEv3ybae0fgsHxzY54S2L11w0c11MS1pZS7zW9GY6gQJOh
# CqUtFpzCRnQI69v02KVJNlQSEVUDs3sW+0f++2j8wEVc4j8SeTOi74LF3kAZB+4V
# xQqbWQVBqdCXnoGeA1v8SuMaLgWsUEcvjPp52B4g+AX7KWsYFPpyBLcLp5pk7BFd
# T0VJjSkaLfvQtglTXzSU4BbEuepzNehX7+HrFsMYxwazO89hhOK2RImUo4YwggW8
# MIIDpKADAgECAgphMyYaAAAAAAAxMA0GCSqGSIb3DQEBBQUAMF8xEzARBgoJkiaJ
# k/IsZAEZFgNjb20xGTAXBgoJkiaJk/IsZAEZFgltaWNyb3NvZnQxLTArBgNVBAMT
# JE1pY3Jvc29mdCBSb290IENlcnRpZmljYXRlIEF1dGhvcml0eTAeFw0xMDA4MzEy
# MjE5MzJaFw0yMDA4MzEyMjI5MzJaMHkxCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpX
# YXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQg
# Q29ycG9yYXRpb24xIzAhBgNVBAMTGk1pY3Jvc29mdCBDb2RlIFNpZ25pbmcgUENB
# MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAsnJZXBkwZL8dmmAgIEKZ
# dlNsPhvWb8zL8epr/pcWEODfOnSDGrcvoDLs/97CQk4j1XIA2zVXConKriBJ9PBo
# rE1LjaW9eUtxm0cH2v0l3511iM+qc0R/14Hb873yNqTJXEXcr6094CholxqnpXJz
# VvEXlOT9NZRyoNZ2Xx53RYOFOBbQc1sFumdSjaWyaS/aGQv+knQp4nYvVN0UMFn4
# 0o1i/cvJX0YxULknE+RAMM9yKRAoIsc3Tj2gMj2QzaE4BoVcTlaCKCoFMrdL109j
# 59ItYvFFPeesCAD2RqGe0VuMJlPoeqpK8kbPNzw4nrR3XKUXno3LEY9WPMGsCV8D
# 0wIDAQABo4IBXjCCAVowDwYDVR0TAQH/BAUwAwEB/zAdBgNVHQ4EFgQUyxHoytK0
# FlgByTcuMxYWuUyaCh8wCwYDVR0PBAQDAgGGMBIGCSsGAQQBgjcVAQQFAgMBAAEw
# IwYJKwYBBAGCNxUCBBYEFP3RMU7TJoqV4ZhgO6gxb6Y8vNgtMBkGCSsGAQQBgjcU
# AgQMHgoAUwB1AGIAQwBBMB8GA1UdIwQYMBaAFA6sgmBAVieX5SUT/CrhClOVWeSk
# MFAGA1UdHwRJMEcwRaBDoEGGP2h0dHA6Ly9jcmwubWljcm9zb2Z0LmNvbS9wa2kv
# Y3JsL3Byb2R1Y3RzL21pY3Jvc29mdHJvb3RjZXJ0LmNybDBUBggrBgEFBQcBAQRI
# MEYwRAYIKwYBBQUHMAKGOGh0dHA6Ly93d3cubWljcm9zb2Z0LmNvbS9wa2kvY2Vy
# dHMvTWljcm9zb2Z0Um9vdENlcnQuY3J0MA0GCSqGSIb3DQEBBQUAA4ICAQBZOT5/
# Jkav629AsTK1ausOL26oSffrX3XtTDst10OtC/7L6S0xoyPMfFCYgCFdrD0vTLqi
# qFac43C7uLT4ebVJcvc+6kF/yuEMF2nLpZwgLfoLUMRWzS3jStK8cOeoDaIDpVbg
# uIpLV/KVQpzx8+/u44YfNDy4VprwUyOFKqSCHJPilAcd8uJO+IyhyugTpZFOyBvS
# j3KVKnFtmxr4HPBT1mfMIv9cHc2ijL0nsnljVkSiUc356aNYVt2bAkVEL1/02q7U
# gjJu/KSVE+Traeepoiy+yCsQDmWOmdv1ovoSJgllOJTxeh9Ku9HhVujQeJYYXMk1
# Fl/dkx1Jji2+rTREHO4QFRoAXd01WyHOmMcJ7oUOjE9tDhNOPXwpSJxy0fNsysHs
# cKNXkld9lI2gG0gDWvfPo2cKdKU27S0vF8jmcjcS9G+xPGeC+VKyjTMWZR4Oit0Q
# 3mT0b85G1NMX6XnEBLTT+yzfH4qerAr7EydAreT54al/RrsHYEdlYEBOsELsTu2z
# dnnYCjQJbRyAMR/iDlTd5aH75UcQrWSY/1AWLny/BSF64pVBJ2nDk4+VyY3YmyGu
# DVyc8KKuhmiDDGotu3ZrAB2WrfIWe/YWgyS5iM9qqEcxL5rc43E91wB+YkfRzojJ
# uBj6DnKNwaM9rwJAav9pm5biEKgQtDdQCNbDPTCCBgcwggPvoAMCAQICCmEWaDQA
# AAAAABwwDQYJKoZIhvcNAQEFBQAwXzETMBEGCgmSJomT8ixkARkWA2NvbTEZMBcG
# CgmSJomT8ixkARkWCW1pY3Jvc29mdDEtMCsGA1UEAxMkTWljcm9zb2Z0IFJvb3Qg
# Q2VydGlmaWNhdGUgQXV0aG9yaXR5MB4XDTA3MDQwMzEyNTMwOVoXDTIxMDQwMzEz
# MDMwOVowdzELMAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAOBgNV
# BAcTB1JlZG1vbmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEhMB8G
# A1UEAxMYTWljcm9zb2Z0IFRpbWUtU3RhbXAgUENBMIIBIjANBgkqhkiG9w0BAQEF
# AAOCAQ8AMIIBCgKCAQEAn6Fssd/bSJIqfGsuGeG94uPFmVEjUK3O3RhOJA/u0afR
# TK10MCAR6wfVVJUVSZQbQpKumFwwJtoAa+h7veyJBw/3DgSY8InMH8szJIed8vRn
# HCz8e+eIHernTqOhwSNTyo36Rc8J0F6v0LBCBKL5pmyTZ9co3EZTsIbQ5ShGLies
# hk9VUgzkAyz7apCQMG6H81kwnfp+1pez6CGXfvjSE/MIt1NtUrRFkJ9IAEpHZhEn
# KWaol+TTBoFKovmEpxFHFAmCn4TtVXj+AZodUAiFABAwRu233iNGu8QtVJ+vHnhB
# MXfMm987g5OhYQK1HQ2x/PebsgHOIktU//kFw8IgCwIDAQABo4IBqzCCAacwDwYD
# VR0TAQH/BAUwAwEB/zAdBgNVHQ4EFgQUIzT42VJGcArtQPt2+7MrsMM1sw8wCwYD
# VR0PBAQDAgGGMBAGCSsGAQQBgjcVAQQDAgEAMIGYBgNVHSMEgZAwgY2AFA6sgmBA
# VieX5SUT/CrhClOVWeSkoWOkYTBfMRMwEQYKCZImiZPyLGQBGRYDY29tMRkwFwYK
# CZImiZPyLGQBGRYJbWljcm9zb2Z0MS0wKwYDVQQDEyRNaWNyb3NvZnQgUm9vdCBD
# ZXJ0aWZpY2F0ZSBBdXRob3JpdHmCEHmtFqFKoKWtTHNY9AcTLmUwUAYDVR0fBEkw
# RzBFoEOgQYY/aHR0cDovL2NybC5taWNyb3NvZnQuY29tL3BraS9jcmwvcHJvZHVj
# dHMvbWljcm9zb2Z0cm9vdGNlcnQuY3JsMFQGCCsGAQUFBwEBBEgwRjBEBggrBgEF
# BQcwAoY4aHR0cDovL3d3dy5taWNyb3NvZnQuY29tL3BraS9jZXJ0cy9NaWNyb3Nv
# ZnRSb290Q2VydC5jcnQwEwYDVR0lBAwwCgYIKwYBBQUHAwgwDQYJKoZIhvcNAQEF
# BQADggIBABCXisNcA0Q23em0rXfbznlRTQGxLnRxW20ME6vOvnuPuC7UEqKMbWK4
# VwLLTiATUJndekDiV7uvWJoc4R0Bhqy7ePKL0Ow7Ae7ivo8KBciNSOLwUxXdT6uS
# 5OeNatWAweaU8gYvhQPpkSokInD79vzkeJkuDfcH4nC8GE6djmsKcpW4oTmcZy3F
# UQ7qYlw/FpiLID/iBxoy+cwxSnYxPStyC8jqcD3/hQoT38IKYY7w17gX606Lf8U1
# K16jv+u8fQtCe9RTciHuMMq7eGVcWwEXChQO0toUmPU8uWZYsy0v5/mFhsxRVuid
# cJRsrDlM1PZ5v6oYemIp76KbKTQGdxpiyT0ebR+C8AvHLLvPQ7Pl+ex9teOkqHQ1
# uE7FcSMSJnYLPFKMcVpGQxS8s7OwTWfIn0L/gHkhgJ4VMGboQhJeGsieIiHQQ+kr
# 6bv0SMws1NgygEwmKkgkX1rqVu+m3pmdyjpvvYEndAYR7nYhv5uCwSdUtrFqPYmh
# dmG0bqETpr+qR/ASb/2KMmyy/t9RyIwjyWa9nR2HEmQCPS2vWY+45CHltbDKY7R4
# VAXUQS5QrJSwpXirs6CWdRrZkocTdSIvMqgIbqBbjCW/oO+EyiHW6x5PyZruSeD3
# AWVviQt9yGnI5m7qp5fOMSn/DsVbXNhNG6HY+i+ePy5VFmvJE6P9MYIErzCCBKsC
# AQEwgZAweTELMAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAOBgNV
# BAcTB1JlZG1vbmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEjMCEG
# A1UEAxMaTWljcm9zb2Z0IENvZGUgU2lnbmluZyBQQ0ECEzMAAAEKLHmu13l7pqwA
# AQAAAQowCQYFKw4DAhoFAKCByDAZBgkqhkiG9w0BCQMxDAYKKwYBBAGCNwIBBDAc
# BgorBgEEAYI3AgELMQ4wDAYKKwYBBAGCNwIBFTAjBgkqhkiG9w0BCQQxFgQUqtAC
# a5jHyxiEbUM6rYjaH7qHPNUwaAYKKwYBBAGCNwIBDDFaMFigNoA0AE0AaQBjAHIA
# bwBzAG8AZgB0ACAASABQAEMAIABQAGEAYwBrACAAMgAwADEAMgAgAFIAMqEegBxo
# dHRwOi8vd3d3Lm1pY3Jvc29mdC5jb20vSFBDMA0GCSqGSIb3DQEBAQUABIIBAHKd
# QPcC3g2HhbPRcR+4I6KjqMcRYQjHYFWLxGlC6K1n5IWNNm+iXrLJlk9U156cBpud
# E/3Ddu+3z9t2SVX4UssjhDlxdJsVb3UVdh6YfDVSuNn40QQd3NFuCHJ/G+Vn3SBe
# F6dKqFEz+4K6LqRBiGss7Y4uJKwPNDQfgm6+ZHFzxOn4WRJVohvl2jbxAaPu61JV
# cT/YHoa4e+l84UVUdW/ow1hkoMbtc/VOeKSa6Z41lPMv1oFgHOlTAMtolhImUlL3
# Lb5n2xVcsyEgrmVlUfW6Er74+jmtIDUhR5XDGA059kiUUslOLIsTqcndlM/U3r2Y
# 9YokS0/sGO9NDXVTvq+hggIoMIICJAYJKoZIhvcNAQkGMYICFTCCAhECAQEwgY4w
# dzELMAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1Jl
# ZG1vbmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEhMB8GA1UEAxMY
# TWljcm9zb2Z0IFRpbWUtU3RhbXAgUENBAhMzAAAAkkyir1UnCC51AAAAAACSMAkG
# BSsOAwIaBQCgXTAYBgkqhkiG9w0BCQMxCwYJKoZIhvcNAQcBMBwGCSqGSIb3DQEJ
# BTEPFw0xNTExMTMwMjQ1MDlaMCMGCSqGSIb3DQEJBDEWBBTvyfL9x7/H+KT0usjC
# JZ9t7pH6qjANBgkqhkiG9w0BAQUFAASCAQA89v2YtnXKKJo4R1s/cCskyy2BuQFK
# g8YEATBI9g6L2CfWuRl5bcBrdpdI7oWCUf4Q/2quYQ2hSCqAPYUxkbKbCl/imYZW
# b76c16KdA/OIuOKXydE4+gqRKHPsSr9pE75lgAXp+2hFgfLgM883+XzzbdEnvIWl
# nkDj53GYhFFLbUMQUjUPatcrXzoVkNSNUWo2VI6lFnHWnw0L+tmVJAOoHncj8ERH
# 85r6LhUj6PljHrjBlkQGr4ZXHPJKKsFLfRi5VS4ncik+AaVzXFy1avs+NGqpyA0x
# MzdxMDV+up5tunaLSjt5PrWQpsr3akdUIa81eMeYMaamFdLBbfjCadNk
# SIG # End signature block
