﻿<#
.SYNOPSIS
   Removes a Microsoft HPC Pack cluster from Microsoft Azure
.DESCRIPTION
   This script removes a Microsoft HPC cluster from Microsoft Azure

   Author :  Microsoft HPC Pack team
   Version:  4.5.0
.EXAMPLE
   .\Remove-HPCIaasCluster.ps1 -ConfigFile ".\HPCIaaSCluster.xml"
.EXAMPLE
   .\Remove-HPCIaasCluster.ps1 -ConfigFile ".\HPCIaaSCluster.xml" -KeepImage
#>
[CmdletBinding(SupportsShouldProcess = $true, ConfirmImpact = "High")]
Param
(
    [Parameter(Mandatory=$true)]
    [String] $ConfigFile,

    [Parameter(Mandatory=$false)]
    [Switch] $KeepImage,

    [Parameter(Mandatory=$false)]
    [Switch] $KeepVNet,

    [Parameter(Mandatory=$false)]
    [Switch] $KeepAffinityGroup,

    [Parameter(Mandatory=$false)]
    [Switch] $KeepStorageAccount,

    [Parameter(Mandatory=$false)]
    [Switch] $Force,

    [Parameter(Mandatory=$false)]
    [ValidateRange(0, 900)]
    [Int] $SecondsWaitingForDiskDeletion = 180
)

# Check the Windows Azure Powershell version
$azureModule = Get-Module -ListAvailable -Name Azure 
if ($null -eq $azureModule)
{
    throw "Azure Powershell not found. Install the latest version from http://azure.microsoft.com/downloads/#cmd-line-tools"
}
if ($azureModule.Version -lt "0.8.10")
{
    $ver = $azureModule.Version
    throw "The Azure Powershell Version $ver is too old. Install the latest version from http://azure.microsoft.com/downloads/#cmd-line-tools"
}

$removeImage = -not $KeepImage.IsPresent
$removeVNet = -not $KeepVNet.IsPresent
$removeAG = -not $KeepAffinityGroup.IsPresent
$removeStoreAcc = -not $KeepStorageAccount.IsPresent

"Checking config file"
$clusterConfig = [xml](Get-Content $ConfigFile -ErrorAction Stop)
$iaasConfig = $clusterConfig.IaaSClusterConfig

if($null -ne $iaasConfig.Subscription["SubscriptionName"])
{
    $subName = $iaasConfig.Subscription.SubscriptionName.Trim()
    $subId = $null
}
else
{
    $subId = $iaasConfig.Subscription.SubscriptionId.Trim()
    $subName = $null
}

$storeAcc = $iaasConfig.Subscription.StorageAccount
$AGName = $iaasConfig.AffinityGroup
$VNetName = $iaasConfig.VNet.VNetName

$roleList = @()
$DNSName = $null

$newDC =  'NewDC' -eq $iaasConfig.Domain.DCOption
$headNodeAsDC = 'HeadNodeAsDC' -eq $iaasConfig.Domain.DCOption
if ($newDC)
{
    # used to remove the DNS server setting
    $DNSName = $iaasConfig.Domain.DomainController.VMName

    $DCSvc = $iaasConfig.Domain.DomainController.ServiceName
    $DCAG = $iaasConfig.Domain.DomainController.AffinityGroup
    $roleList += @{ RoleName = "Domain Controller"; Svcs = @($DCSvc); AG = $DCAG }
}
elseif ($headNodeAsDC)
{
    # HeadNodeAsDC also adds DNS server setting
    $DNSName = $iaasConfig.HeadNode.VMName
}

$newDB = 'NewRemoteDB' -eq $iaasConfig.Database.DBOption
if ($newDB)
{
    $DBSvc = $iaasConfig.Database.DBServer.ServiceName
    $DBAG = $iaasConfig.Database.DBServer.AffinityGroup
    $roleList += @{ RoleName = "DB Server"; Svcs = @($DBSvc); AG = $DBAG }
}

$HNSvc = $iaasConfig.HeadNode.ServiceName
$HNAG = $iaasConfig.HeadNode.AffinityGroup
$roleList += @{ RoleName = "Head Node"; Svcs = @($HNSvc); AG = $HNAG }

$hasCN = $null -ne $iaasConfig.ComputeNodes -and 0 -ne $iaasConfig.ComputeNodes.NodeCount
if ($hasCN)
{
    $CNAG = $iaasConfig.ComputeNodes.AffinityGroup
    $CNHPCInstalled = ($null -ne $iaasConfig.ComputeNodes.ImageName) -and ($iaasConfig.ComputeNodes.ImageName.HPCPackInstalled -eq $true)
    if ($CNHPCInstalled -eq $true)
    {
        $CNImageName = $iaasConfig.ComputeNodes.ImageName.InnerText.Trim()
    }

    $CNSvc = $iaasConfig.ComputeNodes.ServiceName
    if ($null -ne $CNSvc)
    {
        # single service
        $roleList += @{ RoleName = "Compute Node"; Svcs = @($CNSvc); AG = $CNAG }
        if ($CNHPCInstalled -eq $true)
        {
            # only if the image was copied to storeAcc by the script
            $CNImagePattern = "${CNImageName}_$storeAcc"
        }
    }
    else
    {
        # service pattern
        $count = [Math]::Ceiling($iaasConfig.ComputeNodes.NodeCount / $iaasConfig.ComputeNodes.MaxNodeCountPerService)

        $svcPattern = $iaasConfig.ComputeNodes.ServiceNamePattern.Trim(" %").Split("%") | ? { -not [String]::IsNullOrWhiteSpace($_) }
        $svcRoot = $svcPattern[0]
        $svcStartsFrom = [Int]::Parse($svcPattern[1])
        $svcDigitWidth = $svcPattern[1].Length

        $storePattern = $iaasConfig.ComputeNodes.StorageAccountNamePattern.Trim(" %").Split("%") | ? { -not [String]::IsNullOrWhiteSpace($_) }
        $storeRoot = $storePattern[0]
        $storeStartsFrom = [Int]::Parse($storePattern[1])
        $storeDigitWidth = $storePattern[1].Length

        if ($CNHPCInstalled -eq $true)
        {
            if ($CNImageName -match "_$storeRoot[0-9]{$storeDigitWidth}$")
            {
                $CNImagePattern = $CNImageName.Substring(0, $CNImageName.Length - $storeDigitWidth)
            }
            else
            {
                $CNImagePattern = "${CNImageName}_$storeRoot"
            }
        }

        $CNSvcs = @()
        $CNStores = @()

        for ($i = 0; $i -lt $count; $i++)
        {
            $CNSvcs += "$svcRoot$(($svcStartsFrom + $i).ToString("D$svcDigitWidth"))"
            $CNStores += "$storeRoot$(($storeStartsFrom + $i).ToString("D$storeDigitWidth"))"
        }

        $roleList += @{ RoleName = "Compute Node"; Svcs = $CNSvcs; Stores = $CNStores; AG = $CNAG }
    }
}

$hasBN = $null -ne $iaasConfig.BrokerNodes -and 0 -ne $iaasConfig.BrokerNodes.NodeCount
if ($hasBN)
{
    $BNHPCInstalled = ($null -ne $iaasConfig.BrokerNodes.ImageName) -and ($iaasConfig.BrokerNodes.ImageName.HPCPackInstalled -eq $true)
    if ($BNHPCInstalled -eq $true)
    {
        $BNImageName = $iaasConfig.BrokerNodes.ImageName.InnerText.Trim()
        $BNImagePattern = "${BNImageName}_$storeAcc"
    }

    $BNSvc = $iaasConfig.BrokerNodes.ServiceName
    $BNAG = $iaasConfig.BrokerNodes.AffinityGroup
    $roleList += @{ RoleName = "Broker Node"; Svcs = @($BNSvc); AG = $BNAG }
}

$hasLN = $null -ne $iaasConfig.LinuxComputeNodes -and 0 -ne $iaasConfig.LinuxComputeNodes.NodeCount
if ($hasLN)
{
    $LNAG = $iaasConfig.LinuxComputeNodes.AffinityGroup

    $LNSvc = $iaasConfig.LinuxComputeNodes.ServiceName
    if ($null -ne $LNSvc)
    {
        # single service
        $roleList += @{ RoleName = "Linux Compute Node"; Svcs = @($LNSvc); AG = $LNAG }
    }
    else
    {
        # service pattern
        $count = [Math]::Ceiling($iaasConfig.LinuxComputeNodes.NodeCount / $iaasConfig.LinuxComputeNodes.MaxNodeCountPerService)

        $svcPattern = $iaasConfig.LinuxComputeNodes.ServiceNamePattern.Trim(" %").Split("%") | ? { -not [String]::IsNullOrWhiteSpace($_) }
        $svcRoot = $svcPattern[0]
        $svcStartsFrom = [Int]::Parse($svcPattern[1])
        $svcDigitWidth = $svcPattern[1].Length

        $storePattern = $iaasConfig.LinuxComputeNodes.StorageAccountNamePattern.Trim(" %").Split("%") | ? { -not [String]::IsNullOrWhiteSpace($_) }
        $storeRoot = $storePattern[0]
        $storeStartsFrom = [Int]::Parse($storePattern[1])
        $storeDigitWidth = $storePattern[1].Length

        $LNSvcs = @()
        $LNStores = @()

        for ($i = 0; $i -lt $count; $i++)
        {
            $LNSvcs += "$svcRoot$(($svcStartsFrom + $i).ToString("D$svcDigitWidth"))"
            $LNStores += "$storeRoot$(($storeStartsFrom + $i).ToString("D$storeDigitWidth"))"
        }

        $roleList += @{ RoleName = "Linux Compute Node"; Svcs = $LNSvcs; Stores = $LNStores; AG = $LNAG }
    }
}

"Scanning Azure deployment"
if([String]::IsNullOrEmpty($subName))
{
    Select-AzureSubscription -SubscriptionId $subId | Out-Null
}
else
{
    Select-AzureSubscription -SubscriptionName $subName | Out-Null
}

$listAGs = @()
$listSvcs = @()
$listImgs = @()
$listStores = @()
$VNetExisting = $false
$DNSExisting = $false

$createdSvcs = @(Get-AzureService | % { $_.ServiceName })
$createdStores = $null
if ($removeStoreAcc) { $createdStores = @(Get-AzureStorageAccount -WarningAction SilentlyContinue | % { $_.StorageAccountName }) }
$createdAGs = $null
if ($removeAG) { $createdAGs = @(Get-AzureAffinityGroup | % { $_.Name }) }

foreach ($role in $roleList)
{
    foreach ($svcName in $role.Svcs)
    {
        if ($svcName -notin $createdSvcs)
        {
            "$($role.RoleName)'s Cloud Service $svcName does not exist"
        }
        else
        {
            if ($listSvcs -notcontains $svcName)
            {
                $listSvcs += $svcName
            }
        }
    }

    if ($removeStoreAcc)
    {
        foreach ($storeName in $role.Stores)
        {
            if ($storeName -notin $createdStores)
            {
                "$($role.RoleName)'s Storage Account $storeName does not exist"
            }
            else
            {
                if ($listStores -notcontains $storeName)
                {
                    $listStores += $storeName
                }
            }
        }
    }

    if ($removeAG -and -not [string]::IsNullOrEmpty($role.AG))
    {
        if ($role.AG -notin $createdAGs)
        {
            "$($role.RoleName)'s Affinity Group $($role.AG) does not exist"
        }
        else
        {
            if ($listAGs -notcontains $role.AG)
            {
                $listAGs += $role.AG
            }
        }
    }
}

if ($removeImage)
{
    $imgs = Get-AzureVMImage
    if($hasCN)
    {
        if ($CNHPCInstalled -eq $true)
        {
            if ($null -ne $CNSvc)
            {
                # $pattern contains the expected full name in single service case
                $ReplicatedCNImage = $imgs | ? { $_.ImageName -eq $CNImagePattern -and [string]::IsNullOrWhiteSpace($_.ServiceName) }
                if ($null -ne $ReplicatedCNImage)
                {
                    $listImgs += $ReplicatedCNImage.ImageName
                }
            }
            else
            {
                # $pattern contains such foamrt: hpcInstalledImage_stgroot (no suffix digits)
                $ReplicatedCNImages = @($imgs | ? { $_.ImageName.StartsWith($CNImagePattern, "OrdinalIgnoreCase") -and [string]::IsNullOrWhiteSpace($_.ServiceName) })
                $listImgs += $ReplicatedCNImages | % { $_.ImageName } | ? { $_ -ne $CNImageName }
            }
        }
        else
        {
            $svcList = @()
            if ($null -ne $CNSvc)
            {
                $svcList += $CNSvc
            }
            else
            {
                $svcList += $CNSvcs
            }

            $rootImages = @($imgs | ? { $_.ImageName -like "HPC-Pack-ComputeNode_*" -and $_.ServiceName -in $svcList })

            if ($rootImages.Count -eq 0)
            {
                "Compute Node Image does not exist"
            }
            else
            {
                $svcsHostingMultiImages = @($rootImages | % { $_.ServiceName } | Group-Object | ? { $_.Count -gt 1 })
                if ($svcsHostingMultiImages.Count -gt 0)
                {
                    Write-Warning "More than one Compute Node Images captured from the same Cloud Service are found. (e.g. from $($svcsHostingMultiImages[0].Name))"
                    Write-Warning "Skipping Compute Node Image"
                }
                else
                {
                    $imgNames = @($rootImages | % { $_.ImageName })
                    foreach ($name in $imgNames)
                    {
                        $listImgs += $imgs | % { $_.ImageName } | ? { $_.StartsWith($name, "OrdinalIgnoreCase") }
                    }
                }
            }
        }
    }

    if($hasBN)
    {
        if ($BNHPCInstalled -eq $true)
        {
            $ReplicatedBNImage = $imgs | ? { $_.ImageName -eq $BNImagePattern -and [string]::IsNullOrWhiteSpace($_.ServiceName) }
            if ($null -ne $ReplicatedBNImage)
            {
                $listImgs += $ReplicatedBNImage.ImageName
            }
        }
        else
        {
            $BNImage = @($imgs | ? { $_.ImageName -like "HPC-Pack-BrokerNode_*" -and $_.ServiceName -eq $BNSvc })
            if ($BNImage.Count -gt 1)
            {
                Write-Warning "More than one Broker Node Images are found. Skipping Broker Node Image"
            }
            elseif ($BNImage.Count -eq 0)
            {
                "Broker Node Image does not exist"
            }
            else
            {
                $listImgs += $BNImage[0].ImageName
            }
        }
    }
}

if ($removeStoreAcc)
{
    if ($storeAcc -notin $createdStores)
    {
        "Config-wide Storage Account $storeAcc does not exist"
    }
    else
    {
        if ($listStores -notcontains $storeAcc)
        {
            $listStores += $storeAcc
        }
    }
}

if ($removeAG -and -not [string]::IsNullOrEmpty($AGName))
{
    if ($AGName -notin $createdAGs)
    {
        "Config-wide Affinity Group $AGName does not exist"
    }
    else
    {
        if ($listAGs -notcontains $AGName)
        {
            $listAGs += $AGName
        }
    }
}

if ($removeVNet -or $null -ne $DNSName)
{
    $VNConfig = [xml](Get-AzureVNetConfig).XMLConfiguration
}
if ($removeVNet)
{
    $sites = $VNConfig.NetworkConfiguration.VirtualNetworkConfiguration.VirtualNetworkSites
    if ($null -ne $sites)
    {
        $targetSite = $sites.ChildNodes | ? { $_.Name -eq $VNetName }
        if ($null -eq $targetSite)
        {
            "Virtual Network $VNetName does not exist"
        }
        else
        {
            $VNetExisting = $true
        }
    }
}

if ($null -ne $DNSName)
{
    $dnsServers = $VNConfig.NetworkConfiguration.VirtualNetworkConfiguration.Dns.DnsServers
    if ($null -ne $dnsServers)
    {
        $targetDns = $dnsServers.ChildNodes | ? { $_.Name -eq $DNSName }
        if ($null -eq $targetDns)
        {
            "DNS Server setting for $DNSName does not exist"
        }
        else
        {
            $ns = New-Object System.Xml.XmlNamespaceManager($VNConfig.NameTable)
            $ns.AddNamespace("ns", $VNConfig.NetworkConfiguration.NamespaceURI)
            $sitesUsingThisDNS = @($VNConfig.SelectNodes("//ns:NetworkConfiguration/ns:VirtualNetworkConfiguration/ns:VirtualNetworkSites/ns:VirtualNetworkSite[ns:DnsServersRef/ns:DnsServerRef/@name='$DNSName']", $ns)) | % { $_.name }
            $okToRemoveDNS = $true
            foreach ($siteName in $sitesUsingThisDNS)
            {
                if ($siteName -ne $VNetName)
                {
                    $okToRemoveDNS = $false
                    Write-Warning "DNS Server $DNSName (Domain Controller of this cluster) is referenced by non-HPC VNet(s). (e.g. $siteName)"
                    Write-Warning "Skipping DNS Server setting"
                    break;
                }
            }
            if ($okToRemoveDNS)
            {
                $DNSExisting = $true
            }
        }
    }
}

if (-not $VNetExisting -and -not $DNSExisting -and $listStores.Count -eq 0 -and $listImgs.Count -eq 0 -and $listAGs.Count -eq 0 -and $listSvcs.Count -eq 0)
{
    "All the resources have already been removed"
    return
}

"The following Azure resource(s) will be removed:"
if ($listSvcs.Count -gt 0)
{
    "Cloud Service: $($listSvcs -join ',')"
}
if ($listImgs.Count -gt 0)
{
    "Image: $($listImgs -join ',')"
}
if ($listAGs.Count -gt 0)
{
    "Affinity Group: $($listAGs -join ',')"
}
if ($listStores.Count -gt 0)
{
    "Storage Account: $($listStores -join ',')"
}
if ($VNetExisting)
{
    "Virtual Network: $VNetName"
}
if ($DNSExisting)
{
    "DNS Server Setting: $DNSName"
}

if (-not $Force.IsPresent)
{
    if ($PSCmdlet.ShouldProcess("The HPC Cluster will be removed from Azure", "Continue?", "Confirm:") -eq $false)
    {
        "Remove operation aborted"
        return
    }
}

if ($listSvcs.Count -gt 0)
{
    "Removing Cloud Services"
    $listSvcs | % { Remove-AzureService -ServiceName $_ -Force -DeleteAll -ErrorAction Stop | Out-Null }

    "Checking Cloud Services Removed"
    $timeout = 900
    $beginTime = [DateTime]::Now
    while ($true)
    {
        if (([DateTime]::Now - $beginTime).TotalSeconds -ge $timeout)
        {
            throw "Failed to remove Cloud Services"
        }
    
        $remainingSvcs = @(Get-AzureService | % { $_.ServiceName })

        Start-Sleep -Seconds 15
        $listSvcs | % { if ($svc -in $remainingSvcs) { continue } }

        break
    }
}

if ($listImgs.Count -gt 0)
{
    "Removing VM Images"
    $listImgs | % { Remove-AzureVMImage -ImageName $_ -DeleteVHD -ErrorAction Stop | Out-Null }
}

if (($listSvcs.Count -gt 0 -or $listImgs.Count -gt 0) -and $SecondsWaitingForDiskDeletion -gt 0)
{
    "Waiting $SecondsWaitingForDiskDeletion seconds for Azure to delete the Services/Images' associated vhds"
    Start-Sleep -Seconds $SecondsWaitingForDiskDeletion
}

"Checking Azure Disks"
$retryLimit = 5
$retryCount = 0

while ($retryCount -lt $retryLimit)
{
    $allDisks = Get-AzureDisk
    $listDisks = @()

    if ($newDC)
    {
        $DCDisks = @($allDisks | ? { $_.DiskName.StartsWith("$DCSvc-", "OrdinalIgnoreCase") })
        if ($DCDisks.Count -gt 0)
        {
            $listDisks += $DCDisks
        }
    }

    if ($newDB)
    {
        $DBDisks = @($allDisks | ? { $_.DiskName.StartsWith("$DBSvc-", "OrdinalIgnoreCase") })
        if ($DBDisks.Count -gt 0)
        {
            $listDisks += $DBDisks
        }
    }

    $HNDisks = @($allDisks | ? { $_.DiskName.StartsWith("$HNSvc-", "OrdinalIgnoreCase") })
    if ($HNDisks.Count -gt 0)
    {
        $listDisks += $HNDisks
    }

    if ($hasCN)
    {
        $svcList = @()
        if ($null -ne $CNSvc)
        {
            $svcList += $CNSvc
        }
        else
        {
            $svcList += $CNSvcs
        }

        foreach ($svc in $svcList)
        {
            $CNDisks = @($allDisks | ? { $_.DiskName.StartsWith("$svc-", "OrdinalIgnoreCase") })
            if ($CNDisks.Count -gt 0)
            {
                $listDisks += $CNDisks
            }
        }
    }

    if ($hasBN)
    {
        $BNDisks = @($allDisks | ? { $_.DiskName.StartsWith("$BNSvc-", "OrdinalIgnoreCase") })
        if ($BNDisks.Count -gt 0)
        {
            $listDisks += $BNDisks
        }
    }
    
    if ($hasLN)
    {
        $svcList = @()
        if ($null -ne $LNSvc)
        {
            $svcList += $LNSvc
        }
        else
        {
            $svcList += $LNSvcs
        }

        foreach ($svc in $svcList)
        {
            $LNDisks = @($allDisks | ? { $_.DiskName.StartsWith("$svc-", "OrdinalIgnoreCase") })
            if ($LNDisks.Count -gt 0)
            {
                $listDisks += $LNDisks
            }
        }
    }

    if ($listDisks.Count -eq 0)
    {
        break
    }

    Start-Sleep -Seconds 30
    $retryCount++
}

if ($listDisks.Count -gt 0)
{
    "Removing Azure Disks"
    $listDisks | Remove-AzureDisk -DeleteVHD -ErrorAction Stop | Out-Null
    "Waiting $SecondsWaitingForDiskDeletion seconds for Azure to delete the remaining VHDs detected in the last step"
    Start-Sleep -Seconds $SecondsWaitingForDiskDeletion
}

$storesFailedToRemove = @()

if ($listStores.Count -gt 0)
{
    "Removing Storage Account"
    foreach ($store in $listStores)
    {
        Remove-AzureStorageAccount -StorageAccountName $store | Out-Null
        if (-not $?)
        {
            $storesFailedToRemove += $store
        }
    }
}

$agsFailedToRemove = @()

if ($listAGs.Count -gt 0)
{
    "Removing Affinity Groups"
    foreach ($agToRemove in $listAGs)
    {
        Remove-AzureAffinityGroup -Name $agToRemove | Out-Null
        if (-not $?)
        {
            $agsFailedToRemove += $agToRemove
        }
    }
}

$netCfgUpdated = $true

if ($DNSExisting -or $VNetExisting)
{
    "Retrieving Azure Virtual Network Configuration"
    $VNConfig = [xml](Get-AzureVNetConfig).XMLConfiguration

    if ($DNSExisting)
    {
        "Removing DNS Server settings from the configuration"
        $dnsServers = $VNConfig.NetworkConfiguration.VirtualNetworkConfiguration.Dns.DnsServers
        $targetDns = $dnsServers.ChildNodes | ? { $_.Name -eq $DNSName }
        $VNConfig.NetworkConfiguration.VirtualNetworkConfiguration.Dns.DnsServers.RemoveChild($targetDns) | Out-Null

        if ($VNetExisting -eq $false)
        {
            $vNetSites = $VNConfig.NetworkConfiguration.VirtualNetworkConfiguration.VirtualNetworkSites
            if ($null -ne $vNetSites)
            {
                $dnsSite = $vNetSites.ChildNodes | ? { $_.Name -eq $VNetName }
                if ($null -ne $dnsSite -and $null -ne $dnsSite.DnsServersRef)
                {
                    $refedDns = $dnsSite.DnsServersRef.ChildNodes | ? { $_.Name -eq $DNSName }
                    if ($null -ne $refedDns)
                    {
                        $dnsSite.DnsServersRef.RemoveChild($refedDns) | Out-Null
                    }
                }
            }
        }
    }

    if ($VNetExisting)
    {
        "Removing Virtual Network from the configuration"
        $sites = $VNConfig.NetworkConfiguration.VirtualNetworkConfiguration.VirtualNetworkSites
        $targetSite = $sites.ChildNodes | ? { $_.Name -eq $VNetName }
        $VNConfig.NetworkConfiguration.VirtualNetworkConfiguration.VirtualNetworkSites.RemoveChild($targetSite) | Out-Null
    }

    $fileName = "$PSScriptRoot\RemoveCluster_NetCfg_$(Get-Random -Minimum 100000 -Maximum 1000000).xml"
    "Temp NetCfg file is saved to $fileName"
    $VNConfig.Save($fileName)
    "Updating the configuration to Azure"
    Set-AzureVNetConfig -ConfigurationPath $fileName | Out-Null
    if (-not $?)
    {
        $netCfgUpdated = $false
    }
    Remove-Item -Path $fileName
}

if ($storesFailedToRemove.Count -gt 0 -or $agsFailedToRemove.Count -gt 0 -or $netCfgUpdated -eq $false)
{
    Write-Warning "Errors occured while removing the following resources:"
    if ($storesFailedToRemove.Count -gt 0) { Write-Warning "Storage Account: $($storesFailedToRemove -join ',')" }
    if ($agsFailedToRemove.Count -gt 0) { Write-Warning "Affinity Group: $($agsFailedToRemove -join ',')" }
    if (-not $netCfgUpdated) { Write-Warning "VNet and/or DNS settings" }
    Write-Warning "Check whether they are held by other Azure resources (e.g. images) and try again"
}

"Finished removing IaaS cluster"
# SIG # Begin signature block
# MIIaxgYJKoZIhvcNAQcCoIIatzCCGrMCAQExCzAJBgUrDgMCGgUAMGkGCisGAQQB
# gjcCAQSgWzBZMDQGCisGAQQBgjcCAR4wJgIDAQAABBAfzDtgWUsITrck0sYpfvNR
# AgEAAgEAAgEAAgEAAgEAMCEwCQYFKw4DAhoFAAQUpgqpM48jagw3leaSa/bPwQB3
# sYCgghWBMIIEwjCCA6qgAwIBAgITMwAAAJJMoq9VJwgudQAAAAAAkjANBgkqhkiG
# 9w0BAQUFADB3MQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4G
# A1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSEw
# HwYDVQQDExhNaWNyb3NvZnQgVGltZS1TdGFtcCBQQ0EwHhcNMTUxMDA3MTgxNDE0
# WhcNMTcwMTA3MTgxNDE0WjCBsjELMAkGA1UEBhMCVVMxEjAQBgNVBAgTCVdhc2lu
# Z3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBv
# cmF0aW9uMQ0wCwYDVQQLEwRNT1BSMScwJQYDVQQLEx5uQ2lwaGVyIERTRSBFU046
# N0QyRS0zNzgyLUIwRjcxJTAjBgNVBAMTHE1pY3Jvc29mdCBUaW1lLVN0YW1wIFNl
# cnZpY2UwggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQC6WVT9G7wUxF8u
# /fnFTnid7MCYX4X58613PUnaf2uYaz291cpmbxNeEsx+HZ8xrgjCHkMC3U9rTUfl
# oyhWqlW3ZdZQdn97Qa++X7wXa/ybE8FeY0Qphe8K0w9hbhxRjbII4fInEEkM4GAd
# HLqPqQw+U+Ul/gAC8U64SnklxtsjxN2faP98po9YqDYGH/IGaej0Y9ojGA2aEpVh
# J6n3TezIbXNZDBZW1ODKX1W0OmKPNvTdGqFYAHCr6osCrVLyg4ROozoI9GnsvjC7
# f9ACbPJf6Xy1B2v0teYREkUmpqc+OC/rZpApjgtL2Y5ymgeuihuSUj/XaKNtDa0Z
# ERONWgyLAgMBAAGjggEJMIIBBTAdBgNVHQ4EFgQUBsPfWqqHee6gVxN8Wohmb0CT
# pgMwHwYDVR0jBBgwFoAUIzT42VJGcArtQPt2+7MrsMM1sw8wVAYDVR0fBE0wSzBJ
# oEegRYZDaHR0cDovL2NybC5taWNyb3NvZnQuY29tL3BraS9jcmwvcHJvZHVjdHMv
# TWljcm9zb2Z0VGltZVN0YW1wUENBLmNybDBYBggrBgEFBQcBAQRMMEowSAYIKwYB
# BQUHMAKGPGh0dHA6Ly93d3cubWljcm9zb2Z0LmNvbS9wa2kvY2VydHMvTWljcm9z
# b2Z0VGltZVN0YW1wUENBLmNydDATBgNVHSUEDDAKBggrBgEFBQcDCDANBgkqhkiG
# 9w0BAQUFAAOCAQEAjgD2Z96da+Ze+YXIxGUX2pvvvX2etiR572Kwk6j6aXOFJrbB
# FaNelpipwJCRAY/V9qLIqUh+KfQFBKQYlRBf50WrCcXz+sx0BxyG597HjjGCmL4o
# Y0j/F0KATLMw60EcOh2I1hotO1a1W5fHB661OxD+T5KC6D9JN9TTP8vxap080i/V
# uNKyr2QubnfuOvs7jTjDJP5l5ZUEAFcxuliihARHhKnyoWxWcvje/fI463+pmRhF
# /nBuA3jTiCC5DWI3vST9I0l/BwrVDVMwvvnn5xf0vHb1U3TrJVeo2VRpHsqsoCA0
# 35Vuya6u01jEDkKhrZHuuMnxTAgCVuIFeXh9xDCCBOwwggPUoAMCAQICEzMAAAEK
# LHmu13l7pqwAAQAAAQowDQYJKoZIhvcNAQEFBQAweTELMAkGA1UEBhMCVVMxEzAR
# BgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1JlZG1vbmQxHjAcBgNVBAoTFU1p
# Y3Jvc29mdCBDb3Jwb3JhdGlvbjEjMCEGA1UEAxMaTWljcm9zb2Z0IENvZGUgU2ln
# bmluZyBQQ0EwHhcNMTUwNjA0MTc0MjQ1WhcNMTYwOTA0MTc0MjQ1WjCBgzELMAkG
# A1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1JlZG1vbmQx
# HjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjENMAsGA1UECxMETU9QUjEe
# MBwGA1UEAxMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMIIBIjANBgkqhkiG9w0BAQEF
# AAOCAQ8AMIIBCgKCAQEAkvxvNrvhA7ko1kbRomG6pUb7YxY+LvlH0sfs7ceZsSZR
# db3azA8GFBGkbBmewF9NVInkTVDYjc2hYaV3E5ocp+0NdPenVnnoKPdT0rF6Y+D1
# lJe37NlH+Gw98yWIs7wKxQSnjnyFSHYcYUaqGcR6YovBwjWq+1hvxWMLk0kwgRt7
# 3398T7RHbV94HK+295YTUu+50U055XPeSE48FKqXDTnMi1HhXNE78I5n6jBgqU1a
# nUO92yO6wA/XSxCdnE3wUaKEquScpz3Wo+8KGEio+rFOpZgOS7/wFPMAyLBI1lv+
# ONeJES0FukMCTyIAliWtfTVuhGirIBg4KP4cohCxpQIDAQABo4IBYDCCAVwwEwYD
# VR0lBAwwCgYIKwYBBQUHAwMwHQYDVR0OBBYEFIn+CjHqJs3mbceRkbmUjKIYYTc0
# MFEGA1UdEQRKMEikRjBEMQ0wCwYDVQQLEwRNT1BSMTMwMQYDVQQFEyozMTU5NSsw
# NDA3OTM1MC0xNmZhLTRjNjAtYjZiZi05ZDJiMWNkMDU5ODQwHwYDVR0jBBgwFoAU
# yxHoytK0FlgByTcuMxYWuUyaCh8wVgYDVR0fBE8wTTBLoEmgR4ZFaHR0cDovL2Ny
# bC5taWNyb3NvZnQuY29tL3BraS9jcmwvcHJvZHVjdHMvTWljQ29kU2lnUENBXzA4
# LTMxLTIwMTAuY3JsMFoGCCsGAQUFBwEBBE4wTDBKBggrBgEFBQcwAoY+aHR0cDov
# L3d3dy5taWNyb3NvZnQuY29tL3BraS9jZXJ0cy9NaWNDb2RTaWdQQ0FfMDgtMzEt
# MjAxMC5jcnQwDQYJKoZIhvcNAQEFBQADggEBAKaoU5HfOwH+YV0GWutm8AzY99mE
# 0wBRDhDRpNEXKOeMMzgr2EPGA4t16zOS9KniZ/0C26UdTkNFXRtJ4+BPFvB+j/CI
# EcqCraT7pqlf9ZdgyHvEv3ybae0fgsHxzY54S2L11w0c11MS1pZS7zW9GY6gQJOh
# CqUtFpzCRnQI69v02KVJNlQSEVUDs3sW+0f++2j8wEVc4j8SeTOi74LF3kAZB+4V
# xQqbWQVBqdCXnoGeA1v8SuMaLgWsUEcvjPp52B4g+AX7KWsYFPpyBLcLp5pk7BFd
# T0VJjSkaLfvQtglTXzSU4BbEuepzNehX7+HrFsMYxwazO89hhOK2RImUo4YwggW8
# MIIDpKADAgECAgphMyYaAAAAAAAxMA0GCSqGSIb3DQEBBQUAMF8xEzARBgoJkiaJ
# k/IsZAEZFgNjb20xGTAXBgoJkiaJk/IsZAEZFgltaWNyb3NvZnQxLTArBgNVBAMT
# JE1pY3Jvc29mdCBSb290IENlcnRpZmljYXRlIEF1dGhvcml0eTAeFw0xMDA4MzEy
# MjE5MzJaFw0yMDA4MzEyMjI5MzJaMHkxCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpX
# YXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQg
# Q29ycG9yYXRpb24xIzAhBgNVBAMTGk1pY3Jvc29mdCBDb2RlIFNpZ25pbmcgUENB
# MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAsnJZXBkwZL8dmmAgIEKZ
# dlNsPhvWb8zL8epr/pcWEODfOnSDGrcvoDLs/97CQk4j1XIA2zVXConKriBJ9PBo
# rE1LjaW9eUtxm0cH2v0l3511iM+qc0R/14Hb873yNqTJXEXcr6094CholxqnpXJz
# VvEXlOT9NZRyoNZ2Xx53RYOFOBbQc1sFumdSjaWyaS/aGQv+knQp4nYvVN0UMFn4
# 0o1i/cvJX0YxULknE+RAMM9yKRAoIsc3Tj2gMj2QzaE4BoVcTlaCKCoFMrdL109j
# 59ItYvFFPeesCAD2RqGe0VuMJlPoeqpK8kbPNzw4nrR3XKUXno3LEY9WPMGsCV8D
# 0wIDAQABo4IBXjCCAVowDwYDVR0TAQH/BAUwAwEB/zAdBgNVHQ4EFgQUyxHoytK0
# FlgByTcuMxYWuUyaCh8wCwYDVR0PBAQDAgGGMBIGCSsGAQQBgjcVAQQFAgMBAAEw
# IwYJKwYBBAGCNxUCBBYEFP3RMU7TJoqV4ZhgO6gxb6Y8vNgtMBkGCSsGAQQBgjcU
# AgQMHgoAUwB1AGIAQwBBMB8GA1UdIwQYMBaAFA6sgmBAVieX5SUT/CrhClOVWeSk
# MFAGA1UdHwRJMEcwRaBDoEGGP2h0dHA6Ly9jcmwubWljcm9zb2Z0LmNvbS9wa2kv
# Y3JsL3Byb2R1Y3RzL21pY3Jvc29mdHJvb3RjZXJ0LmNybDBUBggrBgEFBQcBAQRI
# MEYwRAYIKwYBBQUHMAKGOGh0dHA6Ly93d3cubWljcm9zb2Z0LmNvbS9wa2kvY2Vy
# dHMvTWljcm9zb2Z0Um9vdENlcnQuY3J0MA0GCSqGSIb3DQEBBQUAA4ICAQBZOT5/
# Jkav629AsTK1ausOL26oSffrX3XtTDst10OtC/7L6S0xoyPMfFCYgCFdrD0vTLqi
# qFac43C7uLT4ebVJcvc+6kF/yuEMF2nLpZwgLfoLUMRWzS3jStK8cOeoDaIDpVbg
# uIpLV/KVQpzx8+/u44YfNDy4VprwUyOFKqSCHJPilAcd8uJO+IyhyugTpZFOyBvS
# j3KVKnFtmxr4HPBT1mfMIv9cHc2ijL0nsnljVkSiUc356aNYVt2bAkVEL1/02q7U
# gjJu/KSVE+Traeepoiy+yCsQDmWOmdv1ovoSJgllOJTxeh9Ku9HhVujQeJYYXMk1
# Fl/dkx1Jji2+rTREHO4QFRoAXd01WyHOmMcJ7oUOjE9tDhNOPXwpSJxy0fNsysHs
# cKNXkld9lI2gG0gDWvfPo2cKdKU27S0vF8jmcjcS9G+xPGeC+VKyjTMWZR4Oit0Q
# 3mT0b85G1NMX6XnEBLTT+yzfH4qerAr7EydAreT54al/RrsHYEdlYEBOsELsTu2z
# dnnYCjQJbRyAMR/iDlTd5aH75UcQrWSY/1AWLny/BSF64pVBJ2nDk4+VyY3YmyGu
# DVyc8KKuhmiDDGotu3ZrAB2WrfIWe/YWgyS5iM9qqEcxL5rc43E91wB+YkfRzojJ
# uBj6DnKNwaM9rwJAav9pm5biEKgQtDdQCNbDPTCCBgcwggPvoAMCAQICCmEWaDQA
# AAAAABwwDQYJKoZIhvcNAQEFBQAwXzETMBEGCgmSJomT8ixkARkWA2NvbTEZMBcG
# CgmSJomT8ixkARkWCW1pY3Jvc29mdDEtMCsGA1UEAxMkTWljcm9zb2Z0IFJvb3Qg
# Q2VydGlmaWNhdGUgQXV0aG9yaXR5MB4XDTA3MDQwMzEyNTMwOVoXDTIxMDQwMzEz
# MDMwOVowdzELMAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAOBgNV
# BAcTB1JlZG1vbmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEhMB8G
# A1UEAxMYTWljcm9zb2Z0IFRpbWUtU3RhbXAgUENBMIIBIjANBgkqhkiG9w0BAQEF
# AAOCAQ8AMIIBCgKCAQEAn6Fssd/bSJIqfGsuGeG94uPFmVEjUK3O3RhOJA/u0afR
# TK10MCAR6wfVVJUVSZQbQpKumFwwJtoAa+h7veyJBw/3DgSY8InMH8szJIed8vRn
# HCz8e+eIHernTqOhwSNTyo36Rc8J0F6v0LBCBKL5pmyTZ9co3EZTsIbQ5ShGLies
# hk9VUgzkAyz7apCQMG6H81kwnfp+1pez6CGXfvjSE/MIt1NtUrRFkJ9IAEpHZhEn
# KWaol+TTBoFKovmEpxFHFAmCn4TtVXj+AZodUAiFABAwRu233iNGu8QtVJ+vHnhB
# MXfMm987g5OhYQK1HQ2x/PebsgHOIktU//kFw8IgCwIDAQABo4IBqzCCAacwDwYD
# VR0TAQH/BAUwAwEB/zAdBgNVHQ4EFgQUIzT42VJGcArtQPt2+7MrsMM1sw8wCwYD
# VR0PBAQDAgGGMBAGCSsGAQQBgjcVAQQDAgEAMIGYBgNVHSMEgZAwgY2AFA6sgmBA
# VieX5SUT/CrhClOVWeSkoWOkYTBfMRMwEQYKCZImiZPyLGQBGRYDY29tMRkwFwYK
# CZImiZPyLGQBGRYJbWljcm9zb2Z0MS0wKwYDVQQDEyRNaWNyb3NvZnQgUm9vdCBD
# ZXJ0aWZpY2F0ZSBBdXRob3JpdHmCEHmtFqFKoKWtTHNY9AcTLmUwUAYDVR0fBEkw
# RzBFoEOgQYY/aHR0cDovL2NybC5taWNyb3NvZnQuY29tL3BraS9jcmwvcHJvZHVj
# dHMvbWljcm9zb2Z0cm9vdGNlcnQuY3JsMFQGCCsGAQUFBwEBBEgwRjBEBggrBgEF
# BQcwAoY4aHR0cDovL3d3dy5taWNyb3NvZnQuY29tL3BraS9jZXJ0cy9NaWNyb3Nv
# ZnRSb290Q2VydC5jcnQwEwYDVR0lBAwwCgYIKwYBBQUHAwgwDQYJKoZIhvcNAQEF
# BQADggIBABCXisNcA0Q23em0rXfbznlRTQGxLnRxW20ME6vOvnuPuC7UEqKMbWK4
# VwLLTiATUJndekDiV7uvWJoc4R0Bhqy7ePKL0Ow7Ae7ivo8KBciNSOLwUxXdT6uS
# 5OeNatWAweaU8gYvhQPpkSokInD79vzkeJkuDfcH4nC8GE6djmsKcpW4oTmcZy3F
# UQ7qYlw/FpiLID/iBxoy+cwxSnYxPStyC8jqcD3/hQoT38IKYY7w17gX606Lf8U1
# K16jv+u8fQtCe9RTciHuMMq7eGVcWwEXChQO0toUmPU8uWZYsy0v5/mFhsxRVuid
# cJRsrDlM1PZ5v6oYemIp76KbKTQGdxpiyT0ebR+C8AvHLLvPQ7Pl+ex9teOkqHQ1
# uE7FcSMSJnYLPFKMcVpGQxS8s7OwTWfIn0L/gHkhgJ4VMGboQhJeGsieIiHQQ+kr
# 6bv0SMws1NgygEwmKkgkX1rqVu+m3pmdyjpvvYEndAYR7nYhv5uCwSdUtrFqPYmh
# dmG0bqETpr+qR/ASb/2KMmyy/t9RyIwjyWa9nR2HEmQCPS2vWY+45CHltbDKY7R4
# VAXUQS5QrJSwpXirs6CWdRrZkocTdSIvMqgIbqBbjCW/oO+EyiHW6x5PyZruSeD3
# AWVviQt9yGnI5m7qp5fOMSn/DsVbXNhNG6HY+i+ePy5VFmvJE6P9MYIErzCCBKsC
# AQEwgZAweTELMAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAOBgNV
# BAcTB1JlZG1vbmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEjMCEG
# A1UEAxMaTWljcm9zb2Z0IENvZGUgU2lnbmluZyBQQ0ECEzMAAAEKLHmu13l7pqwA
# AQAAAQowCQYFKw4DAhoFAKCByDAZBgkqhkiG9w0BCQMxDAYKKwYBBAGCNwIBBDAc
# BgorBgEEAYI3AgELMQ4wDAYKKwYBBAGCNwIBFTAjBgkqhkiG9w0BCQQxFgQUdfJC
# NVlgfsuZaaF/vWg10jsYloswaAYKKwYBBAGCNwIBDDFaMFigNoA0AE0AaQBjAHIA
# bwBzAG8AZgB0ACAASABQAEMAIABQAGEAYwBrACAAMgAwADEAMgAgAFIAMqEegBxo
# dHRwOi8vd3d3Lm1pY3Jvc29mdC5jb20vSFBDMA0GCSqGSIb3DQEBAQUABIIBADJh
# kp+br2EQ7N9PBWfv2ruBcqTQQBgnBuUVyAQtCTCcUNx8yuj5wzEgAkjk53L3OA8f
# BOXrZei4JhkLdc+V/vMZ2fItHeL2b2lnAfwEm0UnKgWZ0mWGaJY68rKhlMidXSlT
# 8lsJnLufuUF9DhpeH3T3Rsn+ofHjC1Rhp7vdCBe5LH7risQ+n/uBloD+Neramfhc
# rXX/ggsPgALs1Q8PicWnDXZUPKoCdekKwU90H6shU6hgR6Tla3WDGB6JfqCmoZV5
# yuvJK1QQAdBhqKYIGXFtTU2YFDdHjozOy5gQ0jS7q7G0BZQe7WPv38Mo0HhRhZHz
# tqr4YU1si+X2JpSE9lShggIoMIICJAYJKoZIhvcNAQkGMYICFTCCAhECAQEwgY4w
# dzELMAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1Jl
# ZG1vbmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEhMB8GA1UEAxMY
# TWljcm9zb2Z0IFRpbWUtU3RhbXAgUENBAhMzAAAAkkyir1UnCC51AAAAAACSMAkG
# BSsOAwIaBQCgXTAYBgkqhkiG9w0BCQMxCwYJKoZIhvcNAQcBMBwGCSqGSIb3DQEJ
# BTEPFw0xNTExMTMwMjQ1MTBaMCMGCSqGSIb3DQEJBDEWBBTeJTVXytM7JSFfVjA9
# 6dIWMrErFzANBgkqhkiG9w0BAQUFAASCAQCr+zvPl21elLIYlwxpwSd2/rBdYwLr
# M8ViiHgoqxWtr7NIQikReDrc4zkNjUaftqoKs7P2qrgSVcTPqUzmBFMo9eM5PXAO
# SSMIXTZsTRmU61xaVRLegZma3O2X3zytci3wKd8yo7w49gjhG6W9GS77InSSgN1D
# xj3H31+NxFITvBGp9Q88dVJDomDexco5fvEAKGCSRiIgtY6mxRhSGtt8lMX083V4
# /T/IFfPC1K4b8ZAPno9163C4qyFaXPsxyzHiXoRRrI+FqTy3dv2cABqD/wTpRAwA
# UzwbPAcCX6MiKoFsQaLbpuN84qy8zX4D7jAeLZZUYojaky2stXf+KVQT
# SIG # End signature block
